# =========================================================================
#
# 'low-level' pco to IUT (connection & coding)
#

proc event2pco { gate exp } {
	if {[main_using_gci]} {
		return [gci_event2pco $gate $exp]
		
	} else {
		return [nogci_event2pco $gate $exp]
	}
}
proc gci_event2pco { gate exp } {
	return [TorxGciGenericPco]
}
proc nogci_event2pco { gate exp } {
	global sidinfo gateinfo pcoinfo
	set g [string tolower $gate]
	if {[llength $gateinfo(pcos,$g)] ==1} {
		return [lindex $gateinfo(pcos,$g) 0]
	} elseif {[info exists gateinfo(event2pco,$g)]} {
		if {[string length $gateinfo(event2pco,$g)] == 0} {
			warning "mismatch in coding: no event2pco for gate $g\n"
			return [lindex $gateinfo(pcos,$g) 0]
		} else {
			if {[catch {$gateinfo(event2pco,$g) $gate $exp} result]} {
				warning "error in event2pco $gateinfo(event2pco,$g): $result\n"
				return [lindex $gateinfo(pcos,$g) 0]
			} else {
				return $result
			}
		}
	} else {
		warning "mismatch in coding: no event2pco for gate $g\n"
		return [lindex $gateinfo(pcos,$g) 0]
	}
}

proc pco2sid { pco } {
	global sidinfo
	set i [string tolower $pco]
	return $sidinfo(iut,$pco)
}

proc gate2sid { gate } {
	global sidinfo
	set g [string tolower $gate]
	return $sidinfo(iut,$g)
}

proc encode_event { g e } {
	global gateinfo
	set gg [string tolower $g]
	if {[info exists gateinfo(conversion,$gg)]} {
		if {[string length $gateinfo(conversion,$gg)] == 0} {
			warning "mismatch in coding: no encoder for gate $g\n"
			set result $e
		} else {
			if [catch {$gateinfo(conversion,$gg) $gg $e} result] {
				warning "error in encoder $gateinfo(conversion,$gg): $result\n"
				set result $e
			}
		}
	} else {
		warning "mismatch in coding: no encoder for events of gate $g\n"
		set result $e
	}
	return $result
}

proc decode_event { i e } {
	global gateinfo pcoinfo
	set ii [string tolower $i]
	if {[info exists pcoinfo(conversion,$ii,1)]} {
		if {[string length $pcoinfo(conversion,$ii,1)] == 0} {
			warning "mismatch in coding: no decoder for pco $i\n"
			set result $e
		} else {
			if [catch {$pcoinfo(conversion,$ii,1) $ii $e} result] {
				warning "error in decoder $pcoinfo(conversion,$ii,1): $result\n"
				set result $e
			}
		}
	} else {
		warning "mismatch in coding: no decoder for events of pco $i\n"
		set result $e
	}
	return $result
}

proc convert_value { i j e } {
	global gateinfo pcoinfo
	set ii [string tolower $i]
	set gate $pcoinfo(gate,output,$ii)
	if {[info exists pcoinfo(conversion,$ii,$j)]} {
		if {[string length $pcoinfo(conversion,$ii,$j)] == 0} {
			set result $e
		} else {
			if [catch {$pcoinfo(conversion,$ii,$j) $e} result] {
				warning "error in convertor $pcoinfo(conversion,$ii,$j): $result\n"
				set result $e
			}
		}
	} else {
		warning "mismatch in coding: no coder for value nr $j at pco $i of gate $g\n"
		set result $e
	}
	return $result
}

proc handle_output_timeout {} {
	global gateinfo iutinfo sidinfo pcoinfo
	set delta 1
	foreach i $pcoinfo(pcos) {
		if {$gateinfo(output,$pcoinfo(gate,output,$i))} {
			expect -i $sidinfo(iut,$i) * {
				set res $expect_out(0,string)
				if {[string compare $res ""] != 0} {
					set delta 0
					# ERROR detected
					nonfatal "Unexpected data received at pco '$i':(\n$res\n) <-- end of unexpected data.\n"
					utlog_tofile_concrete "output" $pcoinfo(gate,output,$i) $i [list $res]
				}
			}
		}
	}
	if {$delta} {
		utlog_tofile_concrete "output" "" "" {}
		return [list delta]
	} else {
		return [list error]
	}
}

proc handle_output { pco values } {
	global maininfo pcoinfo
	set gate $pcoinfo(gate,output,$pco)
	utlog_tofile_concrete "output" $gate $pco $values
	set res [decode_event $pco $values ]
	return [list output $pco [list $gate $res ]]
}

proc generate_expect_commands {} {
	global gateinfo iutinfo sidinfo pcoinfo
	# generate expect commands
	# prefix
	set output(start) "expect {\n"
	set output(end) "timeout { set result \[handle_output_timeout \] }\n}\n"
	set iutinfo(expect_command) $output(start)
	foreach i $pcoinfo(pcos) {
	 if {[info exists pcoinfo(gate,output,$i)]} {
	  if {[info exists gateinfo(output,$pcoinfo(gate,output,$i))]} {
		if {$gateinfo(output,$pcoinfo(gate,output,$i))} {
			set pat(start) "-i \$sidinfo(iut,$i) -re \"^"
			set pat(middle) ""
			set pat(end) "\""
			set ass(start) "{ set result \[handle_output $i \[list "
			set ass(middle) ""
			set ass(end) "\]\] }\n"
			set j 1
			while { $j <= $pcoinfo(noofexps,$i) } {
				# each $e has the form:   spectype:iuttype conversionroutine regexp-option
				# add a \r after each \n (we should check that there isn't one yet)
				## regsub -all -- "\n" $gateinfo(regexp,$g,$j) "\\r\\n" res
				## regsub -all -- "\\n" $gateinfo(regexp,$g,$j) "\\r\\n" res
				regsub -all -- "\\n" $pcoinfo(regexp,$i,$j) "\r\n" res
				# escape the [ characters in the string
				regsub -all -- {\[} $res "\\\[" res
				append pat(middle) "($res)"
				append ass(middle) " \$expect_out($j,string) "
				incr j
			}
	 		# set s "-i \$sidinfo(iut,$g) -re \"(\\\[0-9\\]+)\\r\\n\" { set result \[list output \[list $g \[list \$expect_out(0,string) \]\]\] }\n"
			set s ""
 			append s $pat(start) $pat(middle) $pat(end) " " $ass(start) $ass(middle) $ass(end)
			append iutinfo(expect_command) $s
		}
	  }
	 }
	}
	append iutinfo(expect_command) $output(end)
}

# =========================================================================
