# =========================================================================
#
# Initialisation / config file processing
#

proc init_variables {} {
	global iutinfo gateinfo specinfo maininfo configinfo pcoinfo addressinfo cmdinterpinfo cmdarginfo instinfo

	# General Configuration values (that can not be configured by the user)
	set maininfo(timeout) 10

	# defaults for user config values
	set iutinfo(timeout) 11
	set configinfo(IUTTIMEOUT) $iutinfo(timeout)

	set iutinfo(timeout,stderr) 0 ;# don't wait for output on stderr
	set iutinfo(context) {}
	set iutinfo(contextflags) {}

	set maininfo(debug) 1
	set configinfo(DEBUG) $maininfo(debug)
	
	set specinfo(flags) {}
	set adaptorinfo(flags) {}
	set iutinfo(flags) {}
	set instinfo(flags) {}
	set instinfo(program) {}

	set gateinfo(igates) {}
	set gateinfo(ogates) {}
	set gateinfo(gates) {}
	set gateinfo(output,) 0
	set gateinfo(input,) 0

	set pcoinfo(pcos) {}
	set addressinfo(addresses) {}

	set maininfo(nrofstepsdone) 0
	set maininfo(stopnow) 0
	set maininfo(chooseInputs) 0
	set configinfo(CHOOSEINPUTS) $maininfo(chooseInputs)

	set maininfo(gci,use) 0
	set configinfo(USEGCI) $maininfo(gci,use)
	set maininfo(gci,package) gcitcl
	set configinfo(GCIADAPTER) $maininfo(gci,package)

	if {[info exists cmdarginfo(value,seed)]} {
		set specinfo(seed) $cmdarginfo(value,seed)
	} else {
		set specinfo(seed) [expr int(srand([clock clicks]) * 1000 * rand())]
		# set specinfo(seed) 5
	}
	set configinfo(SEED) $specinfo(seed)

	if {[info exists cmdarginfo(value,configfile)]} {
		set maininfo(configfile) $cmdarginfo(value,configfile)
		set configinfo(CONFIGFILE) $maininfo(configfile)
	}

	if {[info exists cmdarginfo(value,tracefile)]} {
		set maininfo(tracefile) $cmdarginfo(value,tracefile)
		set configinfo(TRACEFILE) $maininfo(tracefile)
	}

	if {[info exists cmdarginfo(value,logfile)]} {
		set maininfo(logfile) $cmdarginfo(value,logfile)
		set configinfo(LOGFILE) $maininfo(logfile)
	}

	set specinfo(timeout) -1 ;# effectively disables timeout for spec!!!
	set configinfo(SPECTIMEOUT) $specinfo(timeout)
	
	if {[info exists cmdarginfo(value,depth)]} {
		set cmdinterpinfo(stepstodo) $cmdarginfo(value,depth)
		set cmdinterpinfo(mode,auto) 1
		set maininfo(nrofstepsmax)  $cmdarginfo(value,depth)
	} else {
		set cmdinterpinfo(stepstodo) 0
		set cmdinterpinfo(mode,auto) 0
		set maininfo(nrofstepsmax) 0
	}
	set configinfo(MAXSTEPS) $maininfo(nrofstepsmax)
	set cmdinterpinfo(automode,usetrace) 0
	set cmdinterpinfo(prompt) "tester> "
	
	# initially everything goes fine -- variable will be set to fail on error
	set maininfo(verdict) inconclusive

	# set initial delta label for us.
	# Note that this (from now on) will be used in a symmetric way, i.e.
	# both in the menu's that we get from the Primer/Combinator,
	# and also in the (input, output) events (or suggestions) that the
	# we send to the Primer/Combinator.
	set specinfo(label,delta) "Delta"
	set configinfo(LABEL-DELTA) $specinfo(label,delta)
	set maininfo(parsed-label,delta) $specinfo(label,delta)
}

proc process_config_line_words { file line infolist } {
	global iutinfo gateinfo specinfo maininfo configinfo pcoinfo addressinfo cmdinterpinfo cmdarginfo instinfo

	set iomap(INPUT) input
	set iomap(OUTPUT) output

	foreach e $infolist {
		set c [string toupper [lindex $e 0]]
		if {[string compare "INCLUDE" $c] == 0} {
			set files [lrange $e 1 end]
			set configinfo($c) $iutinfo(timeout)
			foreach file $files {
				read_config_file $file
			}
		} elseif {[string compare "IUTTIMEOUT" $c] == 0} {
			set iutinfo(timeout) [lindex $e 1]
			set configinfo($c) $iutinfo(timeout)
		} elseif {[string compare "SPECTIMEOUT" $c] == 0} {
			set specinfo(timeout) [lindex $e 1]
			set configinfo($c) $specinfo(timeout)
		} elseif {([string compare "MAXSTEPS" $c] == 0)} {
			if {![info exists cmdarginfo(flag,depth)]} {
				set maininfo(nrofstepsmax)  [lindex $e 1]
				set configinfo($c) $maininfo(nrofstepsmax)
			} else {
				warning "config $file:$line: commandline flag $cmdarginfo(flag,depth) overrules $c\n"
			}
		} elseif {([string compare "CHOOSEINPUTS" $c] == 0)} {
			set maininfo(chooseInputs)  [lindex $e 1]
			set configinfo($c) $maininfo(chooseInputs)
		} elseif {([string compare "USEGCI" $c] == 0)} {
			set maininfo(gci,use)  [lindex $e 1]
			set configinfo($c) $maininfo(gci,use)
		} elseif {([string compare "GCIADAPTER" $c] == 0)} {
			set maininfo(gci,package)  [lindex $e 1]
			set configinfo($c) $maininfo(gci,package)
		} elseif {([string compare "LOGFILE" $c] == 0)} {
			if {![info exists cmdarginfo(flag,logfile)]} {
				set maininfo(logfile)  [lindex $e 1]
				set configinfo($c) $maininfo(logfile)
			} else {
				warning "config $file:$line: commandline flag $cmdarginfo(flag,logfile) overrules $c\n"
			}
		} elseif {([string compare "LOGMON" $c] == 0)} {
			# skip
		} elseif {([string compare "TRACEFILE" $c] == 0)} {
			if {![info exists cmdarginfo(flag,tracefile)]} {
				set maininfo(tracefile)  [lindex $e 1]
				set configinfo($c) $maininfo(tracefile)
			} else {
				warning "config $file:$line: commandline flag $cmdarginfo(flag,tracefile) overrules $c\n"
			}
		} elseif {[string compare "IUTSOURCE" $c] == 0} {
			set specinfo(source,iut) [lindex $e 1]
			set configinfo($c) $specinfo(source,iut)
		} elseif {[string compare "SOURCEIUT" $c] == 0} {
			# this one remains only for backwards compatibility
			set new_c IUTSOURCE
			warning "config $file:$line: $c is deprecated. Use $new_c instead.\n"
			set specinfo(source,iut) [lindex $e 1]
			set configinfo($new_c) $specinfo(source,iut)
		} elseif {[string compare "SPECSOURCE" $c] == 0} {
			set specinfo(source,spec) [lindex $e 1]
			set configinfo($c) $specinfo(source,spec)
		} elseif {[string compare "SOURCESPEC" $c] == 0} {
			# this one remains only for backwards compatibility
			set new_c SPECSOURCE
			warning "config $file:$line: $c is deprecated. Use $new_c instead.\n"
			set specinfo(source,spec) [lindex $e 1]
			set configinfo($new_c) $specinfo(source,spec)
		} elseif {([string compare "SOURCE" $c] == 0)} {
			# skip
		} elseif {[string compare "PRIMERS" $c] == 0} {
			set configinfo($c) [lindex $e 1]
		} elseif {[string compare "PRIMER" $c] == 0} {
			set configinfo($c) [lindex $e 1]
		} elseif {[string compare "GUIDES" $c] == 0} {
			set configinfo($c) [lindex $e 1]
		} elseif {[string compare "GUIDE" $c] == 0} {
			set configinfo($c) [lindex $e 1]
		} elseif {[string compare "MUTANTS" $c] == 0} {
			set configinfo($c) [lindex $e 1]
		} elseif {[string compare "MUTANT" $c] == 0} {
			set configinfo($c) [lindex $e 1]
		} elseif {([string compare "SEED" $c] == 0)} {
			if {![info exists cmdarginfo(flag,seed)]} {
				set specinfo(seed)  [lindex $e 1]
				set configinfo($c) $specinfo(seed)
			} else {
				warning "config $file:$line: commandline flag $cmdarginfo(flag,seed) overrules $c\n"
			}
		} elseif {([string compare "PROMPT" $c] == 0)} {
			set cmdinterpinfo(prompt)  [lindex $e 1]
			set configinfo($c) $cmdinterpinfo(prompt)
		} elseif {([string compare "DEBUG" $c] == 0)} {
			set maininfo(debug)  [lindex $e 1]
			set configinfo($c) $maininfo(debug)
		} elseif {([string compare "CODING" $c] == 0)} {
			# first try to find coding dir relative to current directory;
			# if that fails, try to find it relative to libdir of the tester;
			# if that fails as well, we give up.
			# We should probably make a proc for this, so that we can use it
			# for more (every?) file(s) that we open.
			if {[catch {glob  [lindex $e 1]} foo] &&
			    [catch {glob [file join $maininfo(libdir) [lindex $e 1]]} foo]} {
				fatal "config $file:$line: Could not find specified Conversion directory: [lindex $e 1]";
			}
			set maininfo(conversiondir)  $foo
			set configinfo($c) $maininfo(conversiondir)
		} elseif {([string compare "IUT" $c] == 0)} {
			if {[catch {glob  [lindex $e 1]} foo]} {
				fatal "config $file:$line: Could not find specified IUT program: [lindex $e 1]";
			}
			set iutinfo(program) $foo
			set configinfo($c) $iutinfo(program)
		} elseif {([string compare "IUTRUNDIR" $c] == 0)} {
			set iutinfo(program,run,dir) [lindex $e 1]
			## we should check here
			set configinfo($c) $iutinfo(program,run,dir)
		} elseif {([string compare "IUTCONTEXT" $c] == 0)} {
			set iutinfo(context) [join [lrange $e 1 end] " "]
			set configinfo($c) $iutinfo(context)
		} elseif {([string compare "IUTCONTEXTFLAGS" $c] == 0)} {
			set iutinfo(contextflags) [join [lrange $e 1 end] " "]
			set configinfo($c) $iutinfo(contextflags)
		} elseif {([string compare "IUTFLAGS" $c] == 0)} {
			lappend iutinfo(flags) [join [lrange $e 1 end] " "]
			lappend configinfo($c) $iutinfo(flags)
		} elseif {([string compare "INST" $c] == 0)} {
			if {![catch {glob  [lindex $e 1]} foo]} {
			} elseif {![catch {glob  [file join $maininfo(bindir) [lindex $e 1]]} foo]} {
			} elseif {![catch {glob   [file join $maininfo(libexecdir) [lindex $e 1]]} foo]} {
			} else {
				fatal "config $file:$line: Could not find specified INST program: [lindex $e 1]";
			}
			set instinfo(program) $foo
			set configinfo($c) $instinfo(program)
		} elseif {([string compare "INSTRUNDIR" $c] == 0)} {
			set instinfo(program,run,dir) [lindex $e 1]
			## we should check here
			set configinfo($c) $instinfo(program,run,dir)
		} elseif {([string compare "INSTFLAGS" $c] == 0)} {
			lappend instinfo(flags) [join [lrange $e 1 end] " "]
			lappend configinfo($c) $instinfo(flags)
		} elseif {([string compare "ADAPTOR" $c] == 0)} {
			# skip
		} elseif {([string compare "ADAPTORRUNDIR" $c] == 0)} {
			# skip
		} elseif {([string compare "ADAPTORCONTEXT" $c] == 0)} {
			# skip
		} elseif {([string compare "ADAPTORFLAGS" $c] == 0)} {
			lappend adaptorinfo(flags) [join [lrange $e 1 end] " "]
			lappend configinfo($c) $adaptorinfo(flags)
		} elseif {([string compare "SPEC" $c] == 0)} {
			if {![catch {glob [lindex $e 1]} foo]} {
			} elseif {![catch {glob [file join $maininfo(libexecdir) [lindex $e 1]]} foo]} {
			} elseif {![catch {glob [file join $maininfo(bindir) [lindex $e 1]]} foo]} {
			} else {
				fatal "config $file:$line: Could not find specified SPEC program: [lindex $e 1]";
			}
			set specinfo(program) $foo
			set configinfo($c) $specinfo(program)
		} elseif {([string compare "SPECRUNDIR" $c] == 0)} {
			set specinfo(program,run,dir) [lindex $e 1]
			## we should check here
			set configinfo($c) $specinfo(program,run,dir)
		} elseif {([string compare "SPECFLAGS" $c] == 0)} {
			lappend specinfo(flags) [join [lrange $e 1 end] " "]
			lappend configinfo($c) $specinfo(flags)
		} elseif {([string compare "LABEL-DELTA" $c] == 0)} {
			set specinfo(label,delta) [lrange $e 1 end]
			set maininfo(parsed-label,delta) $specinfo(label,delta)
			set configinfo($c) $specinfo(label,delta)
		} elseif {([string compare "INPUT" $c] == 0) || ([string compare "OUTPUT" $c] == 0)} {
			lappend configinfo($c) [lrange $e 1 end]
			set g [string tolower [lindex $e 1]]
			lappend gateinfo(gates) $g
			if {([string compare "INPUT" $c] == 0)} {
				lappend gateinfo(igates) $g
				set gateinfo(input,$g) 1
				set gateinfo(output,$g) 0
			} else {
				lappend gateinfo(ogates) $g
				set gateinfo(input,$g) 0
				set gateinfo(output,$g) 1
			}
			set gateinfo(io,$g) $iomap($c)
			set gateinfo(val,$g) [lindex $e 2]
			set i 1
			foreach ee $gateinfo(val,$g) {
				set gateinfo(types,$g,$i) [lindex $ee 0]
				incr i
			}
			set gateinfo(noofexps,$g) [llength $gateinfo(val,$g)]
			set gateinfo(conversion,$g) [lindex $e 3]
			set gateinfo(event2pco,$g) [lindex $e 4]
		} elseif {[string compare "ADDRESS" $c] == 0} {
			lappend configinfo($c) [lrange $e 1 end]
			set a [string tolower [lindex $e 1]]
			lappend addressinfo(addresses) $a
			set addressinfo(val,$a) [lrange $e 2 end]
			set addressinfo(proto,$a) [lindex $e 2]
			set addressinfo(comm,$a) [lindex $e 2]
			set addressinfo(host,$a) [lindex $e 3]
			set addressinfo(port,$a) [lindex $e 4]
			if {[string compare [lindex $e 6] ""] != 0} {
				set addressinfo(comm,$a) [lindex $e 5]
				set addressinfo(args,$a) [lindex $e 6]
			} else {
				set addressinfo(args,$a) [lindex $e 5]
			}
			if {[string compare [string tolower $addressinfo(proto,$a)] "pipe"] == 0} {
				set addressinfo(iut,$a) 1
			} else {
				set addressinfo(iut,$a) 0
			}
			if {([string compare [string toupper $addressinfo(host,$a)] "CURRENTHOST"] == 0)} {
				#set addressinfo(host,$a) [info hostname]
				## HACK to get rid of domain name
				#regexp {^([^.]+)\..*} $addressinfo(host,$a) \
				#	 match addressinfo(host,$a)
				# HACK to get name from network database:
				#   gethostbyaddr( gethostbyname( hostname() ), AF_INET )
				set addressinfo(host,$a) [exec torx-hostname]
			}
			if {([string compare $addressinfo(host,$a) ""] != 0) &&
			    ([string compare $addressinfo(port,$a) ""] != 0)} {
				set addressinfo(hostport,$addressinfo(host,$a),$addressinfo(port,$a)) $a
			}
		} elseif {[string compare "PCO" $c] == 0} {
			lappend configinfo($c) [lrange $e 1 end]
			# pco (pco) name:
			set i [string tolower [lindex $e 1]]
			# send_error "PCO bla: [llength $e]\n"
			lappend pcoinfo(pcos) $i
			if {[llength $e] > 2} {
			  # gate name:
			  set gates [lindex $e 2]
			  set io_list {}
			  foreach gate $gates {
				set g [string tolower $gate]
				lappend gateinfo(pcos,$g) $i
				if {$gateinfo(output,$g)} {
					set io output
				} else {
					set io input
				}
				if {[info exists pcoinfo(gate,$io,$i)]} {
						warning "config $file:$line: $io gate for PCO $i already set to $pcoinfo(gate,$io,$i), ignoring $g\n"
				} else {
					set pcoinfo(gate,$io,$i) $g
					lappend io_list $io
				}
			  }
			  set pcoinfo(io,$i) [join [lsort $io_list] "/"]
			  foreach v [list input output] {
				if {![info exists pcoinfo(gate,$v,$i)]} {
					set pcoinfo(gate,$v,$i) ""
				}
			  }
			  # address name:
			  set a [string tolower [lindex $e 3]]
			  set pcoinfo(address,$i) $a
			  set pcoinfo(pco,$g,$a) $i
			  if {[string compare $pcoinfo(address,$i) "iutpipe"] == 0} {
				set pcoinfo(iut,$i) 1
			  } else {
				set pcoinfo(iut,$i) 0
			  }
			  # type, conversion routine, regexp:
			  set pcoinfo(val,$i) [lrange $e 4 end]
			  set j 1
			  foreach ee $pcoinfo(val,$i) {
				# each $e has the form:
				#	INPUT: iuttype conversionroutine
				#	OUTPUT: iuttype conversionroutine regexp
				set pcoinfo(types,$i,$j) [lindex $ee 0]
				set pcoinfo(conversion,$i,$j) [lindex $ee 1]
				set pcoinfo(regexp,$i,$j) [lindex $ee 2]
				incr j
			  }
			  set pcoinfo(noofexps,$i) [llength $pcoinfo(val,$i)]
			}
		} else {
			warning "config $file:$line: unexpected: $e\n"
			lappend configinfo(unexpected) $e
		}
	}
}

proc config_consistency_check {} {
	global pcoinfo
	if {[llength $pcoinfo(pcos)] == 0} {
		warning "config: no PCOs defined in config file(s)\n"
		return 0
	}
	return 1
}

proc generate_spec_command {} {
	global specinfo gateinfo
	if {[info exists specinfo(program)]} {
		if {[info exists specinfo(program,run,dir)]} {
			set cdcmd "cd $specinfo(program,run,dir)"
			set program $specinfo(program)
		} else {
			set cdcmd "cd [file dirname $specinfo(program)]"
			set program [file join . [file tail $specinfo(program)]]
		}
		if {[info exists specinfo(label,delta)]} {
			set delta_gate_string "-d \"$specinfo(label,delta)\""
		} else {
			set delta_gate_string ""
		}
		set specinfo(cmd) [list spawn -noecho "/bin/sh" -c "stty -echo; $cdcmd; $program -s $specinfo(seed) -i [join $gateinfo(igates) {,}] -o [join $gateinfo(ogates) {,}] $delta_gate_string "]
		return 1
	} else {
		return 0
	}
}

proc initialise_random_seed {} {
	global specinfo
	RandomInit $specinfo(seed)
}

proc main_using_gci {} {
	global maininfo
	if {[info exists maininfo(gci,use)]} {
		return $maininfo(gci,use)
	} else {
		return 0
	}
}

# =========================================================================
