# =========================================================================

# Based on 
# Library chapter 'Practical Programming in Tcl and Tk' by Brent Welch
proc Package_UpdateIndex { libdir } {
	global maininfo
	set sharedlibsuffix [TorXSharedLibSuffix]
	if ![file exists $libdir/pkgIndex.tcl] {
		set doit 1
	} else {
		set age [file mtime $libdir/pkgIndex.tcl]
		set doit 0
		foreach file [glob -nocomplain $libdir/*$sharedlibsuffix] {
			if {[file mtime $file] > $age} {
				set doit 1
				break
			}
		}
	}
	if { $doit } {
		if {[catch {pkg_mkIndex $libdir *$sharedlibsuffix} foo]} {
			warning "could not update $libdir/pkgIndex.tcl: $foo"
		}
	}
}

proc gci_get_object_library {} {
	global maininfo auto_path
	# send_error "get_obj_lib: \n"
	if {[info exists maininfo(conversiondir)]} {
		# send_error "get_obj_lib: $maininfo(conversiondir)\n"
		Library_UpdateIndex $maininfo(conversiondir)
		Package_UpdateIndex $maininfo(conversiondir)
		lappend auto_path $maininfo(conversiondir)
	} else {
		warning "no CODING directory specified;"
		warning "so, loading package $maininfo(gci,package) may fail"
		warning "as fallback, we add the lib directory of the installation"
		lappend auto_path $maininfo(plugindir)
		# send_error "get_obj_lib: maininfo(conversiondir) does not exist\n"
	}
	package require $maininfo(gci,package)
}

proc GciQueueIsEmpty {} {
	global gciqueue
	if {[llength $gciqueue] <= 0} {
		return 1
	} else {
		return 0
	}
}
proc GciQueueHead {} {
	global gciqueue
	set queue_el [lindex $gciqueue 0]
	set gciqueue [lrange $gciqueue 1 end]
	return $queue_el
}
#
# We have seen it happen that within a single call to GciSnapshot
# _both_ GciTimeout _and_ GciReceive have been called, so
# we take precautions to remove 'superflouous' delta events
# from the queue (i.e. if we have both deltas and outputs
# in the queue, then we remove the deltas).
#
# Because we know how the GciSnapshot C routine is implemented,
# and we know that it first calls UpdateTimers and then UpdatePCOs,
# we know that the timeoutCallbackProc will be called before
# receiveCallbackProc, so in timeoutCallbackProc we set a flag when
# we insert a delta in the queue, and in receiveCallbackProc
# we check this flag and, if necessary, get rid of unwanted delta's.
#
# The Tcl GciSnapshot routine calls gci_queue_reset to reset
# the information that we have about the queue (like: the flag
# that says whether deltas have been inserted or not).
#
# Additionally, the Tcl GciSnapshot routine calls gci_queue_cleanup
# when the C code has returned, to allow us to clean up the queue.
# Currently we don't do that, but in principle we might use that
# routine to walk through the queue, and remove deltas from the
# queue if it turns out that the queue contains both deltas and
# outputs.
#
proc receiveCallbackProc {args} {
	global gciqueue gci_queue_delta_inserted
	send_error "receiveCallbackProc $args\n"
	if {$gci_queue_delta_inserted} {
		send_error "receiveCallbackProc $args: cleared gciqueue from deltas\n"
		set gciqueue {}
	}
	lappend gciqueue [list output [lindex $args 1]]
	send_error "receiveCallbackProc-gciqueue $gciqueue\n"
}
proc timeoutCallbackProc {args} {
	global gciqueue gci_queue_delta_inserted
	send_error "timeoutCallbackProc $args\n"
	incr gci_queue_delta_inserted
	lappend gciqueue [list delta]
	send_error "timeoutCallbackProc-gciqueue $gciqueue\n"
}
proc gci_queue_reset {} {
	global gciqueue gci_queue_delta_inserted
	set gci_queue_delta_inserted 0
}
proc gci_queue_cleanup {} {
	global gciqueue gci_queue_delta_inserted
	set temp_queue {}
	foreach q_el $gciqueue {
		
	}
}
proc GciSnapshot {} {
	send_error "GciSnapshot\n"
	gci_queue_reset
	set result [gci_snapshot_cmd receiveCallbackProc timeoutCallbackProc]
	gci_queue_cleanup
	return $result
}
proc GciSend {pco value} {
	send_error "GciSend $pco $value\n"
	return [gci_send_cmd $pco $value]
}

proc GciStartTimer {timer value unit} {
	send_error "GciStartTimer $timer $value $unit\n"
	return [gci_start_timer_cmd $timer $value $unit]
}
proc GciCancelTimer {timer} {
	send_error "GciCancelTimer $timer\n"
	return [gci_cancel_timer_cmd $timer]
}
proc GciReadTimer {timer} {
	send_error "GciReadTimer $timer\n"
	return [gci_read_timer_cmd $timer]
}

# =========================================================================
#
proc GciInit {} {
	global gciqueue
	send_error "GciInit\n"
	gci_get_object_library
	set gciqueue {}
	return [gci_init_cmd]
}
proc GciFinalize {} {
	send_error "GciFinalize\n"
	return [gci_finalize_cmd]
}
proc GciGetGenericPco {} {
	send_error "GciGetGenericPco\n"
	return [findPcoByName TorXAnyGeneric]
}

# =========================================================================
#

proc mkEvent {ev} {
	return [list nodes [event2term $ev]]
}
# =========================================================================
#
# interface to GCI
#  - initialize (to do)
#  - translating event in a version in which the PCO is identified;
#  - build a term representation that can be passed via GCI
#  - send a term via GCI
#  - receive a term from GCI
#


proc buildPcoNameMapping {args} {
	global pco2name name2pco
	set i [array size pco2name]
	foreach p $args {
		if {[info exists name2pco($p)]} {
			send_error "buildPcoNameMapping: called more than once for PCO: $p\n"
		} else {
			set pco2name($i) $p
			set name2pco($p) $i
			incr i
		}
	}
}
proc findPcoByName {pcoName} {
	global name2pco
	if {[info exists name2pco($pcoName)]} {
		return $name2pco($pcoName)
	} else {
		return -1
	}
}

# XXX: DO NOT USE THIS but instead use info from the .if file !!!!
## NOW done in gci.tcl TorxGciInit
# 'real' PCOs: rc5 bic
# TorXAnyGeneric is a generic PCO to be 'mapped' on one of the 'real' PCOs,
# based on the value that is to be Send

## buildPcoNameMapping rc5 bic TorXAnyGeneric

# int GciGetNoOfPCOs()
proc GciGetNoOfPCOs { args } {
	global pco2name
	send_error "GciGetNoOfPCOs $args\n"
	if {[info exists pco2name]} {
		send_error "GciGetNoOfPCOs $args: [array size pco2name]\n"
		return [array size pco2name]
	} else {
		send_error "GciGetNoOfPCOs: internal error: array does not exist: pco2name\n"
		return 0
	}
}

# char* GciGetPCOName(int PCOD)
proc GciGetPCOName { pcod args } {
	global pco2name
	send_error "GciGetPCOName $pcod $args\n"
	if {[info exists pco2name($pcod)]} {
		set result $pco2name($pcod)
	} else {
		send_error "GciGetPCOName $args: unexpected pcod $pcod\n"
		set result ""
	}
	return $result
}

# char* GciGetPCOType(int PCOD)
proc GciGetPCOType { pcod  args } {
	send_error "GciGetPCOType $pcod $args\n"
}

# GciPCORole GciGetPCORole(int PCOD)
proc GciGetPCORole { pcod args } {
	send_error "GciGetPCORole $pcod $args\n"
}

# GciValueSTAR GciGetValue(CONST char *value)
proc GciGetPCORole { v args } {
	send_error "GciGetValue $v $args\n"
}

# int GciGetPCO(int PCOD)
proc GciGetPCO { pcod args } {
	send_error "GciGetPCO $pcod $args\n"
	return $pcod
}

# =========================================================================
