# =========================================================================
#
# start IUT (if needed) and connect to it - interface routine
#

proc start_iut {} {
	global maininfo
	if {[main_using_gci]} {
	     gci_start_iut
	} else {
		nogci_start_iut
	}
}

# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
#
# start IUT (if needed) and connect to it - nogci version
#

proc nogci_start_iut {} {
	global iutinfo sidinfo maininfo iutsids sid2pco
	match_max -d 10000
	if { [info exists iutinfo(program)] && 
	     (![string compare $iutinfo(program) ""] == 0) } {
		### set iutinfo(pid) [eval exec $iutinfo(program) &]
		# open a pty from which we will read the stderr output of IUT
		# (we will redirect the stderr of IUT to this pty)
		if {[catch {spawn -pty} foo]} {
			fatal "spawn for stderr of $iutinfo(program) fails: $foo"
		}
		set sidinfo(iut,stderr) $spawn_id
		set iutinfo(pty,stderr) $spawn_out(slave,name)
		# build the command to start the IUT:
		# - we disable echo-ing through the shell stty command 
		# - the iut runs in the 'hexcontext' context
		# - the stderr of the iut is redirected to the pty we spawned above
		# set command [list spawn -noecho "/bin/sh" -c "stty -echo; hexcontext -- $iutinfo(program) 2> $iutinfo(pty,stderr)"] 
		# set command [list spawn -noecho "/bin/sh" -c "stty -echo; $iutinfo(program) > /dev/pts/3 2>&1"] 
		if {[info exists iutinfo(program,run,dir)]} {
			set cdcmd "cd $iutinfo(program,run,dir)"
			set program $iutinfo(program)
		} else {
			set cdcmd "cd [file dirname $iutinfo(program)]"
			set program [file join . [file tail $iutinfo(program)]]
		}
		set args [join $iutinfo(flags) " "]
		set command [list spawn -noecho "/bin/sh" -c "stty -echo; $cdcmd; $program > $iutinfo(pty,stderr) 2>&1"] 
		set command [list spawn -noecho "/bin/sh" -c "stty -echo; exec 2> $iutinfo(pty,stderr); $cdcmd; exec $iutinfo(context) $iutinfo(contextflags) $program $args"] 
		debug 0 "$command\n"
		send_error "Adapter starting IUT: $command\n"
		if {[catch $command foo]} {
			fatal "spawn for $iutinfo(program) fails: $foo"
		}
		set sidinfo(iut) $spawn_id
		set iutinfo(pid) [exp_pid -i $sidinfo(iut)]
		set timeout 1
		expect {
			-i $sidinfo(iut) -re "No such file or directory" {
				fatal "Could not start IUT: program can not be found\n"
			}
			-i $sidinfo(iut,stderr) -re "No such file or directory" {
				fatal "Could not start IUT: program can not be found\n"
			}
			-i $sidinfo(iut) -re "exec: (.*): not found" {
				set s $expect_out(1,string)
				fatal "Could not start IUT: program $s can not be found\n"
			}
			-i $sidinfo(iut,stderr) -re "exec: (.*): not found" {
				set s $expect_out(1,string)
				fatal "Could not start IUT: program $s can not be found\n"
			}
			-i $sidinfo(iut) eof {
				unset sidinfo(iut)
				fatal "Could not start IUT: eof from IUT: $expect_out(buffer)\n"
			}
			-i $sidinfo(iut,stderr) eof {
				unset sidinfo(iut)
				fatal "Could not start IUT: eof from IUT: $expect_out(buffer)\n"
			}
			timeout {
				send_error "IUT started\n"
			}
		}
		set timeout $maininfo(timeout)
		expect_background -i $sidinfo(iut,stderr) -re "(\[^\n\r]*)\r*\n" {
			set l $expect_out(1,string)
			if {[regexp -- "^LOG\[ \t\]*(\[^\n\r]*)" $l match txt]} {
				send_user "A_LOG $txt\n"
			} else {
				send_error "IUT Stderr: $l\n"
			}
		}
		expect_before -i $sidinfo(iut) -re "(ERROR (\[^\n]*)\n)+" {
			foreach l [split $expect_out(buffer) "\n"] {
				regexp -- "^ERROR (\[^\n\]*)" $l match sub
				send_error "IUT Error: $sub\n"
			}
		}
	} else {
		set iutinfo(pid) ""
	}
}

# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
#
# start IUT (if needed) and connect to it - gci version
#

proc gci_start_iut {} {
	global env iutinfo
	# HACK set env var TORXIUTPROG, to pass the IUT value
	# to the simulator IUT Adapter.
	if { [info exists iutinfo(program)] && 
	     (![string compare $iutinfo(program) ""] == 0) } {
		set env(TORXIUTPROG) $iutinfo(program)
	}
	if { [info exists iutinfo(program,run,dir)] && 
	     (![string compare $iutinfo(program,run,dir) ""] == 0) } {
		set env(TORXIUTRUNDIR) $iutinfo(program,run,dir)
	} elseif { [info exists iutinfo(program)] && 
	     (![string compare $iutinfo(program) ""] == 0) } {
		set env(TORXIUTRUNDIR) [file dirname $iutinfo(program) ]
		set env(TORXIUTPROG) [file join . [file tail $iutinfo(program) ]]
	}
	TorxGciInit
}

# =========================================================================
#
# connect to running IUT - interface routine
#

proc connect_to_iut {} {
	global maininfo
	if {[main_using_gci]} {
	     gci_connect_to_iut
	} else {
		nogci_connect_to_iut
	}
}

# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
#
# connect to running IUT - nogci version
#

proc nogci_connect_to_iut {} {
	global pcoinfo sidinfo addressinfo gateinfo maininfo iutsids sid2pco
	set iutsids {}
	expect_before -i iutsids eof {
		global iutsids sid2pco pcoinfo
		set sid $expect_out(spawn_id)
		set p $sid2pco($sid)
		set ind [lsearch -exact $sid $iutsids]
		if {$ind >= 0} {
			set iutsids [lreplace $iutsids $ind $ind]
		}
		to_user "IUT Error" $expect_out(buffer)
		fatal "Eof from IUT on $pcoinfo(io,$p) pco \`\`$p\'\': $expect_out(buffer)\n"
		set result error
		exp_continue
	}
	foreach i $pcoinfo(pcos) {
		if {($addressinfo(iut,$pcoinfo(address,$i)))} {
			# nothing to do
			if {[info exists sidinfo(iut)]} {
				set sidinfo(iut,$i) $sidinfo(iut)
				if {![info exists sidinfo(iutexpect)]} {
					lappend iutsids $sidinfo(iut)
					set sid2pco($sidinfo(iut)) $i
					set sidinfo(iutexpect) 1
				}
			} elseif {[info exists iutinfo(program)]} {
				fatal "Error: missing connection info for $pcoinfo(io,$i) pco \`\`$i\'\' (failed to spawn IUT?)"
			} else {
				fatal "Error: missing connection info for $pcoinfo(io,$i) pco \`\`$i\'\' (not configured to start IUT)"
			}
		} elseif {![info exists sidinfo(connect,$addressinfo(comm,$pcoinfo(address,$i)),$addressinfo(host,$pcoinfo(address,$i)),$addressinfo(port,$pcoinfo(address,$i))) ]} {
			if {[catch {spawn -pty} foo]} {
				fatal "spawn for stderr of pco $i fails: $foo"
			}
			set sidinfo(iut,$i,stderr) $spawn_id
			set sidinfo(name,$spawn_id) $i
			set iutinfo(pty,$i,stderr) $spawn_out(slave,name)
			if {[string compare $addressinfo(args,$pcoinfo(address,$i)) ""] == 0} {
				set command [list spawn -noecho "/bin/sh" -c "stty -echo; exec 2> $iutinfo(pty,$i,stderr); exec $addressinfo(comm,$pcoinfo(address,$i)) $addressinfo(host,$pcoinfo(address,$i)) $addressinfo(port,$pcoinfo(address,$i))"]
			} else {
				# set up variables that we want to offer to the command
				set args $addressinfo(args,$pcoinfo(address,$i))
				regsub -all -- {\$\{P\}} $args $addressinfo(port,$pcoinfo(address,$i)) args
				regsub -all -- {\$\{H\}} $args $addressinfo(host,$pcoinfo(address,$i)) args
				set command [list spawn -noecho "/bin/sh" -c "stty -echo; exec 2> $iutinfo(pty,$i,stderr); exec $addressinfo(comm,$pcoinfo(address,$i)) $args"]
			}
			debug 0 "$command\n"
			send_error "Adapter connecting to IUT: $command\n"
			if {[catch $command foo]} {
				fatal "spawn for pco \`\`$i\'\' (command: $command) fails: $foo"
			}
			set sidinfo(iut,$i) $spawn_id
			set timeout $maininfo(timeout)
			expect_background -i $sidinfo(iut,$i,stderr) -re "(\[^\n\r]*)\r*\n" {
				set l $expect_out(1,string)
				set i $expect_out(spawn_id)
				if {[regexp -- "^LOG\[ \t\]*(\[^\n\r]*)" $l match txt]} {
					send_user "A_LOG $txt\n"
				} else {
					send_error "PCO $sidinfo(name,$i) Stderr: $l\n"
				}
			}
			set progress_routine_name "connection_progress_$addressinfo(comm,$pcoinfo(address,$i))"
			if {[catch {$progress_routine_name $sidinfo(iut,$i) $pcoinfo(io,$i) $i} foo]} {
				# no specific connection-checking routine
				# so, we have to run expect to check for errors
				send_error "(No specific connection setup checker available for \`\`$addressinfo(comm,$pcoinfo(address,$i))\'\' ($foo), using generic one)\n"
				set timeout 1
				expect {
					-i $sidinfo(iut,$i) -re "No such file or directory" {
						fatal "Connection program \`\`$addressinfo(comm,$pcoinfo(address,$i))\'\' can not be found while connecting to $pcoinfo(io,$i) pco \`\`$i\'\'\n"
					}
					-i $sidinfo(iut,$i,stderr) -re "No such file or directory" {
						fatal "Connection program \`\`$addressinfo(comm,$pcoinfo(address,$i))\'\' can not be found while connecting to $pcoinfo(io,$i) pco \`\`$i\'\'\n"
					}
					-i $sidinfo(iut,$i) -re "exec: $addressinfo(comm,$pcoinfo(address,$i)): not found" {
						fatal "Connection program \`\`$addressinfo(comm,$pcoinfo(address,$i))\'\' can not be found while connecting to $pcoinfo(io,$i) pco \`\`$i\'\'\n"
					}
					-i $sidinfo(iut,$i,stderr) -re "exec: $addressinfo(comm,$pcoinfo(address,$i)): not found" {
						fatal "Connection program \`\`$addressinfo(comm,$pcoinfo(address,$i))\'\' can not be found while connecting to $pcoinfo(io,$i) pco \`\`$i\'\'\n"
					}
					-i $sidinfo(iut,$i) eof {
						fatal "Eof from IUT while connecting to $pcoinfo(io,$i) pco \`\`$i\'\': $expect_out(buffer)\n"
					}
					-i $sidinfo(iut,$i,stderr) eof {
						fatal "Eof from IUT while connecting to $pcoinfo(io,$i) pco \`\`$i\'\': $expect_out(buffer)\n"
					}
					-i $sidinfo(iut,$i) timeout {
						# fatal "Timeout from IUT while connecting to $pcoinfo(io,$i) pco \`\`$i\'\'\n"
						send_error "Connection to IUT at $pcoinfo(io,$i) pco \`\`$i\'\' established\n"
					}
				}
				set timeout $maininfo(timeout)
			}
			lappend iutsids $sidinfo(iut,$i)
			set sid2pco($sidinfo(iut,$i)) $i
			set sidinfo(connect,$addressinfo(comm,$pcoinfo(address,$i)),$addressinfo(host,$pcoinfo(address,$i)),$addressinfo(port,$pcoinfo(address,$i))) $i
		} else {
			set bound_pco $sidinfo(connect,$addressinfo(comm,$pcoinfo(address,$i)),$addressinfo(host,$pcoinfo(address,$i)),$addressinfo(port,$pcoinfo(address,$i)))
			send_error "Connection to IUT at $pcoinfo(io,$i) pco \`\`$i\'\' shares\nconnection to IUT at $pcoinfo(io,$bound_pco) pco \`\`$bound_pco\'\'\n"
			set sidinfo(iut,$i) $sidinfo(iut,$bound_pco)
		}
	}
}

# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
#
# connect to running IUT - gci version
#

proc gci_connect_to_iut {} {
	# nothing to do
}
# =========================================================================
