# =========================================================================
#
# main loop
#

# =========================================================================

proc main_loop {} {
	global adaptor
	adaptor_main
	vwait adaptor(quit)
}

proc adaptor_interface {seed file} {
	if {[catch {do_adaptor_interface $seed $file} foo]} {
		bgerror $foo
	}
}
proc do_adaptor_interface {seed file} {
	global adaptor
	adaptor_init $file
	RandomInit $seed
	adaptor_main
	vwait adaptor(quit)
}
proc adaptor_main {} {
	## Currently, we prefere to use blocking IO
	## fconfigure -nonblocking stdin
	fileevent stdin readable adaptor_try_handle_line	
}

# We set bgerror, because we use fileevent to
# handle the input, and fileevent runs the input-handler
# as a background command
proc bgerror {msg} {
	global errorInfo
	debugMsg "Fatal InternalError: $msg"
	set lines [split $errorInfo \n]
	debugMsg "Stack Begin"
	foreach l $lines {
		debugMsg "Stack: $l"
	}
	debugMsg "Stack End"
	exit
}

# ====================================

proc adaptor_try_handle_line {} {
	fileevent stdin readable ""
	set stop 0
	set skip 0
	set line ""
	if {[gets stdin line] < 0} {
		if {[fblocked stdin]} {
			set skip 1
		} elseif {[eof stdin]} {
			puts "A_DEBUG eof on stdin"
			set line "C_QUIT"
			set stop 1
		}
	}
	if {! $skip} {
		set speed [time {adaptor_handle_line $line}]
		# puts "A_DEBUG time: $speed"
	}
	if {! $stop} {
		fileevent stdin readable adaptor_try_handle_line
	}
}

# =========================================================================

proc adaptor_handle_line {line} {
	global adaptor
	# debugMsg "Line: ($line)"
	set line [string trim $line]
	if {![string compare $line ""] == 0} {
		set parsed_line [adaptor_parse $line]
		set parsed_ok [lindex $parsed_line 0]
		# debugMsg "Parsed: ($parsed_line)"
		if {$parsed_ok} {
			set parsed_cmd [lindex $parsed_line 1]
			set cmd_keyword [lindex $parsed_cmd 0]
			switch -- [string toupper $cmd_keyword] {
			    C_QUIT {
			    	set cont 0
			    	set adaptor(quit) 1
			    	puts "A_QUIT"
			  } C_CHANNELS {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set kind_sug [field_value $fields iokind]
			  		set channel_sug [field_value $fields channel]
			  		adaptor_do_channels $kind_sug $channel_sug
			  } C_PCOS {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set kind_sug [field_value $fields iokind]
			  		set channel_sug [field_value $fields channel]
			  		set pco_sug [field_value $fields pco]
			  		adaptor_do_pcos $kind_sug $channel_sug $pco_sug
			  } C_IOKIND {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set chan_sug [field_value $fields channel]
			  		set kind_sug [field_value $fields iokind]
			  		adaptor_do_iokind $kind_sug $chan_sug
			  } C_INPUT {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set ev [field_value $fields event]
			  		set chan [field_value $fields channel]
			  		set s [field_value $fields suspension]
			  		adaptor_do_input $ev $chan $s
			  } C_OUTPUT {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set chan [field_value $fields channel]
			  		adaptor_do_output $chan
			  } C_STATS {
			    	puts "A_STATS no stats yet"
			  } default {
					puts "A_ERROR UnknownCommand: $cmd_keyword"
			  }
			}
		} else {
			puts "A_ERROR UnknownCommand: $line"
		}
	}
}

# =========================================================================
