proc build_config_table {} {
	global config
	foreach k $config(field,list) {
		foreach n $config(field,list,$k) {
			debugMsg config "$k $n:"
			foreach f $config(field,list,$k,$n) {
				set v $config(field,list,$k,$n,$f)
				debugMsg config "  $f=$v"
			}
		}
	}
	foreach name [array names config] {
		debugMsg config "config($name)=$config($name)"
	}
}

# ==============================
# 
# The routines below use:
#   -  var_mk <type>=<inst> <name> <value>
#      enter a Makefile name=value tuple for block
#      <inst> of type <type> (used to 'place' the vars)
#   -  arrayvar_mk <type>=<inst> <name> <value>
#      append to Makefile name=value tuple for block
#      <inst> of type <type> (used to 'place' the vars)
#      (the difference with var_mk is that reinvacation
#       for the same variable appends to it (except when the
#       value is already there), instead of overwriting it)
#   -  var_cfg [component_top] <ctxt> <name> <value>
#      enter a torx-config-file name value tuple
#   -  arrayvar_cfg [component_top] <ctxt> <name> <subname> <value>
#      enter a torx-config-file name subname value tuple
#      meaning that we may have several lines for <name>,
#      that each have a different <subname>, like e.g. for PCO,
#      INPUT, OUTPUT, ADDRESS, etc.
#   -  gen <blocktype> <name>
#      if gen_<blocktype> exists, invoke it for <name>
#      otherwise, try to invoke gen_<block>_<field> for all fields
#      in the block.
#   -  array_gen <blocktype> <namelist>
#      for each <name> in <namelist> invoke gen <blocktype> <name>

proc gen_torx_reset {} {
	var_reset
}
proc gen_torx_defaults {} {
	var_cfg experiment= ADAPTORRUNDIR .
	var_cfg experiment= SPECRUNDIR .
	var_cfg experiment= IUTRUNDIR .
}

proc gen_torx_config {} {
	global config
	if {[info exists config(field,list,campaign)]} {
		set c $config(field,list,campaign)
		if {[llength $c] != 1} {
			warningMsg "more than a single campaign in config: $c"
		}
		array_gen campaign $c
	} else {
		warningMsg "no campaign in config"
	}
}

# ==============================
#
# All routines below fit one of the schemes below:
# - pre_gen_<block>      : called before generation for block
# - gen_<block>          : handle generation for complete block
# - gen_<block>_<field>  : handle generation for field of block
# - post_gen_<block>     : called after generation for block
#
# ==============================

proc pre_gen_campaign {value} {
	global gen
	set gen(campaign) "."
	set gen(makefile) "Makefile"
	rule_mk {} default \$(experiments) {}
	set mkrules [rulegen_mk_ctxt campaign= $value]
	debugMsg output "===== mk rules campaign $value ========="
	debugMsg output  [join $mkrules \n]
	debugMsg output "==========================="
	genfile_append_mkrules_ctxt campaign= $mkrules
}
proc post_gen_campaign {value} {
	global gen
	set mkvars [vargen_mk_ctxt {} $value]
	set mkrules [rulegen_mk_ctxt {} $value]
	set mkincludes [includegen_mk_ctxt campaign=$value $value]
	genfile_append_mkvars_ctxt {} $mkvars
	genfile_append_mkrules_ctxt {} $mkrules
	genfile_append_mkincludes_ctxt {} $mkincludes
	debugMsg output "===== mk vars  $value ========="
	debugMsg output [join $mkvars \n]
	debugMsg output "==========================="
	debugMsg output "===== mk vars  $value $gen(makefile) ========="
	genfile_create [lindex $gen(makefile) 0] \
		[join [concat \
		         [genfile_contents {} {mkvars}] \
		         [genfile_contents campaign=$value {mkincludes}] \
		         [genfile_contents {} {mkrules}] \
			  ] \
			\n]
}
proc gen_campaign_experiment {parent value} {
	array_gen experiment $value
}
proc gen_campaign_dir {parent value} {
	global gen
	set gen(campaign) $value
}
proc gen_campaign_makefile {parent value} {
	global gen
	set gen(makefile) [var_subst $value]
}
proc gen_campaign_mkinclude {parent value} {
	include_mk campaign=$parent [var_subst $value]
}
proc gen_campaign_product {parent value} {
	array_gen product $value
}

proc pre_gen_experiment {value} {
	global gen
#	gen_torx_reset
	set gen(experiment) $value
	gen_torx_defaults
	set gen(mk,exp,$value) Makefile
	var_mk experiment=$value experiment-$value \$(experiment-$value-dir)/run
	component_push experiment=$value
}
proc post_gen_experiment {value} {
	global gen
	set cfg [concat \
	            [vargen_cfg_ctxt experiment=  $value]         \
	            [vargen_cfg_ctxt experiment=$value  $value]   \
	        ]
	set mkvars [vargen_mk_ctxt experiment=$value $value]
	set mkrules [rulegen_mk_ctxt experiment=$value $value]
	set mkincludes [includegen_mk_ctxt experiment=$value $value]
	genfile_append_mkvars_ctxt experiment=$value $mkvars
	genfile_append_mkrules_ctxt experiment=$value $mkrules
	genfile_append_config_ctxt experiment=$value $cfg
	genfile_append_mkincludes_ctxt experiment=$value $mkincludes
	debugMsg output "===== mk vars  experiment $value ========="
	debugMsg output [join $mkvars \n]
	debugMsg output "===== mk rules experiment $value ========="
	debugMsg output [join $mkrules \n]
	debugMsg output "=====   cfg    experiment $value ========="
	debugMsg output [join $cfg \n]
	debugMsg output "==========================="
	set defvars {}
	foreach v {makefile runs seed maxdepth log msg primer driver adapter impl pre preflags post postflags config driverflags} {
		lappend defvars "$v=\$(experiment-$value-$v)"
	}
	if {[info exists gen(cfg,exp,$value)] &&
	    [info exists gen(mk,exp,$value)]} {
		genfile_create [lindex $gen(cfg,exp,$value) 0] \
			[join [genfile_contents experiment=$value {config}] \n]
		set c_mkvars {}
		set c_mkrules {}
		foreach c [component_list] {
			set c_mkvars [concat $c_mkvars [genfile_contents $c {mkvars}]]
			set c_mkrules [concat $c_mkrules [genfile_contents $c {mkrules}]]
		}
		genfile_create [lindex $gen(mk,exp,$value) 0] \
			[join [concat \
			     [genfile_contents experiment=$value {mkvars}] \
			     $c_mkvars \
			     $defvars \
			     [genfile_contents experiment= {mkincludes}] \
			     [genfile_contents experiment=$value {mkincludes}] \
			     [genfile_contents experiment= {mkrules}] \
			     [genfile_contents experiment=$value {mkrules}] \
			     $c_mkrules \
			  ] \
			\n]
	} else {
		warningMsg "experiment=$value: errors while generating (no output)"
	}
	varclear_mk_ctxt experiment=$value $value
	varclear_cfg_ctxt {} $value
	varclear_cfg_ctxt experiment= $value
	varclear_cfg_ctxt experiment=$value $value
	ruleclear_mk_ctxt experiment=$value $value
	arrayvar_mk {} experiments \$(experiment-$value)
	component_reset
	component_pop
}


proc gen_experiment_config {parent value} {
	global gen
	var_mk experiment=$parent experiment-$parent-config $value
	set gen(cfg,exp,$parent) [var_subst $value]
}
proc gen_experiment_makefile {parent value} {
	global gen
	var_mk experiment=$parent experiment-$parent-makefile $value
	set gen(mk,exp,$parent) [var_subst $value]
}
proc gen_experiment_mkinclude {parent value} {
	include_mk experiment=$parent [var_subst $value]
}
proc gen_experiment_runs {parent value} {
	var_mk experiment=$parent experiment-$parent-runs $value
}
proc gen_experiment_seed {parent value} {
	var_cfg experiment= SEED $value
	var_mk experiment=$parent experiment-$parent-seed $value
}
proc gen_experiment_maxdepth {parent value} {
	var_mk experiment=$parent experiment-$parent-maxdepth $value
}
proc gen_experiment_log {parent value} {
	var_cfg experiment= LOGFILE $value
	var_mk experiment=$parent experiment-$parent-log $value
#	arrayvar_mk experiment=$parent experiment-$parent-results \$(experiment-$parent-log)
}
proc gen_experiment_msg {parent value} {
	var_mk experiment=$parent experiment-$parent-msg $value
#	arrayvar_mk experiment=$parent experiment-$parent-results \$(experiment-$parent-msg)
}
proc gen_experiment_dir {parent value} {
	var_mk experiment=$parent experiment-$parent-dir $value
}
proc gen_experiment_pre {parent value} {
	var_mk experiment=$parent experiment-$parent-pre $value
}
proc gen_experiment_preparams {parent value} {
	var_mk experiment=$parent experiment-$parent-preflags $value
}
proc gen_experiment_post {parent value} {
	var_mk experiment=$parent experiment-$parent-post $value
}
proc gen_experiment_postparams {parent value} {
	var_mk experiment=$parent experiment-$parent-postflags $value
}
proc gen_experiment_driver {parent value} {
	gen driver $value
	if {[string compare $value torx] != 0} {
		warningMsg "unknown driver configured: $value"
	}
	var_mk experiment=$parent experiment-$parent-driver \$(driver-$value)
	var_mk experiment=$parent experiment-$parent-driverflags \$(driver-$value-flags)
	rule_mk experiment=$parent \
			\$(experiment-$parent) \$(experiment-$parent-deps) \
			[list [concat cd \$(experiment-$parent-dir) \
			          && \$(MAKE) -f \$(experiment-$parent-makefile) run \
			            	runs=\$(experiment-$parent-runs) \
			            	seed=\$(experiment-$parent-seed) \
			            	maxdepth=\$(experiment-$parent-maxdepth) \
			            	pre=\$(experiment-$parent-pre) \
			            	post=\$(experiment-$parent-post)
			    ]]
}
proc gen_experiment_primer {parent value} {
	gen primer $value
	var_mk experiment=$parent experiment-$parent-primer \$(primer-$value)
	arrayvar_mk experiment=$parent experiment-$parent-deps \$(primer-$value)
	var_cfg experiment= PRIMER $value
}
proc gen_experiment_partitioner {parent value} {
	gen partitioner $value
	var_mk experiment=$parent experiment-$parent-partitioner \$(partitioner-$value)
	arrayvar_mk experiment=$parent experiment-$parent-deps \$(partitioner-$value-deps)
	var_cfg experiment= PRIMER $value
}
proc gen_experiment_combinator {parent value} {
	gen combinator $value
	var_mk experiment=$parent experiment-$parent-primer \$(combinator-$value)
	arrayvar_mk experiment=$parent experiment-$parent-deps \$(combinator-$value-deps)
	var_cfg experiment= PRIMER $value
}
proc gen_experiment_impl {parent value} {
	gen impl $value
	var_mk experiment=$parent experiment-$parent-impl \$(impl-$value)
	arrayvar_mk experiment=$parent experiment-$parent-deps \$(impl-$value)
}
proc gen_experiment_mutant {parent value} {
	var_cfg experiment= MUTANT $value
}
proc gen_experiment_adapter {parent value} {
	gen adapter $value
	var_mk experiment=$parent experiment-$parent-adapter \$(adapter-$value)
}

# ==============================

proc gen_driver_exec {parent value} {
	if {[string compare $value torx] != 0} {
		warningMsg "unknown driver configured: $value"
	}
	var_mk driver=$parent driver-$parent $value
}
proc gen_driver_execparams {parent value} {
	arrayvar_mk driver=$parent driver-$parent-flags [join $value " "]
}

proc post_gen_driver {value} {
	component_add driver $value
	set mkvars [vargen_mk_ctxt driver=$value $value]
	set mkrules [rulegen_mk_ctxt driver=$value $value]
	genfile_append_mkvars_ctxt driver=$value $mkvars
	genfile_append_mkrules_ctxt driver=$value $mkrules
	debugMsg output "===== mk vars  driver $value ========="
	debugMsg output [join $mkvars \n]
	debugMsg output "===== mk rules driver $value ========="
	debugMsg output [join $mkrules \n]
	debugMsg output "==========================="
	varclear_mk_ctxt driver=$value $value
	ruleclear_mk_ctxt driver=$value $value
}

# ==============================

proc gen_combinator_exec {parent value} {
	var_cfg [component_top] SPEC[component_sfx $parent] $value
	var_mk combinator=$parent combinator-$parent $value
}
proc gen_combinator_execdir {parent value} {
	var_cfg [component_top] SPECRUNDIR[component_sfx $parent] $value
	var_mk combinator=$parent combinator-$parent-dir $value
}
proc gen_combinator_execparams {parent value} {
	var_cfg [component_top] SPECFLAGS[component_sfx $parent] $value
}
proc gen_combinator_config {parent value} {
	global gen
	var_mk combinator=$parent combinator-$parent-config $value
	set gen(cfg,comb,$parent) [var_subst $value]
}
proc gen_combinator_primer {parent value} {
	component_push combinator=$parent
	array_gen primer $value
	foreach v $value {
		arrayvar_mk combinator=$parent combinator-$parent-deps \$(primer-$v)
	}
	component_pop
}
proc gen_combinator_partitioner {parent value} {
	component_push combinator=$parent
	array_gen partitioner $value
	foreach v $value {
		arrayvar_mk combinator=$parent combinator-$parent-deps \$(partitioner-$v)
	}
	component_pop
}
proc gen_combinator_tp {parent value} {
	component_push combinator=$parent
	array_gen tp $value
	foreach v $value {
		arrayvar_mk combinator=$parent combinator-$parent-deps \$(tp-$v)
	}
	component_pop
}
proc gen_combinator_test {parent value} {
	component_push combinator=$parent
	array_gen test $value
	foreach v $value {
		arrayvar_mk combinator=$parent combinator-$parent-deps \$(test-$v)
	}
	component_pop
}
proc gen_combinator_combinator {parent value} {
	component_push combinator=$parent
	array_gen combinator $value
	foreach v $value {
		arrayvar_mk combinator=$parent combinator-$parent-deps \$(combinator-$v-deps)
	}
	component_pop
}

proc post_gen_combinator {value} {
	global gen
	component_add combinator $value
	var_cfg combinator=$value CHOOSEINPUTS 1
	set mkvars [vargen_mk_ctxt combinator=$value $value]
	set mkrules [rulegen_mk_ctxt combinator=$value $value]
	set cfg [vargen_cfg_ctxt combinator=$value  $value]
	genfile_append_mkvars_ctxt combinator=$value $mkvars
	genfile_append_mkrules_ctxt combinator=$value $mkrules
	genfile_append_config_ctxt combinator=$value $cfg
	debugMsg output "===== mk vars  combinator $value ========="
	debugMsg output [join $mkvars \n]
	debugMsg output "===== mk rules combinator $value ========="
	debugMsg output [join $mkrules \n]
	debugMsg output "==========================="
	if {[info exists gen(cfg,comb,$value)]} {
		genfile_create [lindex $gen(cfg,comb,$value) 0] \
			[join [genfile_contents combinator=$value {config}] \n]
	} else {
		warningMsg "combinator=$value: errors while generating (no output)"
	}
	varclear_cfg_ctxt combinator=$value $value
	varclear_mk_ctxt combinator=$value $value
	ruleclear_mk_ctxt combinator=$value $value
}

# ==============================

proc gen_partitioner_exec {parent value} {
	var_cfg [component_top] SPEC[component_sfx $parent] $value
	var_mk partitioner=$parent partitioner-$parent $value
}
proc gen_partitioner_execdir {parent value} {
	var_cfg [component_top] SPECRUNDIR[component_sfx $parent] $value
	var_mk partitioner=$parent partitioner-$parent-dir $value
}
proc gen_partitioner_execparams {parent value} {
	var_cfg [component_top] SPECFLAGS[component_sfx $parent] $value
}
proc gen_partitioner_config {parent value} {
	global gen
	var_mk partitioner=$parent partitioner-$parent-config $value
	set gen(cfg,comb,$parent) [var_subst $value]
}
proc gen_partitioner_partfile {parent value} {
	var_cfg partitioner=$parent PARTFILE $value
	var_mk partitioner=$parent partitioner-$parent-partfile $value
	arrayvar_mk partitioner=$parent partitioner-$parent-deps \$(partitioner-$parent-partfile)
}
proc gen_partitioner_primer {parent value} {
	component_push partitioner=$parent
	array_gen primer $value
	foreach v $value {
		arrayvar_mk partitioner=$parent partitioner-$parent-deps \$(primer-$v)
	}
	component_pop
}
proc gen_partitioner_partitioner {parent value} {
	component_push partitioner=$parent
	array_gen partitioner $value
	foreach v $value {
		arrayvar_mk partitioner=$parent partitioner-$parent-deps \$(partitioner-$v-deps)
	}
	component_pop
}

proc post_gen_partitioner {value} {
	global gen
	component_add partitioner $value
	var_cfg partitioner=$value CHOOSEINPUTS 1
	set mkvars [vargen_mk_ctxt partitioner=$value $value]
	set mkrules [rulegen_mk_ctxt partitioner=$value $value]
	set cfg [vargen_cfg_ctxt partitioner=$value  $value]
	genfile_append_mkvars_ctxt partitioner=$value $mkvars
	genfile_append_mkrules_ctxt partitioner=$value $mkrules
	genfile_append_config_ctxt partitioner=$value $cfg
	debugMsg output "===== mk vars  partitioner $value ========="
	debugMsg output [join $mkvars \n]
	debugMsg output "===== mk rules partitioner $value ========="
	debugMsg output [join $mkrules \n]
	debugMsg output "==========================="
	if {[info exists gen(cfg,comb,$value)]} {
		genfile_create [lindex $gen(cfg,comb,$value) 0] \
			[join [genfile_contents partitioner=$value {config}] \n]
	} else {
		warningMsg "partitioner=$value: errors while generating (no output)"
	}
	varclear_cfg_ctxt partitioner=$value $value
	varclear_mk_ctxt partitioner=$value $value
	ruleclear_mk_ctxt partitioner=$value $value
}
# ==============================

proc gen_primer_spec {parent value} {
	gen spec $value
	var_mk primer=$parent genprimer-$parent-deps \$(spec-$value-deps)
	var_mk primer=$parent genprimer-$parent-source \$(spec-$value-source)
}
proc gen_primer_gen {parent value} {
	var_mk primer=$parent genprimer-$parent $value
	rule_mk primer=$parent \
	        \$(primer-$parent) \$(genprimer-$parent-deps) \
	[list [concat cd \$(primer-$parent-dir) && \$(genprimer-$parent) \$(genprimer-$parent-flags) \$(genprimer-$parent-source) ]]
}
proc gen_primer_genparams {parent value} {
	var_mk primer=$parent genprimer-$parent-flags $value
}
proc gen_primer_exec {parent value} {
	var_cfg [component_top] SPEC[component_sfx $parent] $value
	var_mk primer=$parent primer-$parent $value
}
proc gen_primer_execdir {parent value} {
	var_cfg [component_top] SPECRUNDIR[component_sfx $parent] $value
	var_mk primer=$parent primer-$parent-dir $value
}
proc gen_primer_execparams {parent value} {
	var_cfg [component_top] SPECFLAGS[component_sfx $parent] $value
}
proc gen_primer_channel {parent value} {
	array_gen channel $value
}

proc post_gen_primer {value} {
	component_add primer $value
	set mkvars [vargen_mk_ctxt primer=$value $value]
	set mkrules [rulegen_mk_ctxt primer=$value $value]
	genfile_append_mkvars_ctxt primer=$value $mkvars
	genfile_append_mkrules_ctxt primer=$value $mkrules
	debugMsg output "===== mk vars  primer $value ========="
	debugMsg output [join $mkvars \n]
	debugMsg output "===== mk rules primer $value ========="
	debugMsg output [join $mkrules \n]
	debugMsg output "==========================="
	varclear_mk_ctxt primer=$value $value
	ruleclear_mk_ctxt primer=$value $value
}

# ==============================

proc gen_spec {value {super ""}} {
	global specinfo
	if {[string compare $super ""] == 0} {
		set name $value
	} else {
		set name $super
	}
	if {[groupexists spec $value]} {
		set name $value
		foreach f {file auxfile} {
			set v [getfield spec $value $f]
			if {[string compare $v ""] != 0} {
				set specinfo($name,$f) $v
				gen $f $specinfo($name,$f)
			} elseif {![info exists specinfo($name,$f)]} {
				set specinfo($name,$f) ""
			}
		}
	} else {
		warningMsg "spec=$value: undefined"
	}
}

proc post_gen_spec {value} {
	global specinfo
	component_add spec $value
	var_mk spec=$value spec-$value-deps \
		[concat $specinfo($value,file) $specinfo($value,auxfile)]
	var_mk spec=$value spec-$value-source $specinfo($value,file)
	set mkvars [vargen_mk_ctxt spec=$value $value]
	set mkrules [rulegen_mk_ctxt spec=$value $value]
	genfile_append_mkvars_ctxt spec=$value $mkvars
	genfile_append_mkrules_ctxt spec=$value $mkrules
	debugMsg output "===== mk vars  spec $value ========="
	debugMsg output [join $mkvars \n]
	debugMsg output "===== mk rules spec $value ========="
	debugMsg output [join $mkrules \n]
	debugMsg output "==========================="
	varclear_mk_ctxt spec=$value $value
	ruleclear_mk_ctxt spec=$value $value
}

# ==============================

proc gen_channel_timeout {parent value} {
	var_cfg experiment= IUTTIMEOUT $value
}
proc gen_channel_iokind {parent value} {
}
proc gen_channel_pco {parent value} {
	array_gen pco $value
}

# ==============================

proc gen_impl_spec {parent value} {
}
proc gen_impl_configgen {parent value} {
}
proc gen_impl_configgenparams {parent value} {
}
proc gen_impl_gen {parent value} {
	var_mk impl=$parent genimpl-$parent $value
	rule_mk impl=$parent \
	        \$(impl-$parent) {} \
	[list [concat cd \$(impl-$parent-dir) && \$(genimpl-$parent) \$(genimpl-$parent-flags) ]]
}
proc gen_impl_genparams {parent value} {
	var_mk impl=$parent genimpl-$parent-flags $value
}
proc gen_impl_exec {parent value} {
	var_cfg experiment= IUT $value
	var_mk impl=$parent impl-$parent $value
}
proc gen_impl_execdir {parent value} {
	var_cfg experiment= IUTRUNDIR $value
	var_mk impl=$parent impl-$parent-dir $value
}
proc gen_impl_execparams {parent value} {
	var_cfg experiment= IUTFLAGS $value
}
proc gen_impl_execcontext {parent value} {
	var_cfg experiment= IUTCONTEXT $value
	var_mk impl=$parent impl-$parent-context $value
}
proc gen_impl_execcontextparams {parent value} {
	var_cfg experiment= IUTCONTEXTFLAGS $value
}
proc gen_impl_iap {parent value} {
	array_gen iap $value
}
proc post_gen_impl {value} {
	component_add impl $value
	set mkvars [vargen_mk_ctxt impl=$value $value]
	set mkrules [rulegen_mk_ctxt impl=$value $value]
	genfile_append_mkvars_ctxt impl=$value $mkvars
	genfile_append_mkrules_ctxt impl=$value $mkrules
	debugMsg output "===== mk vars  impl $value ========="
	debugMsg output [join $mkvars \n]
	debugMsg output "===== mk rules impl $value ========="
	debugMsg output [join $mkrules \n]
	debugMsg output "==========================="
	varclear_mk_ctxt impl=$value $value
	ruleclear_mk_ctxt impl=$value $value
}

# ==============================

proc gen_adapter_exec {parent value} {
	var_cfg experiment= ADAPTOR $value
	var_mk adapter=$parent adapter-$parent $value
}
proc gen_adapter_execdir {parent value} {
	var_cfg experiment= ADAPTORRUNDIR $value
	var_mk adapter=$parent adapter-$parent-dir $value
}
proc gen_adapter_execparams {parent value} {
	var_cfg experiment= ADAPTORFLAGS $value
}
proc gen_adapter_codingdir {parent value} {
	var_cfg experiment= CODING $value
}
proc gen_adapter_pco {parent value} {
	array_gen pco $value
}
proc post_gen_adapter {value} {
	component_add adapter $value
	set mkvars [vargen_mk_ctxt adaptermpl=$value $value]
	set mkrules [rulegen_mk_ctxt adapter=$value $value]
	genfile_append_mkvars_ctxt adapter=$value $mkvars
	genfile_append_mkrules_ctxt adapter=$value $mkrules
	debugMsg output "===== mk vars  adapter $value ========="
	debugMsg output [join $mkvars \n]
	debugMsg output "===== mk rules adapter $value ========="
	debugMsg output [join $mkrules \n]
	debugMsg output "==========================="
	varclear_mk_ctxt adapter=$value $value
	ruleclear_mk_ctxt adapter=$value $value
}

# ==============================

proc getgate {value type events} {
	set res ""
	foreach e $events {
		set evl [split $e !]
		set gate [lindex $evl 0]
		if {[string compare $res ""] != 0} {
			if {[string compare $res $gate] != 0} {
				warningMsg "pco=$value->$type: gate $res ignored; using $gate"
			}
		}
		set res $gate
	}
	return $res
}

proc gen_pco {value {super ""}} {
	global pcoinfo
	if {[string compare $super ""] == 0} {
		set name $value
	} else {
		set name $super
	}
	if {[groupexists pco $value]} {
		foreach f {address encoder decoder multiplexer regexp} {
			set v [getfield pco $value $f]
			if {[string compare $v ""] != 0} {
				set pcoinfo($name,$f) $v
				# get the name in addition to the 'id',
				# because we need it for address in post_gen_pco
				set vname [getfield $f $v name]
				if {[string compare $vname ""] != 0} {
					set pcoinfo($name,$f,name) $vname
				} else {
					set pcoinfo($name,$f,name) $v
				}
				gen $f $pcoinfo($name,$f)
			} elseif {![info exists pcoinfo($name,$f)]} {
				set pcoinfo($name,$f) ""
				set pcoinfo($name,$f,name) ""
			}
		}
		foreach f {ievent oevent} {
			set v [getfield pco $value $f]
			if {[string compare $v ""] != 0} {
				set pcoinfo($name,$f) [getgate $value $f $v]
				gen $f $pcoinfo($name,$f)
			} elseif {![info exists pcoinfo($name,$f)]} {
				set pcoinfo($name,$f) ""
			}
		}
	} else {
		warningMsg "pco=$value: undefined"
	}
}
proc post_gen_pco {value} {
	global pcoinfo
	set name $value
	arrayvar_cfg experiment= PCO $name [list [list \
			[list $pcoinfo($name,oevent) $pcoinfo($name,ievent) ] \
			$pcoinfo($name,address,name) \
			[list {} $pcoinfo($name,decoder) [join $pcoinfo($name,regexp) ""] ] \
	] ]
	arrayvar_cfg experiment= INPUT $pcoinfo($name,ievent) [list [list \
			{} \
			$pcoinfo($name,encoder) \
			$pcoinfo($name,multiplexer) \
	] ]
	arrayvar_cfg experiment= OUTPUT $pcoinfo($name,oevent) [list [list \
			{} \
			{} \
	] ]
	catch {unset pcoinfo}
}

# ==============================

proc gen_iap_address {parent value} {
	gen address $value
}

# ==============================

proc gen_address_value {parent value} {
	set name [getfield address $parent name]
	if {[string compare $name ""] == 0} {
		set name $parent
	}
	set vl [split $value !]
	if {[llength $vl] > 1} {
		switch -- [lindex $vl 0] {
			udp {
				set host [lindex $vl 1]
				if {[string compare $host *] == 0} {
					set host currenthost
				}
				set port [lindex $vl 2]
				if {[string compare $port ""] == 0} {
					set port 0
				}
				set suf [list {--port ${P}}]
				arrayvar_cfg experiment= ADDRESS $name [list udp $host $port $suf]
			}
			tcp {
				set host [lindex $vl 1]
				if {[string compare $host *] == 0} {
					set host currenthost
				}
				set port [lindex $vl 2]
				if {[string compare $port ""] == 0} {
					set port 0
				}
				arrayvar_cfg experiment= ADDRESS $name [list udp $host port]
			}
			default {
				arrayvar_cfg experiment= ADDRESS $name [list $value]
			}
		}
	} else {
		arrayvar_cfg experiment= ADDRESS $name [list $value]
	}
}

# ==============================
