# This file implements the machinery underlying the PRIMER, GUIDE and MUTANT
# commands that we use in primer, guide and mutant config files.
#
# This file is derived from ../Gui/extension.tcl
#
# The idea is that
#    - when we have a Gui (xtorx), the Gui is responsible for extension handling
#    - when we do not have a Gui, the Driver is responsible for loading the
#       extension file(s) and enabling the chosen one.
#
# For the Driver it makes no sense to load an extension file
# if no item of it is used (loading of an extension file implies
# running (end)user code so we have no idea what is done --
# it may be expensive, so only do it when useful.



# This routine will be invoked from xtorx to force auto-loading of this file.
proc do_first_initExtension {{arglist {}}} {
}

proc do_initExtension {kind pname
                       fvarname flagvarname cfgvarname
                       {arglist {}}} {
	global tester
	global debug

	set tester(ext,$kind,procname) $pname
	set tester(ext,$kind,file,varname) $fvarname
	set tester(ext,$kind,torx,cmdline,flags) $flagvarname
	set tester(ext,$kind,torx,cmdline,config) $cfgvarname

	set tester(ext,$kind,nrof) 0
	set $tester(ext,$kind,file,varname) ""
}

proc do_cleanupExtension {kind {arglist {}}} {
	global tester
	if {[info exists tester(ext,$kind,variable,prev)]} {
		# test to make sure that we indeed have a constructor name/string
		if {$tester(ext,$kind,constructor,$tester(ext,$kind,variable,prev)) != {}} {
			Debug_Msg "do_cleanupExtension" "$kind Destructor called for: $tester(ext,$kind,variable,prev)"
			if {[catch {eval "$tester(ext,$kind,destructor,$tester(ext,$kind,variable,prev)) $tester(ext,$kind,torx,cmdline,flags) $tester(ext,$kind,torx,cmdline,config)" } foo]} {
				catch {puts stderr "Error executing destructor for entry nr. $tester(ext,$kind,variable,prev) \"$tester(ext,$kind,entryname,$tester(ext,$kind,variable,prev))\": $foo"}
			}
		} else {
			Debug_Msg "do_cleanupExtension" "$kind No destructor for: $tester(ext,$kind,variable,prev)"
		}
		catch { unset tester(ext,$kind,variable,prev) }
	}
}

proc do_do_open_extension {kind filename} {
	global tester
	if {$filename != ""} {
		do_do_close_extension $kind
		set tester(ext,$kind,file) $filename
		set basename [file tail $filename]
		set $tester(ext,$kind,file,varname) [file rootname $basename]
		set tester(dirname) [file dirname $filename]
		set tester(open,initialdir) $tester(dirname)
		
		if {[update_extension_menu $kind $filename]} {
			return 1
		}
	}
	return 0
}
proc update_extension_menu_add {kind file} {
	global tester
	do_cleanupExtension $kind
	# add entry 'none' and make it active
	# the user may add his own entry named none, if he wants to;
	extension_register $tester(ext,$kind,nrof) $kind "none" "" {} {}
	set tester(ext,$kind,variable,cur) $tester(ext,$kind,nrof)
	set tester(ext,$kind,none) $tester(ext,$kind,nrof)
	
	if {[catch { source $file } foo ]} {
		catch {puts stderr "Error loading $kind file $file: $foo"}
		update_extension_menu_clear $kind
		return 0
	} else {
		return 1
	}
}
proc update_extension_menu_clear {kind} {
	global tester
	do_cleanupExtension $kind
	while {$tester(ext,$kind,nrof) > 0} {
		catch {unset $tester(ext,$kind,entryname,$tester(ext,$kind,nrof))}
		catch {unset $tester(ext,$kind,constructor,$tester(ext,$kind,nrof))}
		catch {unset $tester(ext,$kind,destructor,$tester(ext,$kind,nrof))}
		incr tester(ext,$kind,nrof) -1
	}
	set tester(ext,$kind,nrof) 0
	return 1
}
proc update_extension_menu {kind {file ""}} {
	global tester
	if {$file != ""} {
		return [update_extension_menu_add $kind $file]
	} else {
		return [update_extension_menu_clear $kind]
	}
}
proc do_do_close_extension {kind} {
	global tester
	set tester(ext,$kind,file) {}
	set $tester(ext,$kind,file,varname) {}
		
	# make sure that we unset the mutant when we close the file
	if {[info exists tester(ext,$kind,none)]} {
		set tester(ext,$kind,variable,cur) $tester(ext,$kind,none)
		do_extension_callback $kind
		update_extension_menu $kind
	}
}

proc do_extension_callback {kind {arglist {}}} {
	global tester
	Debug_FnArgs "do_extension_callback" $kind $arglist
	if {[info exists tester(ext,$kind,variable,prev)]} {
		if {$tester(ext,$kind,variable,cur) == $tester(ext,$kind,variable,prev)} {
			return
		}
	}
	do_cleanupExtension $kind
	set tester(ext,$kind,variable,prev) $tester(ext,$kind,variable,cur)
	# test to make sure that we indeed have a constructor name/string
	if {$tester(ext,$kind,constructor,$tester(ext,$kind,variable,prev)) != {}} {
		Debug_Msg "do_extension_callback" "$kind Constructor called for: $tester(ext,$kind,variable,prev)"
		if {[catch {eval "$tester(ext,$kind,constructor,$tester(ext,$kind,variable,prev)) $tester(ext,$kind,torx,cmdline,flags) $tester(ext,$kind,torx,cmdline,config)" } foo]} {
			catch {puts stderr "Error executing constructor for entry nr. $tester(ext,$kind,variable,prev) \"$tester(ext,$kind,entryname,$tester(ext,$kind,variable,prev))\": $foo"}
		}
	} else {
		Debug_Msg "do_extension_callback" "$kind No constructor for: $tester(ext,$kind,variable,prev)"
	}
}

proc do_do_select_extension {kind name} {
	global tester

	if {[info exists tester(ext,$kind,entry2key,$name)]} {
		set tester(ext,$kind,variable,cur) $tester(ext,$kind,entry2key,$name)
		do_extension_callback $kind
	}
}
proc extension_register {key kind entry title create destruct {arglist {}}} {
	global tester
	set tester(ext,$kind,entryname,$key) $entry
	set tester(ext,$kind,constructor,$key) $create
	set tester(ext,$kind,destructor,$key) $destruct
	set tester(ext,$kind,args,$key) $arglist
	if {![info exists tester(ext,$kind,entry2key,$entry)]} {
		set tester(ext,$kind,entry2key,$entry) $key
	}
}

proc EXTENSION {kind entry title create destruct {arglist {}}} {
	global tester
	Debug_FnArgs "EXTENSION" $kind $entry $title $create $destruct $arglist
	incr tester(ext,$kind,nrof)
	extension_register $tester(ext,$kind,nrof) $kind $entry $title $create $destruct $arglist
	
	# the name "none" is special: it will be the default setting
	if {[string compare $entry none] == 0} {
		set tester(ext,$kind,variable,cur) $tester(ext,$kind,nrof)
	}
}

# ===================================

proc do_initGuides {args} {
	global tester
	global debug

	# set debug(Msg,do_extension_callback) 1
	# set debug(Msg,do_cleanupExtension) 1

	do_initExtension \
		guides \
		GUIDE \
		tester(basenamebare,guides_file) \
		tester(torx,cmdline,guides,flags) \
		tester(torx,cmdline,guides,config) \
		$args
}

proc GUIDE {entry title create destruct args} {
	EXTENSION guides $entry $title $create $destruct $args
}

# ===================================

proc do_initMutants {args} {
	global tester
	global debug

	# set debug(Msg,do_extension_callback) 1
	# set debug(Msg,do_cleanupExtension) 1

	do_initExtension \
		mutants \
		MUTANT \
		tester(basenamebare,mutants_file) \
		tester(torx,cmdline,mutants,flags) \
		tester(torx,cmdline,mutants,config) \
		$args
}

proc MUTANT {entry title create destruct args} {
	EXTENSION mutants $entry $title $create $destruct $args
}

# ===================================

proc do_initPrimers {args} {
	global tester
	global debug

	# set debug(Msg,do_extension_callback) 1
	# set debug(Msg,do_cleanupExtension) 1

	do_initExtension \
		primers \
		PRIMER \
		tester(basenamebare,primers_file) \
		tester(torx,cmdline,primers,flags) \
		tester(torx,cmdline,primers,config) \
		$args
}

proc PRIMER {entry title create destruct args} {
	EXTENSION primers $entry $title $create $destruct $args
}

# ===================================

proc extension_init {} {
	do_initPrimers
	do_initGuides
	do_initMutants
}

proc extension_process_config {} {
	global tester
	global maininfo
	global debug

	foreach kind {primers guides mutants} {
		set cfgkey $tester(ext,$kind,procname)
		set cfgkey_s [set cfgkey]S
		set flagsvarname $tester(ext,$kind,torx,cmdline,flags)
		set configvarname $tester(ext,$kind,torx,cmdline,config)

		if {[info exists maininfo($cfgkey_s)] &&
		    [info exists maininfo($cfgkey)] &&
		    ([string compare $maininfo($cfgkey_s) ""] != 0) &&
		    ([string compare $maininfo($cfgkey) ""] != 0)} {
			catch {puts stderr "trying $cfgkey_s $maininfo($cfgkey_s)"}
			catch {puts stderr "trying $cfgkey $maininfo($cfgkey)"}
			if {[do_do_open_extension $kind [file_mk_absolute $maininfo($cfgkey_s) $tester(dirname)]]} {
				catch {puts stderr "using $cfgkey_s $maininfo($cfgkey_s)"}
				do_do_select_extension $kind $maininfo($cfgkey)
				if {[info exists $flagsvarname]} {
					parse_args [set $flagsvarname]
				}
				if {[info exists $configvarname]} {
					foreach file [set $configvarname] {
						read_config_file $file
					}
				}
			} else {
				catch {puts stderr "failing to use $cfgkey_s $maininfo($cfgkey_s)"}
			}
		}
	}
}

# ===================================
