# =========================================================================
# $Id: smileexp.tcl,v 1.1.2.6 2001/11/19 14:03:21 belinfan Exp $
# =========================================================================

# ====================================
# Explorer Tcl API Interface routines

proc explorer_init {} {
	global unfold_depth
	global rewrite_always
	global rewrite_minimal
	global unparse_incremental
	set unfold_depth 1
	set rewrite_always 1
	set rewrite_minimal 1
	# set unparse_incremental 0
	global selexp_stack
	set selexp_stack {}
}

proc explorer_start {} {
	set root [simulate [specification_name]]
	# currently, simulate does not return a value,
	# so we return the root node explicitly.
	set root 0
	return [unparse_state $root]
}

proc explorer_unfold {s} {
	# Do not use 'expand', because it may refuse to
	# expand states visited already!!!
	# Use unfold instead (make sure: unfold_depth == 1)
	# set evs [expand $s]
	set evs [unfold $s]
	if {[llength $evs] <= 0} {
		warningMsg "empty menu"
		return -code error -errorinfo "empty menu"
	}
	set res {}
	foreach e $evs {
		lappend res [unparse_event $e]
	}
	return $res
}

# return a tuple of either:
#  1 new-event-as-list
#  0 solve-predicates-result-as-list
proc explorer_instantiate {ev preds} {
	transmit_predicates $ev
	add_to_goal $preds
	set res [solve_goal]
	set comp_ok [lindex $res 0]
	set sol_found [lindex $res 1]
	set txt [lindex $res 2]
	set msg [join [lindex $res 3] " ; "]
	set err [join [lindex $res 4] " ; "]
	set sol [join [lindex $res 5] " ; "]
	set answer [list $comp_ok $sol_found $sol $msg $err]
	if {$comp_ok > 0} {
		set new_ev [instantiate_event]
		return [list 1 [unparse_event $new_ev]]
	} else {
		return [list 0 $answer]
	}
}

proc explorer_solve {preds} {
	clear_goal
	add_to_goal $preds
	set res [solve_goal]
	# debugMsg "solve_predicates ($preds) ($res)"
	set comp_ok [lindex $res 0]
	set sol_found [lindex $res 1]
	set txt [lindex $res 2]
	set msg [join [lindex $res 3] " ; "]
	set err [join [lindex $res 4] " ; "]
	set sol [join [lindex $res 5] " ; "]
	return [list $comp_ok $sol_found $sol $msg $err]
}

proc explorer_match_events {ev evs} {
	foreach e $evs {
		set res [match $ev $e]
		if {$res} {
			return [list 1 $e]
		}
	}
	return [list 0]
}

# =========================================================================
# Explorer Tcl API Utility routines

proc unparse_event {e} {
	set preds_solved [predicates_solved $e] 
	set repr [show_event $e]
	set gate [lindex $repr 0]
	set exps [string trim [lindex $repr 1]]
	set preds [string trim [lindex $repr 2]]
	set freevars [free_variables $e]
	set normalised [normalise_exps_preds_vars $exps $preds $freevars]
	set nexps [lindex $normalised 0]
	set npreds [lindex $normalised 1]
	set nfreevars [lindex $normalised 2]
	if {[catch { set elsewhere [identical_node $e] } ]} {
		set elsewhere ""
	}
	if {[string compare $gate "i"] == 0} {
		set visible 0
	} else {
		set visible 1
	}
	return [list $e $visible $preds_solved [concat $gate $nexps] $npreds $nfreevars $elsewhere]
}

proc unparse_state {e} {
	set preds_solved [predicates_solved $e] 
	set preds [string trim [predicates $e]]
	set freevars [free_variables $e]
	set normalised [normalise_exps_preds_vars {} $preds $freevars]
	set npreds [lindex $normalised 1]
	set nfreevars [lindex $normalised 2]
	if {[catch { set elsewhere [identical_node $e] } ]} {
		set elsewhere ""
	}
	return [list $e $preds_solved $npreds $nfreevars $elsewhere]
}

proc normalise_experiments {exps} {
	set s ""
	foreach e $exps {
		set exquest [lindex $e 0]
		set exp [lindex $e 1]
		set type [lindex $e 2]
		append s " ! $exp"
	}
	return [string trim $s]
}

proc normalise_predicates {preds} {
	set s [join $preds " ; "]
	return [string trim $s]
}

proc normalise_freevars {varpairs} {
	set ls {}
	foreach vp $varpairs {
		lappend ls [join $vp " "]
	}
	return [string trim [join $ls " "]]
}

proc normalise_exps_preds_vars {exps preds varpairs} {
	set exps_s [normalise_experiments $exps]
	set preds_s [normalise_predicates $preds]
	set freevars {}
	foreach vp $varpairs {
		set v [lindex $vp 0]
		set t [lindex $vp 1]
		if {![info exists vartype($t)]} {
			set vartype($t) 0
		}
		incr vartype($t)
		# set newvar "var_[set t]_[set vartype($t)]"
		set newvar "var_[set t]\$[set vartype($t)]"
		regsub -all -- "[set v](\[^a-zA-Z0-9_\$\]|\$)" $preds_s "$newvar\\1" preds_s
		regsub -all -- "[set v](\[^a-zA-Z0-9_\$\]|\$)" $exps_s "$newvar\\1" exps_s
		lappend freevars [list $newvar $v $t]
	}
	set freevars_s [normalise_freevars $freevars]
	return [list $exps_s $preds_s $freevars_s]
}

# =========================================================================
# Explorer Tcl API Smile specific Interface/Utility routines
# What are we going to do with these????

proc disable_selection_exportation {} {
	global selexp_stack
	lappend selexp_stack [export_selection 0]
}
proc restore_selection_exportation {} {
	global selexp_stack
	set old [lindex $selexp_stack end]
	set selexp_stack [lreplace $selexp_stack end end]
	update idletasks ;# make selections before enabling export
	export_selection $old
}

# ====================================
