#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "spexSet.h"


DSetP	getNewDSet();


// (DSetP) NULL and emptySet are used interchangeable.
const DSetP emptySet = NULL;


// Defining SetAppendElem macro.
// Usage: SetAppendElem(DSetP fs, DSetP ns, DRangeP el)
// 'el' will be evaluated only once.
void SetAppendElem(DSetP, DSetP, DRangeP);

#define SetAppendElem(fs, ns, el) \
if((fs) == emptySet) {\
	(fs) = (ns) = getNewDSet();\
} else {\
	(ns)->next = getNewDSet();\
	(ns) = (ns)->next;\
}\
(ns)->elem = RangeClone(el);
// End of SetAppendElem macro.



// Create empty Set
DSetP	SetMakeEmpty() {
	return emptySet;
}

// Checks if Set is empty
char	SetIsEmpty(DSetP set) {
	if (set == emptySet) {
		return 1;
	} else {
		return 0;
	}
}

// Checks if a Range is contained in the set. Returns 1 if is partially contained,
// 2 if it is fully contained.
char	SetContains(DSetP set, DRangeP elem) {
	DSetP curSet;
	
	for(curSet = set; curSet != emptySet; curSet = curSet->next) {
		if (RangeContainedIn(curSet->elem, elem)) {
			return 2;
		} else if (RangeHasOverlap(curSet->elem, elem)) {
			return 1;
		}
	}
	return 0;
}

DSetP	SetIntersect(DSetP setA, DSetP setB) {
	DSetP curSetA;
	DSetP curSetB;
	
	DSetP firstNewSet = emptySet;
	DSetP newSet;
	
	curSetB = setB;
	
	for(curSetA = setA; curSetA != emptySet; curSetA = curSetA->next) {
		while ( curSetB != emptySet && 
		        (RangeCmp(curSetB->elem, curSetA->elem) < 0) &&
		        (! RangeHasOverlap(curSetB->elem, curSetA->elem)) ) {
			curSetB = curSetB->next;
		}
		if (curSetB == emptySet) {
			break;
		}
		if ( RangeHasOverlap(curSetA->elem, curSetB->elem) ) {
			SetAppendElem(firstNewSet, newSet, RangeOverlap(curSetA->elem, curSetB->elem));
		}
	}
			 
	return firstNewSet;
}


DSetP	SetUnion(DSetP setA, DSetP setB) {
	DSetP curSetA, curSetB;
	DSetP longestSet = emptySet;
	
	DSetP firstNewSet = emptySet;
	DSetP newSet = NULL;
	
	DSetP *combineSet = NULL;
	DRangeP combineRange;

	curSetA = setA;
	curSetB = setB;
	
#define	rangeA (curSetA->elem)
#define	rangeB (curSetB->elem)
#define	rangeNew (newSet->elem)

	while ( curSetA != emptySet && curSetB != emptySet) {
		
		// Determine if we need one of the sets to be combined with the previous one (if there is one)
		if(newSet && RangeCanCombine(rangeA, rangeNew)) {
			combineSet = &curSetA;
		} else if (newSet && RangeCanCombine(rangeB, rangeNew)) {
			combineSet = &curSetB;
		} else {
			combineSet = NULL;
		}
		
		
		if (combineSet) {
			// We need to combine a set with a previously saved one.
			combineRange = RangeCombine((*combineSet)->elem, rangeNew);
			free(rangeNew);
			rangeNew = combineRange;
			(*combineSet) = (*combineSet)->next;
		} else {
			// We don't have to combine the current set with a previous one
			if (RangeCanCombine(rangeA, rangeB) ) {
				SetAppendElem(firstNewSet, newSet, RangeCombine(rangeA, rangeB));
				curSetA = curSetA->next;
				curSetB = curSetB->next;
			} else {
				if ( RangeCmp(rangeA, rangeB) < 0 ) {
					SetAppendElem(firstNewSet, newSet, rangeA);
					curSetA = curSetA->next;
				} else {
					SetAppendElem(firstNewSet, newSet, rangeB);
					curSetB = curSetB->next;
				}
			} 
		}
	}
	
	if ( curSetA != emptySet ) {
		longestSet = curSetA;
	} else {
		longestSet = curSetB;
	}
	
	while (longestSet != emptySet) {
		if (newSet && RangeCanCombine(rangeNew, longestSet->elem)) {
			combineRange = RangeCombine(rangeNew, longestSet->elem);
			free(rangeNew);
			rangeNew = combineRange;				
		} else {
			SetAppendElem(firstNewSet, newSet, longestSet->elem);
		}
	
		longestSet = longestSet->next;	
	}

#undef	rangeA
#undef	rangeB
#undef	rangeNew
	
	return firstNewSet;
}

DSetP	SetRemove(DSetP setA, DSetP setB) {
	DSetP curSetA, curSetB;
	
	DSetP firstNewSet = emptySet;
	DSetP newSet = NULL;
	
	DRangeP curRange;
	DRangeP *ranges;
	
	for (curSetA = setA, curSetB = setB; curSetA != emptySet && curSetB != emptySet; curSetA = curSetA->next) {
		curRange = curSetA->elem;

		while ( curSetB != emptySet && RangeSmallerThan(curSetB->elem, curRange) && ! RangeHasOverlap(curRange, curSetB->elem) ) {
			curSetB = curSetB->next;
		}
		if (curSetB == emptySet)
			break;
		
		while(curRange != NULL && RangeHasOverlap(curRange, curSetB->elem)) {
			// Remove from the A element, the values in the B element.
			ranges = RangeDelete(curRange, curSetB->elem);
			
			if ( ranges[0] != NULL) {
				SetAppendElem(firstNewSet, newSet, ranges[0]);
				free(ranges[0]);
			}
			
			curRange = ranges[1];

			free(ranges);

			if (curRange != NULL) {
				curSetB = curSetB->next;
				if (curSetB == emptySet)
					break;
			}
		}
		if ( curRange != NULL ) {
			SetAppendElem(firstNewSet, newSet, curRange);
			free(curRange);
		}

		
	}
	while ( curSetA != emptySet && curSetB == emptySet ) {
		SetAppendElem(firstNewSet, newSet, curSetA->elem);
		curSetA = curSetA->next;
	}
	
	return firstNewSet;

}

int	SetIsSane(DSetP set) {
	DSetP curSet;
	DRangeP prevRange;
	
	if (set == emptySet) {
		return 1;
	} else {
		prevRange = set->elem;
		curSet = set->next;
		
		while( curSet != emptySet && 
		       (! RangeCanCombine(prevRange, curSet->elem)) &&
		       (RangeCmp(prevRange, curSet->elem) < 0)  ) 
		{
		       	prevRange = curSet->elem;
		       	curSet = curSet->next;
		}
		return (curSet == emptySet);
	}	
}
	

DSetP	SetSanitize(DSetP set) {
	DSetP firstNewSet = NULL;
	DSetP newSet = NULL;
	
	DSetP curSet;
	DRangeP combineRange;
	
	// TODO Sanitize even if order was mixed up. (Might need SetMakeArrayUnsort)
	
	for(curSet = set; curSet != emptySet; curSet = curSet->next) {
		if (newSet && RangeCanCombine(newSet->elem, curSet->elem)) {
			combineRange = RangeCombine(newSet->elem, curSet->elem);
			free(newSet->elem);
			newSet->elem = combineRange;
		} else {
			SetAppendElem(firstNewSet, newSet, curSet->elem);
		}
	}
	
	return firstNewSet;
}


// Searches the set for a range element, for which the function 'cmp' returns true.
// If it finds such an element, it returns the first element that does. Otherwise the function returns NULL.
DRangeP	SetFind(DSetP set, char cmp(DRangeP)) {
	
	DSetP curSet;
	
	for (curSet = set; curSet != emptySet; curSet = curSet->next) {
		if (cmp(curSet->elem)) {
			return curSet->elem;
		}
	}
	
	return NULL;
}

// Maps the function 'fcn' on every range element. If 'fcn' returns NULL for an
// element, this element will not be added to the resulting set.
DSetP	SetMap(DSetP set, DRangeP fcn(DRangeP)) {
	DSetP curSet;

	DSetP firstNewSet = NULL;
	DSetP newSet;
	
	DRangeP newRange = NULL;	
	
	for (curSet = set; curSet != emptySet; curSet = curSet->next) {
		newRange = fcn(curSet->elem);
		if (newRange != NULL) {
			SetAppendElem(firstNewSet, newSet, newRange);
		}
	}
	return firstNewSet;
}
	

char	SetEq(DSetP setA, DSetP setB) {
	DSetP curSetA, curSetB;
	
	if (SetIsSane(setA)) {
		curSetA = setA;
	} else {
		curSetA = SetSanitize(setA);
	}
	if (SetIsSane(setB)) {
		curSetB = setB;
	} else {
		curSetB = SetSanitize(setB);
	}
	
	while ( curSetA != emptySet && curSetB != emptySet && RangeEq(curSetA->elem, curSetB->elem) ) {	
		curSetA = curSetA->next;
		curSetB = curSetB->next;
	}
	
	return (curSetA == emptySet && curSetB == emptySet);
}

DSetP	getNewDSet() {
	DSetP set;
	
	set = calloc(1, sizeof(DSet));
	
	if (set == NULL ) {
		printf("Error. Could not allocate space for DSet element. Aborting.\n");
		exit(EXIT_FAILURE);
	}
	
	return set;
}

DSetP	SetMakeFromString(const char *input) {
	char *	rangeStr;
	char *	rangeTok;
	char *	tokSplit;
	
	
	RANGE_VALUE_TYPE *values = NULL;
	int     valCount = 0;
	int	maxValCount = 0;
	
	DRangeP *rElts;

	rangeStr = strdup(input);
	
	maxValCount = 30;
	values = malloc(2 * maxValCount * sizeof(RANGE_VALUE_TYPE));
	
	rangeTok = strtok(rangeStr, ",");
	do {
		if (rangeTok == NULL) {
			break;
		}

		while (isspace(*rangeTok)) {
			rangeTok += 1;
		}
		if ( *rangeTok == '\0') {
			break;
		}

		if (valCount == maxValCount) {
			maxValCount += 20;
			values = realloc(values, 2 * maxValCount * sizeof(RANGE_VALUE_TYPE));
		}
		
		tokSplit = strstr(rangeTok, "..");
		if (tokSplit == NULL) {
			// We have only one value at 'rangeTok'
			values[valCount * 2] = values[valCount * 2 + 1] = atoi(rangeTok);
		} else {
			// TODO is the removal of the periods necessary? 
			tokSplit[0] = tokSplit[1] = '\0';
			// We have two values at 'rangeTok' and 'tokSplit + 2'
			values[valCount * 2] = atoi(rangeTok);
			values[valCount * 2 + 1] = atoi(tokSplit + 2);
		}
		valCount += 1;
	} while ( (rangeTok = strtok(NULL, ",")) != NULL );		
	
	free(rangeStr);
	
	rElts = RangeMakeArray(values, valCount);
	
	free(values);
	
	return SetMakeArray(rElts, valCount);
}


DSetP	SetMakeSingle(DRangeP elem) {
	DSetP set;
	
	set = getNewDSet();
	set->elem = elem;
	set->next = emptySet;
	
	return set;
}


DSetP	SetMakeArray(DRangeP elements[], int nrElements) {
	return SetInsertArray(SetMakeEmpty(), elements, nrElements);
}


// Creates a new DSet, that contains all elements from 'set', as well as 'elem'.
// The elements are not copied, only the reference is.
DSetP	SetInsertElem(DSetP set, DRangeP elem) {
	return SetInsertArray(set, &elem, 1);
}



// Creates a new DSet, that contains all elements from 'set', as well as the ones in 'elements'.
// The elements from the original DSet are cloned.
DSetP	SetInsertArray(DSetP set, DRangeP elements[], int nrElements) {
	DSetP curSet;
	DSetP firstNewSet = NULL;
	DSetP newSet = NULL;
	//DRangeP curElem;
	int elemCursor;
	char insertedAll = 0;
	
	curSet = set;
	elemCursor = 0;
	while (curSet != emptySet) {
		if (firstNewSet == NULL) {
			firstNewSet = newSet = getNewDSet();
		} else {
			newSet->next = getNewDSet();
			newSet = newSet->next;
		}
		
		
		// If we haven't inserted the element yet...
		if ( ! insertedAll ) {
			if ( RangeContainedIn(curSet->elem, elements[elemCursor]) ) {
			// If the element is already in the set, discard it.
				insertedAll = (elemCursor == nrElements - 1);
				elemCursor++;
			} else if ( RangeCmp(curSet->elem, elements[elemCursor]) > 0 ) {
			// If we have arrived at the first element that is bigger, insert
			// the element.
				insertedAll = (elemCursor == nrElements - 1);
				newSet->elem = elements[elemCursor];
				elemCursor++;
				continue;	// Process next element
			}
		
		}
		
		newSet->elem = RangeClone(curSet->elem);
		
		curSet = curSet->next;
	}
	
	// Append all remaining elements
	if (insertedAll == 0) {
		for (; elemCursor < nrElements; elemCursor++) {
			if (firstNewSet == NULL) {
				firstNewSet = newSet = getNewDSet();
			} else {
				newSet->next = getNewDSet();
				newSet = newSet->next;
			}
			newSet->elem = elements[elemCursor];
			//SetAppendElem(firstNewSet, newSet, elements[elemCursor]);
		}
	}
	
	if ( ! SetIsSane(firstNewSet) ) {		
		firstNewSet = SetSanitize(firstNewSet);
	}
	return firstNewSet;
}


void	SetFree(DSetP set) {
	DSetP curSet;
	DSetP tmpSet;
	
	for ( curSet = set; curSet!= emptySet; curSet = tmpSet ) {
		tmpSet = curSet->next;
		
		free(curSet->elem);
		free(curSet);
	}
	return;
}


void	SetPrint(DSetP set) { 
	DSetP curSet;
	
	fprintf(stderr, "[");
	for(curSet = set; curSet != emptySet; curSet = curSet->next) {
		RangePrint(curSet->elem);
		fprintf(stderr, ", ");
	}
	fprintf(stderr, "]\n");
}

char *	SetToString(DSetP set) {
	static char *line;	
	static int line_length; // Total allocated space
	
	int incr_amount = 15;
	
	char *tmp_line;

	int cur_place = 0; // Index for line, to place the next character
	int range_size = 0;
	
	DSetP curSet;
	
	if ( line == NULL ) {
		line = (char *) calloc(5, sizeof(char));
		if (line == NULL) {
			fprintf(stderr, "Error. Could not allocate initial storage for Set printing.\n");
			fprintf(stderr, "Program will abort now.\n");
			exit(-1);
		}
		line_length = 5;
	}

	// We're always able to print the opening brace.	
	cur_place += sprintf(&line[cur_place], "{");
	
	// For each element from the set...
	for(curSet = set; curSet != emptySet; curSet = curSet->next) {
		
		// Put the element in the buffer. 
		while ( (range_size = snRangePrint(&line[cur_place], line_length - cur_place, curSet->elem)) < 0) {
			tmp_line = realloc(line, line_length + incr_amount);
			if (tmp_line == 0) {
				fprintf(stderr, "Warning. Could not allocate additional storage for set printing.\n");
				return line;
			}
			
			line = tmp_line;
			line_length += incr_amount;
		}

		cur_place += range_size;
		
		if (line_length - cur_place < 3 ) {
			tmp_line = realloc(line, line_length + incr_amount);
			if (tmp_line == 0) {
				fprintf(stderr, "Warning. Could not allocate additional storage for set printing.\n");
				return line;
			}
			line = tmp_line;
			line_length += incr_amount;
		}


		if (curSet->next != NULL)
			cur_place += sprintf(&line[cur_place], ", ");
	}



	sprintf(&line[cur_place], "}");
	
	return line;
}	
		
			

int	SetSize(DSetP set) {
	DSetP curSet;
	int prev = 0;
	
	for (curSet = set; curSet != emptySet; curSet = curSet->next) {
		prev += RangeSize(curSet->elem);
	}
	return prev;
}

 

