/* SPEX Grammar by Jeroen v. Yperen (JvY)
 * Version: 1.1 (19-03-07)
 *
 *
 * Changes in version 1.1:
 * - Recognize variable declarations
 * - Allow identifiers to be used with and without "P..->" or "now."
 */


// Declaration section

%{
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define YYERROR_VERBOSE
// #define YYDEBUG 1
 
void yyerror(const char *str);
int yywrap();
int main(void);

int yydebug = 0;
int fatalerr = 0;

extern int lineno;

/*
 * Structs for various rules
 */

typedef struct ident {
	char *str;
	char *name;
	struct exprpart *array_index;
	char *str_ind;
	struct ident *str_part;
} Ident;

struct exprpart {
	// Either 'expr' or 'id' will hold a value, the other should be NULL
	char *expr;
	struct ident *id;
	struct exprpart *next;
};

struct lrangepart {
	char *lb;
	char *ub;
	struct lrangepart *next;
};

struct range {
	struct ident *id;
	struct lrangepart *parts;
};

struct largs {
	struct exprpart *parts;
	struct largs *next;
};

struct lrange {
	struct range *ra;
	struct lrange *next;
};

typedef enum {
	t_range,
	t_arg,
} arg_or_range;
struct largrange {
	arg_or_range t;
	union {
		struct range *ra;
		struct exprpart *parts;
	} u;
	struct largrange *next;
};

struct compare {	
	struct ident *id;
	const char *op;
	struct exprpart *expr;
};

struct cmp_entry {
	char parentheses;
	char or_op;
	struct compare *cmp;
	struct cmp_entry *lhs;
	struct cmp_entry *rhs;
};

struct stringlist {
	char *label;
	struct stringlist *next;
};

struct mtypelist {
	char *label;
	char *intrepr;
	struct mtypelist *next;
};

struct chandecllist {
	char *chan;
	struct stringlist *types;
	struct chandecllist *next;
};

struct id_inst {
	struct ident *id;
	struct exprpart *expart;
	struct id_inst *next;
};

struct arg_list {
	char *type;
	struct id_inst *id_list;
	struct arg_list *next;
};
	
/*
 * Private structs
 */

typedef struct id_entry {
	char *type;
	char *label;
	struct id_entry *next;
} IdEntry;


/* 
 * Private functions
 */
const char *op_recognize(char *);
const char *dhas_op(const char *op);
const char *domain_op(const char *op);

struct exprpart	*expr_append(struct exprpart *, struct exprpart *);
struct exprpart	*expr_string(char *str);
struct exprpart	*args_flatten(struct largs *arguments);
struct largs 	*args_append(struct largs *, struct largs *);
struct lrangepart *num_range_append(struct lrangepart *, struct lrangepart *);
struct lrange 	*lrange_append(struct lrange *, struct lrange *);
struct largrange	*largrange_append(struct largrange *, struct largrange *);

void print_ident(struct ident *, char);
void print_str(struct ident *id);
void print_expr(struct exprpart *, char);

void print_cmpe_cond(struct cmp_entry *cmpe, char c_prefix);
void print_cmpe_eff(struct cmp_entry *cmpe, char c_prefix);
void print_cmpe(struct cmp_entry *cmpe);
void print_cmp(struct compare *cmp, char printStruct);
void print_cmp_cond(struct compare *cmp);
void print_cmp_eff(struct compare *cmp);
void print_cmpe_eff(struct cmp_entry *cmpe, char c_prefix);
void print_cmpe_or_eff(struct cmp_entry *cmpe);
void print_cmpe_and_eff(struct cmp_entry *cmpe);

IdEntry *findId(char *name);
int isRangeBetween(struct lrangepart *, int, int);
char * mtypenr(char *s);
struct stringlist *findchantypes(char *s);

/*
 * Private variables
 */
const char *OP_EQ = "==";
const char *OP_NEQ = "!=";
const char *OP_GTEQ = ">=";
const char *OP_GT = ">";
const char *OP_STEQ = "<=";
const char *OP_ST = "<";

char proctype_seen = 0;
char *last_proctype = NULL;

char *file_name =  NULL;


// Linked list for global identifiers.
IdEntry *glIds = NULL;
IdEntry *curGlId = NULL;
IdEntry *localIds = NULL;
IdEntry *curLclId = NULL;

struct mtypelist *mtypes = NULL;
struct chandecllist  *channels = NULL;

%}

%union { 
	struct largrange *arl;
	struct lrange *rl;
	struct range *r;
	struct lrangepart *rp;
	struct largs *al;
	struct exprpart *ep;
	struct ident *id;
	struct compare *cmp;
	struct cmp_entry *cmpe;
	struct stringlist *sl;
	struct mtypelist *ml;
	struct id_inst *ids;
	struct arg_list *args;
	char *str;
}

%token <str> SPEX_OUTPUT SPEX_INPUT SPEX_IO SPEX_CHAN
%token <str> NOW_STRUCT PROC_STRUCT
%token <str> DOMAIN
%token <str> SPEX
%token <str> IN DOTS OF
%token <str> STRING NUMBER OPERAND
%token <str> LOGIC_OR LOGIC_AND
%token <str> OTHER SINGLE_OTHER
%token <str> TYPE MTYPE CHAN
%token <str> PROCTYPE
%token <str> STRUCT_MEMB
%token <str> ARROW
%token <str> AMPERSAND


%type <str>	program lines line linemarker spexinput spexoutput spexio domain spexcmp
%type <str> 	anything
%type <r>	range
%type <rl>	ranges ranges_opt
%type <rp>	num_range num_ranges
%type <al> 	arguments arguments_opt 
%type <arl>	args_ranges_opt args_ranges arg_range
%type <id> 	identifier std_identifier norm_identifier 
%type <id>	norm_ident_ind norm_ident std_ident_ind
%type <str>	struct_ind
%type <cmpe> 	compare or_cmp and_cmp logic_cmp
%type <cmp>	op_cmp
%type <ep> 	expression expression_opt argument 
%type <ep> 	prom_thing prom_things prom_things_opt
%type <str>	prom_other
%type <str>	declaration declaration_2
%type <ids>	id_insts id_inst
%type <ids>	m_id_insts m_id_inst
%type <str>	proc_decl
%type <args>	arg_decl_opt arg_decl_list arg_decls
%type <ids>	arg_id_list
%type <sl>	datatypelist strings
%type <ml>	other_list
%type <str>	datatype
%type <str> 	eos

%%
// Rules section

program	: lines
	;
	
lines	: line
	| lines line
	;
	
line	: spexinput
	| spexoutput
	| spexio
	| domain
	| spexcmp
	| declaration
	| proc_decl
	| anything { printf("%s", $1); }
	| linemarker
	;
	
linemarker	: '#' NUMBER STRING
		  {
			char *tmp_file;
			int len;
	
			tmp_file = strdup($3 + 1);
			len = strlen(tmp_file);
			tmp_file[len-1] = '\0';
			
			if (file_name)
				free(file_name);
		  	file_name = tmp_file;
		  	lineno = atoi($2) - 1; /* number is for the next line */
		  	
		  	printf("# %s %s", $2, $3);
		  	
		  	$$ = NULL;
		  }
		  | '#' OTHER
		  {
		  	printf("# %s", $2);
		  }
		;	
	
spexio	: SPEX_IO
		  '('
		  identifier
		  args_ranges_opt
		  ')'
		  {
		  	struct largrange *cur_lr;
		  	struct range *cur_range;
		  	struct lrangepart *cur_lrp;
		  	struct stringlist *types;
		  	struct stringlist *cur_t;
		  	
		  	// TODO: make sure %% works all right
		  	// TODO: stricter type checking
		  	// TODO: make sure argument passing works
		  	//             (should we add casts?)
		  	// TODO: better way to deal with mtypes (add enums in C?)
		  	
		  	// based on:
		  	
		  	// spex_input(string, id1 in range1, id2 in range2, ...);
		  	// ->
		  	// #if SPEX
		  	// c_code {
		  	// spex__inputs(string,
		  	//		"id1", &id1, "range1",
		  	//		"id2", &id2, "range2",
		  	// 		...
		  	// ); };
		  	// #else
		  	// c_code {
		  	//	id1 = newDomain("range1");
		  	//	id2 = newDomain("range2");
		  	//	...
		  	// };
		  	// #endif
		  	
		  	// TODO Should check the string for amount and type of %?'s
		  	
		  	types = findchantypes($3->name);
		  	if (types == NULL) {
				fprintf(stderr, "%s:%d: error: no types defined for channel: %s\n", file_name, lineno, $3->name);
				fatalerr = 1;
				return NULL;
			}
		  	
		  	printf("\n#ifdef SPEX /*spexio*/\n");
		  	printf("c_code {\n\t");
		  	printf("spex__io(");
		  	printf("\"");
		  	printf("%s", $3->name);
		  	for (cur_lr = $4, cur_t = types; cur_lr != NULL && cur_t != NULL; cur_lr = cur_lr->next, cur_t = cur_t->next) {
			  	printf(" ! %%");
			  	if (cur_lr->t == t_range)
				  	printf("&");
				if (strcmp(cur_t->label, "int") == 0)
				  	printf("d");
				else if (strcmp(cur_t->label, "short") == 0)
				  	printf("h");
				else if (strcmp(cur_t->label, "mtype") == 0)
				  	printf("e");
				else if (strcmp(cur_t->label, "char") == 0)
				  	printf("c");
				else {
					fprintf(stderr, "%s:%d: error: unknown type of channel %s: %s\n", file_name, lineno, $3->name, cur_t->label);
					fatalerr = 1;
				  	printf("X");
				}
			}
		  	printf("\"");
		  	if (cur_lr == NULL && cur_t != NULL) {
				fprintf(stderr, "%s:%d: error: too few parameters for spexio on: %s\n", file_name, lineno, $3->name);
				fatalerr = 1;
		  	} else  if (cur_lr != NULL && cur_t == NULL) {
				fprintf(stderr, "%s:%d: error: too many parameters for spexio on: %s\n", file_name, lineno, $3->name);
				fatalerr = 1;
			}

		  	for (cur_lr = $4; cur_lr != NULL; cur_lr = cur_lr->next) {
			  	printf(",\n\t\t");
		  		if (cur_lr->t == t_range) {
			  		cur_range = cur_lr->u.ra;
			  		printf("\"");
			  		print_ident(cur_range->id, 0);
			  		printf("\", ");
			  		printf("&");
			  		print_ident(cur_range->id, 1);
			  		printf(", \"");
			  		for (cur_lrp = cur_range->parts; cur_lrp != NULL; cur_lrp = cur_lrp->next) {
			  			if (cur_lrp->lb == cur_lrp->ub) {
			  				printf("%s, ", cur_lrp->lb);
			  			} else {
			  				printf("%s..%s, ", cur_lrp->lb, cur_lrp->ub);
			  			}
			  		}
			  		printf("\"");
		  		} else {
			  		print_expr(cur_lr->u.parts, 1);	
		  		}
			  	if (cur_lr->next == NULL) {
			  		printf("\n\t");
			  	}
		  	}
		  	printf(");\n");
		  	printf("}\n");
		  	
		  	printf("#else\n");
		  	
		  	// If no assignments have to be made, insert a 'skip' command.
		  	if( $4 == NULL ) {
		  		printf("skip\n");
		  	} else {
			  	printf("c_code {\n");
			  	
			  	for (cur_lr = $4; cur_lr != NULL; cur_lr = cur_lr->next) {
			  		if (cur_lr->t == t_range && cur_lr->u.ra != NULL) {
				  		cur_range = cur_lr->u.ra;
				  		printf("\t");
				  		print_ident(cur_range->id, 1);
				  		printf(" = newDomain(\"");
				  		for (cur_lrp = cur_range->parts; cur_lrp != NULL; cur_lrp = cur_lrp->next) {
				  			if (cur_lrp->lb == cur_lrp->ub) {
				  				printf("%s, ", cur_lrp->lb);
				  			} else {
				  				printf("%s..%s, ", cur_lrp->lb, cur_lrp->ub);
				  			}
				  		}
				  		printf("\");\n");
				  	}
			  	}
			  	
			  	printf("}\n");
			  	
			  	//printf("printf(%s \"\\n\"", $3); /* String concatination like this doesn't work in Spin */
			  	printf("printf(%s ", $3->name);
			  	for ( cur_lr = $4; cur_lr != NULL; cur_lr = cur_lr->next) {
			  		if (cur_lr->t == t_arg && cur_lr->u.parts != NULL) {
			  			printf(", ");	
			  			print_expr(cur_lr->u.parts, 0);
			  		}
			  	}
			  	printf(");\n");
			}
		  	printf("#endif\n");
		  	
		  }
		;

spexinput	: SPEX_INPUT
		  '('
		  STRING
		  ranges_opt
		  ')'
		  {
		  	struct lrange *cur_lr;
		  	struct range *cur_range;
		  	struct lrangepart *cur_lrp;
		  	
		  	// spex_input(string, id1 in range1, id2 in range2, ...);
		  	// ->
		  	// #if SPEX
		  	// c_code {
		  	// spex__inputs(string,
		  	//		"id1", &id1, "range1",
		  	//		"id2", &id2, "range2",
		  	// 		...
		  	// ); };
		  	// #else
		  	// c_code {
		  	//	id1 = newDomain("range1");
		  	//	id2 = newDomain("range2");
		  	//	...
		  	// };
		  	// #endif
		  	
		  	// TODO Should check the string for amount and type of %?'s
		  	
		  	printf("\n#ifdef SPEX /*spexinput*/\n");
		  	printf("c_code {\n\t");
		  	printf("spex__inputs(%s", $3);

		  	for (cur_lr = $4; cur_lr != NULL && cur_lr->ra != NULL; cur_lr = cur_lr->next) {
		  		cur_range = cur_lr->ra;
		  		printf(",\n\t\t");
		  		printf("\"");
		  		print_ident(cur_range->id, 0);
		  		printf("\", ");
		  		printf("&");
		  		print_ident(cur_range->id, 1);
		  		printf(", \"");
		  		for (cur_lrp = cur_range->parts; cur_lrp != NULL; cur_lrp = cur_lrp->next) {
		  			if (cur_lrp->lb == cur_lrp->ub) {
		  				printf("%s, ", cur_lrp->lb);
		  			} else {
		  				printf("%s..%s, ", cur_lrp->lb, cur_lrp->ub);
		  			}
		  		}
		  		printf("\"");
		  		if (cur_lr->next == NULL) {
		  			printf("\n\t");
		  		}
		  	}
		  	printf(");\n");
		  	printf("}\n");
		  	
		  	printf("#else\n");
		  	
		  	// If no assignments have to be made, insert a 'skip' command.
		  	if( $4 == NULL ) {
		  		printf("skip\n");
		  	} else {
			  	printf("c_code {\n");
			  	
			  	for (cur_lr = $4; cur_lr != NULL && cur_lr->ra != NULL; cur_lr = cur_lr->next) {
			  		cur_range = cur_lr->ra;
			  		printf("\t");
			  		print_ident(cur_range->id, 1);
			  		printf(" = newDomain(\"");
			  		for (cur_lrp = cur_range->parts; cur_lrp != NULL; cur_lrp = cur_lrp->next) {
			  			if (cur_lrp->lb == cur_lrp->ub) {
			  				printf("%s, ", cur_lrp->lb);
			  			} else {
			  				printf("%s..%s, ", cur_lrp->lb, cur_lrp->ub);
			  			}
			  		}
			  		printf("\");\n");
			  	}
			  	
			  	printf("}\n");
			}
		  	printf("#endif\n");
		  	
		  }
		;

spexoutput	: SPEX_OUTPUT
		  '('
		  strings
		  arguments_opt
		  ')'
		  {
		  	struct largs *cur_arg;
		  	struct stringlist *cur_sl;
		  	
		  	//#if SPEX
		  	// c_code { spex__output(string, expr1, expr2, ...); };
		  	//#else
		  	// printf(string, expr1, expr2, ...);
		  	//#endif
		  	
		  	// TODO Should change %c and %h to %d in printf string.
		  	
		  	printf("\n#ifdef SPEX /*spexoutput*/\n");
		  	printf("c_code { spex__outputs(");
		  	
		  	for ( cur_sl = $3; cur_sl != NULL ; cur_sl = cur_sl->next ) {
		  		printf(" %s", cur_sl->label);
		  	}
		  	
		  	for ( cur_arg = $4; cur_arg != NULL; cur_arg = cur_arg->next) {
		  		printf(", ");
		  		print_expr(cur_arg->parts, 1);	
		  	}
		  	printf("); }\n#else\n");
		  	//printf("printf(%s \"\\n\"", $3); /* String concatination like this doesn't work in Spin */
		  	printf("printf(");
		  	for ( cur_sl = $3; cur_sl != NULL ; cur_sl = cur_sl->next ) {
		  		printf(" %s", cur_sl->label);
		  	}
		  	for ( cur_arg = $4; cur_arg != NULL; cur_arg = cur_arg->next) {
		  		printf(", ");	
		  		print_expr(cur_arg->parts, 0);	
		  	}
		  	printf(");\n#endif\n");

		  	
		  }

		;

domain		: DOMAIN 
		  '('
		  compare
		  ')'
		  { 
		  	struct compare *cmp;
		  	
		  	cmp = $3->cmp;

			// 'DOMAIN( id __ expr );'
			// ->
		  	// 'atomic { c_expr { DHas__(id, (expr)) } -> c_code { id = Domain__(id, (expr)); }; };'

//		  	printf("atomic { c_expr { %s(", dhas_op(cmp->op));
//		  	print_ident(cmp->id, 1);
//		  	printf(", (");
//
//		  	print_expr(cmp->expr, 1);				
//		  	printf(")) } -> ");
//		  	
//		  						
//			printf("c_code {");
//			print_ident(cmp->id, 1);
//			printf(" = %s(", domain_op(cmp->op));
//			print_ident(cmp->id, 1);
//			printf(", (");
//
//			print_expr(cmp->expr, 1);
//		  	printf(")); }; }");

//			print_cmpe($3, 1);

			printf("atomic { ");
			print_cmpe_cond($3, 1);
			printf(" ->\n\t");
			print_cmpe_eff($3, 1);
			printf("; }");
		  	
		  }
		| DOMAIN 
		  '('
		  ranges
		  ')'
		  { 
		  	struct lrange *cur_lr;
		  	struct range *cur_range;
		  	struct lrangepart *cur_lrp;
		  	
		  	// DOMAIN( id1 in {range1}, id2 in {range2}, ...)
		  	// ->
		  	// c_code {
		  	// id1 = newDomain(range1);
		  	// id2 = newDomain(range2);
		  	// ...
		  	// };
		  	
		  	printf("\nc_code{\n");
		  	
		  	for (cur_lr = $3; cur_lr != NULL && cur_lr->ra != NULL; cur_lr = cur_lr->next) {
		  		cur_range = cur_lr->ra;
		  		printf("\t");
		  		print_ident(cur_range->id, 1);
		  		printf(" = newDomain(\"");
		  		for (cur_lrp = cur_range->parts; cur_lrp != NULL; cur_lrp = cur_lrp->next) {
		  			if (cur_lrp->lb == cur_lrp->ub) {
		  				printf("%s, ", cur_lrp->lb);
		  			} else {
		  				printf("%s..%s, ", cur_lrp->lb, cur_lrp->ub);
		  			}
		  		}
		  		printf("\");\n");
		  	}
		  	printf("}\n");
		  	
		  }
		;

spexcmp		: SPEX
		  '('
		  compare
		  ')'
		  {
		  	//#if SPEX
		  	// (id __ (expr));
		  	//#else
		  	//atomic { c_expr { DHas__(id, (expr)) } -> c_code { id = D__(id, (expr)); }; };
		  	//#endif


		  	printf("\n#ifdef SPEX /*spexcmp*/\n(");
		  	print_cmpe($3);
		  	printf(")\n#else\n");

			printf("atomic { ");
			print_cmpe_cond($3, 1);
			printf(" ->\n\t");
			print_cmpe_eff($3, 1);
			printf("; }");
		  	printf("\n#endif\n");
		  }
		;	

strings		: STRING
		  {
		  	struct stringlist *newol;
		  	newol = calloc(1, sizeof(struct stringlist));
		  	newol->label = $1;
		  	
		  	$$ = newol;
		  }
		| strings STRING
		  {
		  	struct stringlist *newol, *endol;
		  	newol = calloc(1, sizeof(struct stringlist));
		  	newol->label = $2;
		  	
		  	for ( endol = $1; endol->next != NULL; endol = endol->next) {
		  		;
		  	}
		  	endol->next = newol;
		  	
		  	$$ = $1;
		  }

		;

declaration	: declaration_2 eos { printf(";"); }
		| declaration_2 '}' { printf("\n}"); }
		;
		
declaration_2	: id_declaration { $$ = NULL; }
		| mtype_declaration { $$ = NULL; }
		| chan_declaration { $$ = NULL; }
		| spexchan_declaration { $$ = NULL; }
		;
		
id_declaration	: data_id_decl
		| mtype_id_decl
		| chan_id_decl
		;
		
mtype_declaration :
		  MTYPE '=' '{' other_list comma_opt '}'
		  {
		  	struct mtypelist *ol;
		  	int nmtypes = 0;
		  	int i;
		  	char buf[15];
		  	
		  	mtypes = $4;
		  	printf("%s = { ", $1);
		  	for ( ol = $4; ol != NULL; ol = ol->next ) {
		  		printf("%s", ol->label);
		  		if ( ol->next != NULL ) {
		  			printf(", ");
		  		}
		  		nmtypes++;
		  	}
		  	printf(" }");
		  	i = nmtypes;
		  	for ( ol = $4; ol != NULL; ol = ol->next ) {
		  		sprintf(buf, "%d", i);
		  		ol->intrepr = strdup(buf);
		  		i--;
		  	}
		  }
		;

other_list	: OTHER
		  {
		  	struct mtypelist *newol;
		  	newol = calloc(1, sizeof(struct mtypelist));
		  	newol->label = $1;
		  	
		  	$$ = newol;
		  }
		| other_list ',' OTHER 
		  {
		  	struct mtypelist *newol, *endol;
		  	newol = calloc(1, sizeof(struct mtypelist));
		  	newol->label = $3;
		  	
		  	for ( endol = $1; endol->next != NULL; endol = endol->next) {
		  		;
		  	}
		  	endol->next = newol;
		  	
		  	$$ = $1;
		  }		  	
		;
		
comma_opt	: /* Can be empty */
		| ','
		;

		
chan_declaration :
		  CHAN std_ident_ind '=' '[' NUMBER ']' OF '{' datatypelist '}'
		  {
		  	struct stringlist *curs;
		  	
		  	printf("chan ");
		  	print_ident($2, 0);
		  	printf(" = [%s] of {", $5);
		  	
		  	for (curs = $9; curs != NULL; curs = curs->next) {
		  		printf(" %s", curs->label);
		  		if ( curs->next != NULL) {
		  			printf(",");
		  		}
		  	}
		  	printf(" }");
		  }
		;
		
spexchan_declaration :
		  SPEX_CHAN std_ident_ind '=' '[' NUMBER ']' OF  '{' datatypelist '}'
		  {
		  	struct chandecllist *newcl;
		  	newcl = calloc(1, sizeof(struct chandecllist));
		  	newcl->chan = $2->name;
		  	newcl->types = $9;
		  	newcl->next = channels;
		  	channels = newcl;
		  	if (atoi($5) != 0) {
		  		fprintf(stderr, "%s:%d: error: spexchan size must be 0 (not %s) for channel: %s\n", file_name, lineno, $5, $2->name);
		  		fatalerr = 1;
		  	}
		  		
		  }
		;
		
data_id_decl	: TYPE id_insts
		  {
		  	struct id_inst *curInst;
	  		IdEntry *newId = NULL;
	  		IdEntry *firstNewId = NULL;
		  	IdEntry *lastNewId = NULL;
		  	
		  	printf("%s ", $1);
		  	for (curInst = $2; curInst != NULL; curInst = curInst->next) {
		  		
		  		print_ident(curInst->id, 0);
		  		if ( curInst->expart ) {
		  			printf(" = ");
		  			print_expr(curInst->expart, 0);
		  		}
		  		
		  		if (curInst->next != NULL) {
		  			printf(", ");
		  		}
		  		
		  		// Append new IdEntry to local list
		  		newId = (IdEntry *) calloc(1, sizeof(IdEntry));
		  		newId->label = curInst->id->name; // Needs strdup?
		  		newId->type = $1;
		  		
		  		if ( firstNewId == NULL ) {
		  			firstNewId = lastNewId = newId;
		  		} else {
		  			lastNewId->next = newId;
		  			lastNewId = newId;
		  		}
		  		
		  	}
		  	//printf(";");

			
		  	// Append local IdEntry list to the appropriate global list
		  	if (proctype_seen) {
		  		if ( localIds == NULL ) {
		  			localIds = curLclId = firstNewId;
		  		} else {
		  			curLclId->next = lastNewId;
		  			curLclId = lastNewId;
		  		}
		  	} else {
		  		if ( glIds == NULL ) {
		  			glIds = curGlId = firstNewId;
		  		} else {
		  			curGlId->next = lastNewId;
		  			curGlId = lastNewId;
		  		}
		  		
		  	}
		  }
		;

mtype_id_decl 	: MTYPE m_id_insts
		  {
		  	struct id_inst *curInst;
	  		IdEntry *newId;

	  		newId = (IdEntry *) calloc(1, sizeof(IdEntry));
	  		newId->label = $2->id->name; // Needs strdup?
	  		newId->type = "mtype";
		  	
		  	printf("%s ", $1);
		  	for (curInst = $2; curInst != NULL; curInst = curInst->next) {
		  		
		  		print_ident(curInst->id, 0);
		  		if ( curInst->expart ) {
		  			printf(" = ");
		  			print_expr(curInst->expart, 0);
		  		}
		  		
		  		if (curInst->next != NULL) {
		  			printf(", ");
		  		}
		  		
		  	}
		  	//printf(";");
		  	
		  	if (proctype_seen) {
		  		if ( localIds == NULL ) {
		  			localIds = curLclId = newId;
		  		} else {
		  			curLclId->next = newId;
		  			curLclId = newId;
		  		}
		  		//printf("/* Local declaration of %s %s */", $1, $2->id->name);
		  	} else {
		  		if ( glIds == NULL ) {
		  			glIds = curGlId = newId;
		  		} else {
		  			curGlId->next = newId;
		  			curGlId = newId;
		  		}
		  		
		  		//printf("/* Global declaration of %s %s */", $1, newId->label);
		  		
		  	}
		  }
		;
	
chan_id_decl	: CHAN std_identifier 
		  {
	  		IdEntry *newId;

	  		newId = (IdEntry *) calloc(1, sizeof(IdEntry));
	  		newId->label = $2->name; // Needs strdup?
	  		newId->type = "chan";
		  	
		  	printf("chan ");
		  	print_ident($2, 0);
		  	
		  	if (proctype_seen) {
		  		if ( localIds == NULL ) {
		  			localIds = curLclId = newId;
		  		} else {
		  			curLclId->next = newId;
		  			curLclId = newId;
		  		}
		  		//printf("/* Local declaration of %s %s */", $1, $2->id->name);
		  	} else {
		  		if ( glIds == NULL ) {
		  			glIds = curGlId = newId;
		  		} else {
		  			curGlId->next = newId;
		  			curGlId = newId;
		  		}
		  		
		  		//printf("/* Global declaration of %s %s */", $1, newId->label);
		  		
		  	}
		  }
		| CHAN std_ident_ind '=' std_identifier
		  {
	  		IdEntry *newId;

	  		newId = (IdEntry *) calloc(1, sizeof(IdEntry));
	  		newId->label = $2->name; // Needs strdup?
	  		newId->type = "chan";

		  	printf("chan ");
		  	print_ident($2, 0);
		  	printf(" = ");
		  	print_ident($4, 0);

		  	if (proctype_seen) {
		  		if ( localIds == NULL ) {
		  			localIds = curLclId = newId;
		  		} else {
		  			curLclId->next = newId;
		  			curLclId = newId;
		  		}
		  		//printf("/* Local declaration of %s %s */", $1, $2->id->name);
		  	} else {
		  		if ( glIds == NULL ) {
		  			glIds = curGlId = newId;
		  		} else {
		  			curGlId->next = newId;
		  			curGlId = newId;
		  		}
		  		
		  		//printf("/* Global declaration of %s %s */", $1, newId->label);
		  		
		  	}
		  }
		;


id_insts	: id_inst
		| id_insts ',' id_inst
		  {
		  	struct id_inst *endol;
		  	for(endol = $1; endol->next != NULL; endol = endol->next) {
		  		;
		  	}
		  	endol->next = $3;
		  	$$ = $1;
		  }
		;
		
m_id_insts	: m_id_inst
		| m_id_insts ',' m_id_inst
		  {
		  	struct id_inst *endol;
		  	for(endol = $1; endol->next != NULL; endol = endol->next) {
		  		;
		  	}
		  	endol->next = $3;
		  	$$ = $1;
		  }
		;

m_id_inst	: std_identifier
		  {
		  	struct id_inst *newid;
		  	newid = calloc(1, sizeof(struct id_inst));
		  	newid->id = $1;
		  	
		  	$$ = newid;
		  }
		| std_identifier '=' OTHER
		  {
		  	struct id_inst *newid;
			struct exprpart *newpart;		  	
		  	newid = calloc(1, sizeof(struct id_inst));
		  	newid->id = $1;
		  	
		  	newpart = calloc(1, sizeof(struct exprpart));
		  	newpart->expr = $3;
		  	newid->expart = newpart;
		  	
		  	$$ = newid;
		  }
		;

id_inst		: std_identifier
		  {
		  	struct id_inst *newid;
		  	newid = calloc(1, sizeof(struct id_inst));
		  	newid->id = $1;
		  	
		  	$$ = newid;
		  }
		| std_identifier '=' expression
		  {
		  	struct id_inst *newid;
		  	newid = calloc(1, sizeof(struct id_inst));
		  	newid->id = $1;
		  	newid->expart = $3;
		  	
		  	$$ = newid;
		  }
		;

datatype	: MTYPE
		| TYPE
		| CHAN
		| OTHER
		;
		
datatypelist	: datatype
		  {
		  	struct stringlist *newsl;
		  	newsl = calloc(1, sizeof(struct stringlist));
		  	newsl->label = $1;
		  	
		  	$$ = newsl;
		  }
		| datatypelist ',' datatype
		  {
		  	struct stringlist *newsl, *endsl;
		  	newsl = calloc(1, sizeof(struct stringlist));
		  	newsl->label = $3;
		  	
		  	for ( endsl = $1; endsl->next != NULL; endsl = endsl->next) {
		  		;
		  	}
		  	endsl->next = newsl;
		  	
		  	$$ = $1;
		  }		  	
		;
		

proc_decl	: PROCTYPE 
		  OTHER 
		  {
		  	IdEntry *tmpId, *nextId;
		  	
		  	printf("%s %s", $1, $2);

		  	last_proctype = strdup($2);
		  	proctype_seen = 1;
		  	
		  	for ( tmpId = localIds; tmpId != NULL; tmpId = nextId ) {
		  		nextId = tmpId->next;
		  		free(tmpId);
		  	}
		  	localIds = curLclId = NULL;
		  }
		  '(' 
		  arg_decl_opt 
		  ')'
		  {
		  	struct arg_list *cur_als;
		  	struct id_inst *cur_ids;
		  	
		  	
		  	printf("( ");
		  	for ( cur_als = $5; cur_als != NULL; cur_als = cur_als->next ) {
		  		printf("%s ", cur_als->type);
		  		for ( cur_ids = cur_als->id_list; cur_ids != NULL; cur_ids = cur_ids->next ) {
		  			IdEntry *newId;

		  			newId = (IdEntry *) calloc(1, sizeof(IdEntry));
		  			newId->label=cur_ids->id->name;
		  			newId->type = cur_als->type;
			  		if ( localIds == NULL ) {
			  			localIds = curLclId = newId;
			  		} else {
			  			curLclId->next = newId;
			  			curLclId = newId;
			  		}
		  			
		  			print_ident(cur_ids->id, 0);
		  			if (cur_ids->next != NULL) {
		  				printf(", ");
		  			}
		  		}
		  		if ( cur_als->next != NULL ) {
		  			printf("; ");
		  		}
		  	}
		  	printf(" ) ");
		  }
		;
		
arg_decl_opt	: /* Can be empty */
		  { $$ = NULL; }
		| arg_decl_list
		;

arg_decl_list	: arg_decls
		| arg_decl_list ';' arg_decls
		  {
		  	struct arg_list *end_als;
		  	for(end_als = $1; end_als->next != NULL; end_als = end_als->next) {
		  		;
		  	}
		  	end_als->next = $3;
		  	$$ = $1;
		  }
		;
		
arg_decls	: datatype arg_id_list
		  {
		  	struct arg_list *als;
		  	als = calloc(1, sizeof(struct arg_list));
		  	
		  	als->type = $1;
		  	als->id_list = $2;
		  	
		  	$$ = als;
		  }
		;
		
arg_id_list	: std_ident_ind 
		  {
		  	struct id_inst *newid;
		  	newid = calloc(1, sizeof(struct id_inst));
		  	newid->id = $1;
		  	
		  	$$ = newid;
		  }
		| arg_id_list ',' std_ident_ind
		  {
		  	struct id_inst *endid;
		  	struct id_inst *newid;
		  	newid = calloc(1, sizeof(struct id_inst));
		  	newid->id = $3;

		  	for(endid = $1; endid->next != NULL; endid = endid->next) {
		  		;
		  	}
		  	endid->next = newid;
		  	$$ = $1;
		  }
		;
	
anything	: eos
		| OTHER
		| SINGLE_OTHER
		| STRING
		| PROC_STRUCT
		| NOW_STRUCT
		| NUMBER
		| OPERAND
		| IN
		| STRUCT_MEMB
		| LOGIC_AND
		| LOGIC_OR
		| ']' { $$ = "]"; }
		| '[' { $$ = "["; }
		| '(' { $$ = "("; }
		| ')' { $$ = ")"; }
		| '{' { $$ = "{"; }
		| '}' { $$ = "}"; }
		| ',' { $$ = ","; }
		| '=' { $$ = "="; }
		//| '#' { $$ = "#"; }
		;

eos		: ';' { $$ = ";"; }
		| ARROW
		;

args_ranges_opt 	: /* can be empty */
		  { $$ = NULL; }
		| ',' args_ranges
		  { $$ = $2; }
		;

args_ranges:  arg_range
		  { $$ = $1; }
		| args_ranges 
		  ',' 
		  arg_range
		  {
			$$ = largrange_append($1, $3);
		  }	
		;

arg_range: AMPERSAND range
	          { 
	          	struct largrange *rp;
	          	rp = calloc(1, sizeof(struct largrange));
	          	rp->u.ra = $2;
	          	rp->t = t_range;
	          	$$ = rp;
	          }
		| argument
		  {
			struct largrange *arg;
			arg = calloc(1, sizeof(struct largrange));
			arg->u.parts = $1;
	          	arg->t = t_arg;
		  	$$ = arg;
		  }
		;
	

ranges_opt 	: /* can be empty */
		  { $$ = NULL; }
		| ',' ranges
		  { $$ = $2; }
		;

ranges		: range
	          { 
	          	struct lrange *rp;
	          	rp = calloc(1, sizeof(struct lrange));
	          	rp->ra = $1;
	          	$$ = rp;
	          }
		| ranges 
		  ',' 
		  range
		  {
	          	struct lrange *rp;
	          	rp = calloc(1, sizeof(struct lrange));
	          	rp->ra = $3;
		  	$$ = lrange_append($1, rp);
		  }
		;

range		: identifier
		  IN
		  '{'
		  num_ranges
		  '}'
		  {
		  	struct range *r;
		  	struct lrangepart *lrp = NULL;
			IdEntry *tmpId;
			char invalidRange = 0;
			
			int lb, ub;
			char *lbs, *ubs;

		  	tmpId = findId($1->name);
		  	
		  	if ( tmpId != NULL ) {
		  		if ( (strcmp(tmpId->type, "bit")) == 0 ) {
		  			lb  =  0; 	ub  =  1;
		  			lbs = "0"; 	ubs = "1";
		  		} else if ( (strcmp(tmpId->type, "bool")) == 0 ) {
		  			lb  =  0; 	ub  =  1;
		  			lbs = "0"; 	ubs = "1";
		  		} else if ( (strcmp(tmpId->type, "byte")) == 0 ) {
		  			lb  =  0; 	ub  =  255;
		  			lbs = "0"; 	ubs = "255";
		  		} else if ( (strcmp(tmpId->type, "mtype")) == 0 ) {
		  			lb  =  1; 	ub  =  255;
		  			lbs = "1"; 	ubs = "255";
		  		} else if ( (strcmp(tmpId->type, "chan")) == 0 ) {
		  			lb  =  0; 	ub  =  255;
		  			lbs = "0"; 	ubs = "255";
		  		} else if ( (strcmp(tmpId->type, "short")) == 0 ) {
		  			lb = -32767;	lbs = "-32767";
		  			ub =  32767;	ubs =  "32767";
		  		} else if ( (strcmp(tmpId->type, "int")) == 0 ) {
		  			lb  =  -2147483647;	ub  =  2147483647;
		  			lbs = "-2147483647";	ubs = "2147483647";
		  		} else {
		  			// Unknown type might be an int, so allow the int range.
		  			lb = -2147483647;
		  			ub =  2147483647;
		  			// But if it's invalid, we will select the default range of 0..255
		  			lbs = "0";
		  			ubs = "255";
		  		}
		  		
		  		if ( ! isRangeBetween($4, lb, ub) ) {
		  			lrp = calloc(1, sizeof(struct lrangepart));
		  			lrp->lb = strdup(lbs);
		  			lrp->ub = strdup(ubs);
		  			invalidRange = 1;
		  		} 
	  		}

			if ( invalidRange ) {
				fprintf(stderr, "%s:%d: warning: invalid range specified for: %s\n", file_name, lineno, $1->name);
				fprintf(stderr, "%s:%d: using default range instead: (%s..%s)\n", file_name, lineno, lrp->lb, lrp->ub);
			}
			
		  	r = calloc(1, sizeof(struct range));
		  	r->id = $1;
		  	r->parts = invalidRange ? lrp: $4;
		  	$$ = r;
		  }
		| identifier
		  {
		  	struct range *r;
		  	struct lrangepart *lrp;
			
			IdEntry *tmpId;

		  	r = calloc(1, sizeof(struct range));
		  	r->id = $1;

		  	lrp = calloc(1, sizeof(struct lrangepart));
		  	
		  	// Check to see if the identifier has been defined
		  	// in either global or local context.
		  	// If it is, create the lower- and upperbound depending
		  	// on the type of it.
		  	
		  	tmpId = findId($1->name);
		  	
		  	if (tmpId == NULL ) {
			  	lrp->lb = strdup("0");
			  	lrp->ub = strdup("255");
		  	} else {
		  		if ( (strcmp(tmpId->type, "bit")) == 0 ) {
				  	lrp->lb = strdup("0");
				  	lrp->ub = strdup("1");
		  		} else if ( (strcmp(tmpId->type, "bool")) == 0 ) {
				  	lrp->lb = strdup("0");
				  	lrp->ub = strdup("1");
		  		} else if ( (strcmp(tmpId->type, "byte")) == 0 ) {
				  	lrp->lb = strdup("0");
				  	lrp->ub = strdup("255");
		  		} else if ( (strcmp(tmpId->type, "short")) == 0 ) {
				  	lrp->lb = strdup("-32767");
				  	lrp->ub = strdup("32767");
		  		} else if ( (strcmp(tmpId->type, "int")) == 0 ) {
				  	lrp->lb = strdup("-2147483647");
				  	lrp->ub = strdup("2147483647");
		  		} else {
		  			// TODO Should limit mtype to the amount of mtypes defined.
		  			
		  			// For chan or mtype
				  	lrp->lb = strdup("0");
				  	lrp->ub = strdup("255");
		  		}
		  	}	


		  	r->parts = lrp;
		  	$$ = r;
		  }
		;

num_ranges	: num_range
		| num_ranges 
		  ',' 
		  num_range
		  {
		  	$$ = num_range_append($1, $3);
		  }
		;

num_range	: NUMBER
		  {
		  	struct lrangepart *lrp;
		  	lrp = calloc(1, sizeof(struct lrangepart));
		  	lrp->lb = $1;
		  	lrp->ub = lrp->lb;
		  	$$ = lrp;
		  }
		| NUMBER DOTS NUMBER
		  {
		  	struct lrangepart *lrp;
		  	lrp = calloc(1, sizeof(struct lrangepart));
		  	lrp->lb = $1;
		  	lrp->ub = $3;
		  	$$ = lrp;
		  }
		;
		
arguments_opt	: /* can be empty */
		  {	$$ = NULL;	}
		| ',' 
		  arguments
		  {  	$$ = $2;	}
		;

arguments	: argument
		  {
			struct largs *arg;
			arg = calloc(1, sizeof(struct largs));
			arg->parts = $1;
		  	$$ = arg;
		  }
		| arguments 
		  ',' 
		  argument
		  {
			struct largs *arg;
			
			arg = calloc(1, sizeof(struct largs));
			arg->parts = $3;
			$$ = args_append($1, arg);
		  }	
		;

argument	: expression
		;

expression_opt  : /* can be empty */
		  { $$ = NULL; }
		| expression
		;
		
expression	: prom_things
		| expression_opt
		  '(' 
		  arguments
		  ')'
		  prom_things_opt
		  {
			struct exprpart *expr;
			
			expr = expr_append(expr_string(")"), $5);
			expr = expr_append(args_flatten($3), expr);
			expr = expr_append(expr_string("("), expr);
			expr = expr_append($1, expr);
			
		  	//expr = expr_append($1, expr_string("("));
		  	//expr = expr_append(expr, args_flatten($3));
		  	//expr = expr_append(expr, expr_string(")"));
		  	//expr = expr_append(expr, $5);
		  	$$ = expr;
		  }
		;

prom_things_opt	: /* can be empty */
		  { $$ = NULL; }
		| prom_things
		;

prom_things	: prom_things 
		  prom_thing
		  { 
		  	$$ = expr_append($1, $2);
		  }
		| prom_thing
		;

prom_thing	: prom_other
		  {
		  	struct exprpart *prom;
		  	prom = calloc(1, sizeof(struct exprpart));
		  	prom->expr = $1;
		  	$$ = prom;
		  }
		| identifier
		  {
		  	struct exprpart *prom;
		  	char * mnr;
		  	
		  	prom = calloc(1, sizeof(struct exprpart));
		  	mnr = mtypenr($1->name);
		  	if (mnr != NULL)
			  	prom->expr = mnr;
		  	else
			  	prom->id = $1;
		  	$$ = prom;
		  }
		;		

prom_other	: STRING
		| NUMBER
		| SINGLE_OTHER
		;

compare		: or_cmp
		;
		
or_cmp		: and_cmp
		| or_cmp LOGIC_OR and_cmp
		  {
		  	struct cmp_entry *newEntr;
		  	
		  	newEntr = (struct cmp_entry *) calloc(1, sizeof(struct cmp_entry));

			newEntr->or_op = 1;
			
			newEntr->lhs = $1;
			newEntr->rhs = $3;
			
			$$ = newEntr;
		  }
		;

and_cmp		: logic_cmp
		| and_cmp LOGIC_AND logic_cmp
		  {
		  	struct cmp_entry *newEntr;
		  	
		  	newEntr = (struct cmp_entry *) calloc(1, sizeof(struct cmp_entry));

			newEntr->or_op = 0;
			
			newEntr->lhs = $1;
			newEntr->rhs = $3;
			
			$$ = newEntr;
		  }
		;
		
logic_cmp	: '(' compare ')'
		  {
		  	$2->parentheses += 1;
		  	
		  	$$ = $2;
		  }
		| op_cmp
		  {
		  	struct cmp_entry *newEntr;
		  	
		  	newEntr = (struct cmp_entry *) calloc(1, sizeof(struct cmp_entry));
		  	newEntr->cmp = $1;
		  	
		  	$$ = newEntr;
		  }
		;

op_cmp		: identifier
		  OPERAND 
		  expression 
		  {
		  	// We need to look only at the first two characters of the operand.
		  	struct compare *cmp;
		  	cmp = malloc(sizeof(struct compare));
		  	cmp->id = $1;
		  	cmp->op = op_recognize($2);
		  	cmp->expr = $3;
		  	$$ = cmp;
		  }
		;

identifier	: std_identifier
		| norm_identifier
		;

std_identifier	: std_ident_ind
		| std_ident_ind struct_ind std_identifier
		  {
		  	$1->str_ind = $2;
		  	$1->str_part = $3;
		  	$$ = $1;
		  }
		;

norm_identifier	: norm_ident_ind
		| norm_ident_ind struct_ind std_identifier
		  {
		  	$1->str_ind = $2;
		  	$1->str_part = $3;
		  	$$ = $1;
		  }
		;

struct_ind	: STRUCT_MEMB
//		| STRUCT_POINT
		;

norm_ident_ind	: norm_ident
		| norm_ident
		  '['
		  expression
		  ']'
		  {
		  	$$ = $1;
		  	$1->array_index = $3;
		  }
		;

std_ident_ind	: OTHER
		  {
			struct ident *id;
			id = calloc(1,sizeof(struct ident));
			id->str = NULL;
			id->name = strdup($1);
			id->array_index = NULL;
			$$ = id;
		  }
		| OTHER
		  '['
		  expression
		  ']'  	
		  {
			struct ident *id;
			id = calloc(1,sizeof(struct ident));
			id->str = NULL;
			id->name = strdup($1);
			id->array_index = $3;
			$$ = id;
		  }
		;

norm_ident	: NOW_STRUCT OTHER { 
			struct ident *id;
			id = calloc(1,sizeof(struct ident));
			id->str = strdup($1);
			id->name = strdup($2);
			id->array_index = NULL;
			$$ = id;
		  }
		| PROC_STRUCT OTHER { 
			struct ident *id;
			id = calloc(1,sizeof(struct ident));
			id->str = strdup($1);
			id->name = strdup($2);
			id->array_index = NULL;
			$$ = id;
		  }
		;
	 								
%%		
// Program section

void yyerror(const char *str)
{
        //fprintf(stderr,"error: \"%s\" at token \"%s\" in %s:%d\n",str, yylval.str, file_name, lineno);
        fprintf(stderr,"%s:%d %s at token \"%s\"\n", file_name, lineno, str, yylval.str);
        fatalerr = 1;

}
 
int yywrap()
{
        return 1;
} 
  
int main(void)
{
	file_name = strdup("<stdin>");
        yyparse();
        return fatalerr;
} 

const char *op_recognize(char *op) {
	if(op[0] == '=' && op[1] == '=') {
		return OP_EQ;
	} else if (op[0] == '!' && op[1] == '=') {
		return OP_NEQ;
	} else if (op[0] == '>' && op[1] == '=') {
		return OP_GTEQ;
	} else if (op[0] == '<' && op[1] == '=') {
		return OP_STEQ;
	} else if (op[0] == '>') {
		return OP_GT;
	} else if (op[0] == '<') {
		return OP_ST;
	} else {
		fprintf(stderr, "%s:%d: error: operand was not recognized: %s\n", file_name, lineno, op);
		fatalerr = 1;
		return NULL;
	}
}

const char *dhas_op(const char *op) {
  	if ( op == OP_EQ ) {
		return "DHas";
	} else if ( op == OP_NEQ) {
		return "! DHas";
	} else if ( op == OP_GTEQ) {
		return "DHasGtEq";
	} else if ( op == OP_GT) {
		return "DHasGt";
	} else if ( op == OP_STEQ) {
		return "DHasStEq";
	} else if ( op == OP_ST) {
		return "DHasSt";
	} else {
		return NULL;
	}
}

const char *domain_op(const char *op) {
  	if ( op == OP_EQ ) {
		return "DomainEq";
	} else if ( op == OP_NEQ) {
		return "DomainNEq";
	} else if ( op == OP_GTEQ) {
		return "DomainGtEq";
	} else if ( op == OP_GT) {
		return "DomainGt";
	} else if ( op == OP_STEQ) {
		return "DomainStEq";
	} else if ( op == OP_ST) {
		return "DomainSt";
	} else {
		return NULL;
	}
}	
struct exprpart *expr_append(struct exprpart *list, struct exprpart *part) {
	struct exprpart *cur = NULL;
	
	if ( list == NULL ) {
		return part;
	}
	for(cur = list; cur->next != NULL; cur = cur->next) {
		;
	}
	cur->next = part;
	return list;
	
}

struct exprpart	*expr_string(char *str) {
	struct exprpart *part;
	part = calloc(1, sizeof(struct exprpart));
	part->expr = strdup(str);
	return part;
}

struct exprpart *args_flatten(struct largs *arguments) {
	struct largs *cur_arg = NULL;
	// First part of the expression
	struct exprpart *first_expr = NULL;
	// Last part of the expression
	struct exprpart *cur_expr = NULL;
	// Temporary part when we need to add a ','
	struct exprpart *comma_expr = NULL;
		
	for ( cur_arg = arguments; cur_arg != NULL; cur_arg = cur_arg->next ) {
		if (first_expr == NULL) {
			first_expr = cur_expr = cur_arg->parts;
			while (cur_expr->next != NULL) {
				cur_expr = cur_expr->next;
			}
		} else {
			// Make a expr for the ','
			comma_expr = expr_string(",");
			// Append the comma_expr to the end of our current expr
			expr_append(cur_expr, comma_expr);
			// Append the expression of the current arg
			cur_expr = expr_append(comma_expr, cur_arg->parts);
			while (cur_expr->next != NULL) {
				cur_expr = cur_expr->next;
			}
		}
	}
	return first_expr;
}

struct largs 	*args_append(struct largs *list, struct largs *part) {
	struct largs *cur_arg;
	
	if ( list == NULL ) {
		return part;
	}
	for ( cur_arg = list; cur_arg->next != NULL; cur_arg = cur_arg->next ) {
		;
	}
	cur_arg->next = part;
	return list;
}

struct lrangepart *num_range_append(struct lrangepart *list, struct lrangepart *part) {
	struct lrangepart *cur_part;
	
	if (list == NULL) {
		return part;
	}
	for (cur_part = list; cur_part->next != NULL; cur_part = cur_part->next) {
		;
	}
	cur_part->next = part;
	return list;
}

struct lrange	*lrange_append(struct lrange *list, struct lrange *part) {
	struct lrange *cur_lrange;
	
	if (list == NULL) {
		return part;
	}
	for (cur_lrange = list; cur_lrange->next != NULL; cur_lrange = cur_lrange->next) {
		;
	}
	cur_lrange->next = part;
	return list;
}	

struct largrange	*largrange_append(struct largrange *list, struct largrange *part) {
	struct largrange *cur_largrange;
	
	if (list == NULL) {
		return part;
	}
	for (cur_largrange = list; cur_largrange->next != NULL; cur_largrange = cur_largrange->next) {
		;
	}
	cur_largrange->next = part;
	return list;
}	

void print_ident(struct ident *id, char printStruct) {
	if (printStruct && id->name[0] != '_') {
		if ( id->str) {
			printf("%s", id->str);
		} else {
			print_str(id);
		}
	}
	printf("%s", id->name);
	if(id->array_index) {
		printf("[");
		print_expr(id->array_index, printStruct);
		printf("]");
	}
	if(id->str_part) {
		printf("%s", id->str_ind);
		print_ident(id->str_part, 0);
	}
}
void print_str(struct ident *id) {
	IdEntry *tmpId;
	
	// If we can find the identifier in the list of global identifiers,
	// we will print it with "now."
	for (tmpId = glIds; tmpId != NULL; tmpId = tmpId->next) {
		if ( (strcmp(id->name, tmpId->label)) == 0 ) {
			printf("now.");
			return;
		}
	}
	// Otherwise, we print it with the current proctype (P..->)
	printf("P%s->", last_proctype);	
}
void print_expr(struct exprpart *expression, char printStruct) {
	struct exprpart *cur_expr;
	
	for(cur_expr = expression; cur_expr != NULL; cur_expr = cur_expr->next) {
		if(cur_expr->id) {
			print_ident(cur_expr->id, printStruct);
		} else if (cur_expr->expr) {
			printf("%s", cur_expr->expr);
		}
	}
}
	
void print_cmpe_cond(struct cmp_entry *cmpe, char c_prefix) {
	int i;

	if ( c_prefix ) {
		printf("c_expr { (");
	}
	for (i = cmpe->parentheses; i > 0; i--) {
		printf("(");
	}

	if ( cmpe->cmp ) {
		print_cmp_cond(cmpe->cmp);
	} else {
		print_cmpe_cond(cmpe->lhs, 0);
		printf(" %s ", cmpe->or_op ? "||" : "&&");
		print_cmpe_cond(cmpe->rhs, 0);
	}

	for (i = cmpe->parentheses; i > 0; i--) {
		printf(")");
	}
	
	if ( c_prefix ) {
		printf(") }");
	}
}

void print_cmpe_eff(struct cmp_entry *cmpe, char c_prefix) {
	if ( c_prefix ) {
		printf("c_code {");
	}

	if ( cmpe->cmp ) {
		print_cmp_eff(cmpe->cmp);
	} else {
		print_cmpe_or_eff(cmpe);
	}
	
	if ( c_prefix ) {
		printf("}");
	}
}

void print_cmpe_or_eff(struct cmp_entry *cmpe) {
	if ( ! cmpe->or_op ) {
		print_cmpe_and_eff(cmpe);
	} else {
		printf("if (");
		print_cmpe_cond(cmpe->lhs, 0);
		printf(") {\n");
		print_cmpe_eff(cmpe->lhs, 0);
		printf("\n} else if (");
		print_cmpe_cond(cmpe->rhs, 0);
		printf(") {\n");
		print_cmpe_eff(cmpe->rhs, 0);
		printf("\n}");
	}
}

void print_cmpe_and_eff(struct cmp_entry *cmpe) {
	print_cmpe_eff(cmpe->lhs, 0);
	printf("\n");
	print_cmpe_eff(cmpe->rhs, 0);
	printf("\n");	
}

void print_cmpe(struct cmp_entry *cmpe) {
	int i;
	
	for (i = cmpe->parentheses; i > 0; i--) {
		printf("(");
	}
	
	if ( cmpe->cmp ) {
		print_cmp(cmpe->cmp, 0);
	} else {
		print_cmpe(cmpe->lhs);
		printf(" %s ", cmpe->or_op ? "||" : "&&");
		print_cmpe(cmpe->rhs);
	}
	
	for (i = cmpe->parentheses; i > 0; i--) {
		printf(")");
	}
	
}

void print_cmp(struct compare *cmp, char printStruct) {
	print_ident(cmp->id, printStruct);
	printf(" %s ", cmp->op);
	print_expr(cmp->expr, printStruct);
}
void print_cmp_cond(struct compare *cmp) {

	printf("%s(", dhas_op(cmp->op));
	print_ident(cmp->id, 1);
	printf(", (");
	print_expr(cmp->expr, 1);
	printf("))");
}
void print_cmp_eff(struct compare *cmp) {
	
	print_ident(cmp->id, 1);
	printf(" = %s(", domain_op(cmp->op));
	print_ident(cmp->id, 1);
	printf(", (");
	print_expr(cmp->expr, 1);
	printf("));");
}

IdEntry *findId(char *name) {
	IdEntry *tmpId = NULL;

  	for ( tmpId = glIds; tmpId != NULL; tmpId = tmpId->next ) {
  		if ( (strcmp(name, tmpId->label)) == 0 ) {
  			break;
  		}
  	}
  	if (tmpId == NULL) {
	  	for ( tmpId = localIds; tmpId != NULL; tmpId = tmpId->next ) {
	  		if ( (strcmp(name, tmpId->label)) == 0 ) {
	  			break;
	  		}
	  	}
  	}
	
	return tmpId;
}

int isRangeBetween(struct lrangepart *lrp, int lb, int ub) {
	long rangeVal;
	
	if (lrp == NULL) {
		return 0;
	}
	
	rangeVal = strtol(lrp->lb, NULL, 0);
	if ( rangeVal < lb ) {
		return 0;
	}
	
	// find last rangepart
	for ( ; lrp->next != NULL; lrp = lrp->next ) {
		;
	}
	
	rangeVal = strtol(lrp->ub, NULL, 0);
	if ( rangeVal > ub ) {
		return 0;
	}

	return 1;
}

char * mtypenr(char *s)
{
	struct mtypelist *ol;
	

	if (mtypes == NULL) {
		if (0) fprintf(stderr, "ismtype(%s) 0\n", s);
		return NULL;
	}
		
	for ( ol = mtypes; ol != NULL; ol = ol->next ) {
		  if (strcmp(ol->label, s) == 0) {
			if (0) fprintf(stderr, "ismtype(%s) %s\n", s, ol->intrepr);
		  	return ol->intrepr;
		  }
	}
	
	if (0) fprintf(stderr, "ismtype(%s) 0\n", s);
	return NULL;
}

struct stringlist *findchantypes(char *s)
{
	struct chandecllist *l;
	

	if (channels == NULL)
		return NULL;
		
	for ( l = channels; l != NULL; l = l->next ) {
		  if (strcmp(l->chan, s) == 0){
			if(0) fprintf(stderr,"findchantypes(%s): trying: (%s) : 1\n", s, l->chan);
		  	return l->types;
		  } else {
			if(0) fprintf(stderr,"findchantypes(%s): trying: (%s) : 0\n", s, l->chan);
		  }
	}
	if(0) fprintf(stderr,"findchantypes(%s): nada\n", s);
	
	return NULL;
}

/* End of file */
