/* SPEX tokens by Jeroen v. Yperen (JvY)
 * Version 1.0 (21-09-06)
 *
 */

%{

#include <stdio.h>
#include <string.h>
#include "spexgram.h"

int lineno = 1;

%}

%%
spexio			return SPEX_IO;
spexchan			return SPEX_CHAN;
spex_output		return SPEX_OUTPUT;
spex_input		return SPEX_INPUT;
in			yylval.str = yytext; return IN;
DOMAIN|domain		yylval.str = yytext; return DOMAIN;
SPEX|spex		yylval.str = yytext; return SPEX;
now\.			yylval.str = strdup(yytext); return NOW_STRUCT;
P[a-zA-Z0-9][a-zA-Z0-9_]*->		yylval.str = strdup(yytext); return PROC_STRUCT;
\"[^"[:cntrl:]]*(\\[[:print:]][^"[:cntrl:]]*)*\"	yylval.str = strdup(yytext); return STRING;
>|>=|==|!=|<=|<		yylval.str = strdup(yytext); return OPERAND;
\|\|			yylval.str = strdup(yytext); return LOGIC_OR;
\&\&			yylval.str = strdup(yytext); return LOGIC_AND;
(\+|\-)?0x[0-9a-fA-F]*([UuLl])?		yylval.str = strdup(yytext); return NUMBER; /* hexadecimal NUMBER */
(\+|\-)?0[0-7]*([UuLl])?		yylval.str = strdup(yytext); return NUMBER; /* octal NUMBER (or 0)*/
(\+|\-)?[1-9][0-9]*([UuLl])?		yylval.str = strdup(yytext); return NUMBER; /* decimal number (non-zero) */
bit|bool|byte|short|int		yylval.str = strdup(yytext); return TYPE;
mtype			yylval.str = strdup(yytext); return MTYPE;
chan			yylval.str = strdup(yytext); return CHAN;
proctype		yylval.str = "proctype"; return PROCTYPE;
of			yylval.str = strdup(yytext); return OF;
[[:alnum:]_]+		yylval.str = strdup(yytext); return OTHER;		/* OTHER */
[][(){},;=#]		yylval.str = yytext; return *yytext;
\n			lineno++; ECHO;	/* ignore newlines */
[[:space:]]		ECHO; /* ignore whitespace */;
\/\*		{	/* Eat multi-line comment, but still count the newlines. */
			register int c;
			
			while (1) {
				c = input();
				while( c != '*' && c != EOF) {
					putchar(c);
					if ( c == '\n') {
						lineno++;
					}
					c = input();
				}
				while ( c == '*' ) {
					putchar(c);
					c = input();
				}
				if ( c == '/' ) {
					putchar(c);
					break;
				}
				if ( c == EOF ) {
					fprintf(stderr, "Found EOF in multi-line comment\n");
					exit(-1);
				}
			}
		}
\.\.			yylval.str = yytext; return DOTS;
\.			yylval.str = strdup(yytext); return STRUCT_MEMB;		/* literal '.' */
&			yylval.str = strdup(yytext); return AMPERSAND;		/* literal '&' */
->			yylval.str = strdup(yytext); return ARROW;
.			yylval.str = strdup(yytext); return SINGLE_OTHER;	/* SINGLE_OTHER */
%%


// \/\/[^\n]*\n			lineno++; ECHO; /* ignore c-style comments */
