# =========================================================================
# $Id: config.tcl,v 1.1.2.10 2004/10/05 18:35:26 belinfan Exp $
# Author: Axel Belinfante
# =========================================================================

# =========================================================================

# utility routines

proc process_config_line_words { file line infolist initialresult args } {
	set result $initialresult
	# Message "process_config_line_words processing ($file)($file)($infolist)($initialresult)($args)\n"
	if {[llength $args] < 1} {
		Message "Warning: Internal error: process_config_line_words: argument missing\n"
		return $result
	}
	set configkeys [lindex $args 0]
	foreach e $infolist {
		set c [string toupper [lindex $e 0]]
		if {[string compare INCLUDE $c] == 0} {
			foreach f [lrange $e 1 end] {
				set result [read_config_file $f process_config_line_words $result $configkeys]
			}
		} else {
			foreach key $configkeys {
				# Message "process_config_line_words processing ($key): $c\n"
				if {[string compare $key $c] == 0} {
					lappend result [list $file $line [join [lrange $e 1 end] " "]]
					# Message "process_config_line_words found $key $file $line\n"
				}
			}
		}
	}
	return $result
}

proc read_config_file { file routine initialresult args} {
	global read_config_file_warned

	set result $initialresult
	if [catch {open $file r} fileId] {
		if {![info exists read_config_file_warned($file)] ||
		     [string compare $read_config_file_warned($file) $fileId] != 0} {
			set read_config_file_warned($file) $fileId
			Message "Warning: Cannot open configuration file $file: $fileId\n"
		}
	} else {
		set linenr 0
		foreach line [string trim [split [read $fileId] \n]] {
			incr linenr
			if {(![regexp {^#} $line match]) && (![string compare $line ""] == 0)} {
				set command [list $routine $file $linenr [list $line] $result]
				foreach a $args {
					lappend command $a
				}
				set result [eval $command]
			}
		}
		close $fileId
	}
	return $result
}

# =========================================================================

# if key is found, returns list of three-tuple (list):  configfile linenr value
# otherwise, returns empty list.
proc find_flv_list_in_configfiles {configkeys configfiles} {
	set result {}
	# Message "find_in_configfiles $configkeys $configfiles\n"
	foreach file $configfiles {
		set result [read_config_file $file process_config_line_words $result $configkeys]
	}
	# puts stderr "find_flv_in_configfiles($configkeys)($configfiles): $result"
	return $result
}

# if key is found, returns three-tuple (list):  configfile linenr value
# otherwise, returns empty list.
proc find_flv_in_configfiles {configkeys configfiles} {
	set result [lindex [find_flv_list_in_configfiles $configkeys $configfiles] 0]
	return $result
}

# NOTE: the routine below is used in the *.primers and *.mutants files!!
# if key is found, returns value
# otherwise, returns empty string.

proc find_in_configfiles {configkeys configfiles} {
	set result ""
	set file_line_string [find_flv_in_configfiles $configkeys $configfiles]
	if {[llength $file_line_string] > 0} {
		set result [lindex $file_line_string 2]
	}
	return $result
}

# =========================================================================
