# =========================================================================
#
# $Id: mutant.tcl,v 1.7.2.5 2006/01/24 21:42:36 belinfan Exp $
#
# Author: Axel Belinfante
#
# This file is _VERY_ simular to the primer.tcl file!
# If you change something here, changes are VERY high that you have to
# change it in primer.tcl as well!
#
# This file implements the MUTANT command that we use in mutant
# config files.
# The idea is to interpret mutant files as tcl files containing
# only invocations of the MUTANT command, to register mutants.
# As result of such registration, the GUI will add the registered
# items to the mutants menu.
# Question: will the loading of a second mutant file override the first one?
# Or can we have items of multiple mutant files 'active' at the same time?
# If so, how will this be visible in the menu?
# (as a sub-menu under the Mutants menu?)
# How will we be able to 'close' different mutant files?
# (via a sub-menu under the Close Mutants menu entry?)
#
# =========================================================================

# This routine will be invoked from xtorx to force auto-loading of this file.
proc do_first_initMutants {args} {
	do_first_initExtension $args
}

proc do_initMutants {args} {
	global tester
	global debug

	# set debug(Msg,do_extension_callback) 1
	# set debug(Msg,do_cleanupExtension) 1

	do_initExtension \
		mutants \
		MUTANT \
		"*mutants" \
		"Open Mutants File" \
		tester(mutants,titlename) \
		tester(basenamebare,mutants_file) \
		tester(torx,cmdline,mutants,flags) \
		tester(torx,cmdline,mutants,config) \
		"mutants_present *mutants_mb,*mutants,*close_mutants_button,*reopen_mutants_button" \
		"mutants_present *open_mutants_button" \
		{
			{{Mutants Files}    {.mutants}    }
			{{Primers Files}    {.primers}    }
			{{Guides Files}    {.guides}    }
			{{Trace Files}      {.trace}      }
			{{Log Files}        {.log}        }
			{{Config Files}     {.if}         }
			{{Text Files}       {.txt}        }
			{{TCL Scripts}      {.tcl}        }
			{{All Files}        *             }
		} \
		$args
}

proc do_cleanupMutants {args} {
	do_cleanupExtension mutants $args
}

proc do_open_mutants {args} {
	do_open_extension mutants $args
}

proc do_do_open_mutants { filename} {
	do_do_open_extension mutants $filename
}

proc do_close_mutants {args} {
	do_close_extension mutants $args
}

proc do_do_close_mutants {} {
	do_do_close_extension mutants
}

proc do_reopen_mutants {args} {
	do_reopen_extension mutants $args
}

proc do_do_select_mutant {titlename} {
	do_do_select_extension mutants $titlename
}

proc MUTANT {entry title create destruct args} {
	EXTENSION mutants $entry $title $create $destruct $args
}
