# =========================================================================
#
# main loop end-of-test routine
#

proc main_go_on {} {
	global maininfo
	if { $maininfo(stopnow) } {
		return 0
	}
	if { ($maininfo(nrofstepsmax) > 0) &&
          ($maininfo(nrofstepsdone) >= $maininfo(nrofstepsmax)) } {
        if {[string compare $maininfo(verdict) inconclusive] == 0} {
            set maininfo(verdict) pass
        }
		return 0
	}
	return 1
}

# =========================================================================
#
# primitives used in the main loop
#

proc spec_test_eot {id} {
	global sidinfo maininfo specinfo
	send -i $sidinfo(spec,$id) "C_EOT\r"
	set result ""
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		expect {
			-i $sidinfo(spec,$id) -re "^A_EOT_YES\r*\n" {
				set result [list yes]
			}
			-i $sidinfo(spec,$id) -re "^A_EOT_NO\r*\n" {
				set result [list no]
			}
			timeout {
				set result [list error]
			}
		}
	}
	set timeout $maininfo(timeout)
	return $result
}

proc spec_get_iokind {id sug} {
	global sidinfo maininfo specinfo
    set iokind_map("") ""
    set iokind_map(input) iokind=input
    set iokind_map(output) iokind=output
	if {[info exists iokind_map($sug)]} {
		set iosug $iokind_map($sug)
	} else {
		# error message XXX
		set iosug ""
	}
	send -i $sidinfo(spec,$id) "C_IOKIND $iosug\r"
	set result ""
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		expect {
			-i $sidinfo(spec,$id) -re "^A_IOKIND\[ \t]*(\[^\r\n]*)\r*\n" {
				set fields [fields_split $expect_out(1,string)]
				set iokind_str [field_value $fields iokind]
				set result [list $iokind_str]
			}
			timeout {
				set result [list error]
			}
		}
	}
	set timeout $maininfo(timeout)
	return $result
}

proc spec_do_input {id sug} {
	return [spec_getdo_input $id $sug INPUT]
}
proc spec_get_input {id sug} {
	return [spec_getdo_input $id $sug GETINPUT]
}
proc spec_getdo_input {id sug keyword} {
	global sidinfo maininfo specinfo
	set pr_ev [event_table_lookup_mapping "" $id $sug]
	if {$maininfo(debug) > 2} {
		send_error "DEBUG: C_[set keyword] event=$pr_ev\n"
	}
	send -i $sidinfo(spec,$id) "C_[set keyword] event=$pr_ev\r"
	set result ""
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		catch { unset expect_out(2,string) }
		expect {
			-i $sidinfo(spec,$id) -re "^A_[set keyword]_OK\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
				set fields [fields_split $expect_out(1,string)]
				set event_str [field_value $fields event]
				set event $event_str
				set result [list input_ok $event]
			}
			-i $sidinfo(spec,$id) -re "^A_[set keyword]_ERROR(\[\t ]\[\t ]*(\[^\r\n]*))?\r*\n" {
				if {[info exists expect_out(2,string)]} {
					set result [list input_error $expect_out(2,string)]
				} else {
					set result [list input_error ]
				}
			}
			timeout {
				set result [list error ]
			}
		}
	}
	if {$maininfo(debug) > 2} {
		send_error "DEBUG: C_[set keyword] event=$pr_ev => $result\n"
	}
	set timeout $maininfo(timeout)
	return $result
}

#
# spec_choose_from_menu
# This routine takes a suggestion as argument.
# If the suggestion is the empty event, and $maininfo(chooseInputs) is set,
# it asks the specification module for the list of input events, and
# chooses one element from this list (using a random number generator).
# We will then invoke spec_get_input with this chosen input event.
#
# So, when $maininfo(chooseInputs) is set, it is _not_ the spec module
# who will randomly choose an event, but this routine will do this.
#
# Why do we want this behaviour?
# We want this, because, for promela, we want our instantiator module
# to be able to instantiate input events. The instantiator can only
# do that if the C_INPUT command sent to the spec module contains the
# complete suggested event (taken from the input menu).
#
proc spec_choose_from_menu {sug} {
	global sidinfo maininfo specinfo menuinfo
	if {$maininfo(debug) > 1} {
		send_error "DEBUG: spec_choose_from_menu ($sug)\n"
	}
	if {[event_in_menu $sug $menuinfo(inputs)]} {
		set result [list input $sug]
	} else {
		set result ""
		if {$maininfo(debug) > 1} {
			send_error "DEBUG: spec_choose_from_menu inputs ($menuinfo(inputs))\n"
		}
	}
	if {([llength $result] == 0) && ($maininfo(chooseInputs))} {
		set inputs $menuinfo(inputs)
		if {$maininfo(debug) > 2} {
			send_error "DEBUG: spec_choose_from_menu inputs ($inputs)\n"
		}
		set nr_of_el [llength $inputs]
		set idx [RandomRange $nr_of_el]
		set result [list input [lindex $inputs $idx]]
		if {$maininfo(debug) > 2} {
			send_error "DEBUG: spec_choose_from_menu (nr_of_el: $nr_of_el)(idx: $idx) result ($result)\n"
		}
	}
	if {$maininfo(debug) > 2} {
		send_error "DEBUG: spec_choose_from_menu result ($result)\n"
	}
	return $result
}

# kind should be INPUT or OUTPUT
proc spec_get_events {id kind} {
	global sidinfo maininfo specinfo
	set result ""
	set events {}
	switch -- $kind {
		INPUT { set type inputs }
		OUTPUT { set type outputs }
		default { return $result }
	}
	send -i $sidinfo(spec,$id) "C_[set kind]S\r"
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		expect {
			-i $sidinfo(spec,$id)  -re "^A_[set kind]S_START\r*\n" {
				while {[string length $result] == 0} {
					expect {
						-i $sidinfo(spec,$id) -re "^A_[set kind]S_END\r*\n" {
							set result [list $type $events ]
						}
						-i $sidinfo(spec,$id) -re "^A_EVENT\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
							set fields [fields_split $expect_out(1,string)]
							set event_str [field_value $fields event]
							set event $event_str
							set susp_str [field_value $fields suspension]
							set susp $susp_str
							set ev [list $event $susp]
							set attr {}
							foreach f $fields {
								set name_val [field_first $f]
								set a_name [lindex $name_val 0]
								set a_val [lindex $name_val 1]
								switch -- $a_name {
									event {}
									suspension {}
									channel {}
									default {
										if {[string compare $a_val ""] != 0} {
											lappend attr [list $a_name $a_val]
										}
									}
								}
							}
							# foreach a_name [list partition weight] {
							# 	set a_val [field_value $fields $a_name]
							#	if {[string compare $a_val ""] != 0} {
							#		lappend attr [list $a_name $a_val]
							#	}
							# }
							lappend ev $attr
							# send_error "DEBUG: inserting $id $kind ($ev)\n"
							lappend events $ev
						}
						timeout {
							set result [list error ]
						}
					}
				}
			}
			timeout {
				set result [list error ]
			}
		}
	}
	set timeout $maininfo(timeout)
	return $result
}

proc spec_get_outputs {id} {
	return [spec_get_events $id OUTPUT]
}

proc spec_get_inputs {id} {
	return [spec_get_events $id INPUT]
}

proc spec_get_state {id} {
	global sidinfo maininfo specinfo
	send -i $sidinfo(spec,$id) "C_STATE\r"
	set result ""
	set state ""
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		expect {
			-i $sidinfo(spec,$id) -re "^A_STATE_START\r*\n" {
				while {[string length $result] == 0} {
					expect {
						-i $sidinfo(spec,$id) -re "^A_STATE_END\r*\n" {
							set result [list state $state ]
						}
						-i $sidinfo(spec,$id) -re "^(\[^\r\n]*)\r*\n" {
				    		set l $expect_out(1,string)
							lappend state $l
						}
						timeout {
							set result [list error ]
						}
					}
				}
			}
			timeout {
				set result [list error ]
			}
		}
	}
	set timeout $maininfo(timeout)
	return $result
}

proc spec_get_stats {id} {
	global sidinfo maininfo specinfo
	send -i $sidinfo(spec,$id) "C_STATS\r"
	set result ""
	set stats ""
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		expect {
			-i $sidinfo(spec,$id) -re "^A_STATS\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
				set result [list stats [list $expect_out(1,string) ]]
			}
			-i $sidinfo(spec,$id) -re "^A_STATS_START\r*\n" {
				while {[string length $result] == 0} {
					expect {
						-i $sidinfo(spec,$id) -re "^A_STATS_END\r*\n" {
							set result [list stats $stats ]
						}
						-i $sidinfo(spec,$id) -re "^A_STATS\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
				    		set l $expect_out(1,string)
							lappend stats $l
						}
						timeout {
							set result [list error ]
						}
					}
				}
			}
			timeout {
				set result [list error ]
			}
		}
	}
	set timeout $maininfo(timeout)
	return $result
}

proc spec_get_stateid {id} {
	global sidinfo maininfo specinfo
	send -i $sidinfo(spec,$id) "C_STATEID\r"
	set result ""
	set stats ""
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		expect {
			-i $sidinfo(spec,$id) -re "^A_STATEID\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
				set result [list stateid [list $expect_out(1,string) ]]
			}
			-i $sidinfo(spec,$id) -re "^A_STATEID_START\r*\n" {
				while {[string length $result] == 0} {
					expect {
						-i $sidinfo(spec,$id) -re "^A_STATEID_END\r*\n" {
							set result [list stateid $stats ]
						}
						-i $sidinfo(spec,$id) -re "^A_STATEID\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
				    		set l $expect_out(1,string)
							lappend stats $l
						}
						timeout {
							set result [list error ]
						}
					}
				}
			}
			timeout {
				set result [list error ]
			}
		}
	}
	set timeout $maininfo(timeout)
	return $result
}

proc spec_goto {super id role} {
	global sidinfo maininfo specinfo
	set result ""
	send -i $sidinfo(spec,$id) "C_GOTO super=$super\r"
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		expect {
			-i $sidinfo(spec,$id) -re "^A_GOTO_OK\[\t ]*(\[^\r\n]*)\r*\n" {
				set result [list goto ]
			}
			-i $sidinfo(spec,$id) -re "^A_GOTO_ERROR\[\t ]*(\[^\r\n]*)\r*\n" {
				set result [list error ]
			}
			timeout {
				set result [list error ]
			}
		}
	}
	set timeout $maininfo(timeout)
	return $result
}

proc spec_check_output {id ev susp chan} {
	global sidinfo maininfo specinfo
	set pr_ev $ev
	if {$maininfo(debug) > 1} {
		send_error "DEBUG: $id C_OUTPUT event=$pr_ev\tsuspension=$susp\tchannel=$chan\n"
	}
	send -i $sidinfo(spec,$id) "C_OUTPUT event=$pr_ev\tsuspension=$susp\tchannel=$chan\r"
	set result ""
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		expect {
			-i $sidinfo(spec,$id) -re "^A_OUTPUT_OK\r*\n" {
				set result [list output_ok]
			}
			-i $sidinfo(spec,$id) -re "^A_OUTPUT_OK\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
				set event_str [field_value [fields_split $expect_out(1,string)] event]
				set result [list output_ok $event_str]
			}
			-i $sidinfo(spec,$id) -re "^A_OUTPUT_ERROR(\[\t ]\[\t ]*\[^\r\n]*)?\r*\n" {
				set result [list output_error]
			}
			timeout {
				set result [list error]
			}
		}
	}
	set timeout $maininfo(timeout)
	return $result
}

proc spec_quit {id} {
	global sidinfo maininfo specinfo
	if {$maininfo(debug) > 2} {
		send_error "DEBUG $id: C_QUIT\n"
	}
	set specinfo(quit,$id) 1
	send -i $sidinfo(spec,$id) "C_QUIT\r"
	set result ""
	set timeout $specinfo(timeout)
	while {[string length $result] == 0} {
		expect {
			-i $sidinfo(spec,$id) -re "^A_QUIT\r*\n" {
				set result [list quit]
				exp_continue ;# NEEDED to match eof via expect_after
			}
			timeout {
				set result [list error]
			}
		}
	}
	set timeout $maininfo(timeout)
	return $result
}

proc we_stop {} {
	clean_exit
}

# =========================================================================

proc send_menu_verdict {} {
	global menuinfo
	if {[string compare $menuinfo(verdict) ""] != 0} {
		send_user "A_WARNING guidance verdict: $menuinfo(verdict)\n"
		send_user "A_GUIDANCE_VERDICT $menuinfo(verdict)\n"
	} elseif {$menuinfo(inconsistent)} {
		send_user "A_WARNING intersection menu is empty, so guidance verdict: miss\n"
#		send_user "A_GUIDANCE_VERDICT INCONSISTENT\n"
		send_user "A_GUIDANCE_VERDICT MISS\n"
	} elseif {$menuinfo(endoftest)} {
		send_user "A_WARNING menu is empty, so guidance verdict: pass\n"
		send_user "A_GUIDANCE_VERDICT PASS\n"
	}
}

# =========================================================================

# Random number generator taken from:
#
# Brent Welch, "Practical Programming in Tcl/Tk", 2nd edition, pages 78,79
#   (adapted from Don Libes, "Exploring Expect", 1995 and from
#                 Press et al. "Numerical Recipes in C", 1998
#   )
#
proc RandomInit {seed} {
	global randomSeed
	set randomSeed $seed
}

proc Random {} {
	global randomSeed
	set randomSeed [expr ($randomSeed*9301 + 49297) % 233280]
	return [expr $randomSeed/double(233280)]
}

proc RandomRange {range} {
	return [expr int ([Random]*$range)]
}

# =========================================================================
