# =========================================================================
# $Id: explorer.tcl,v 1.1.2.14 2008/07/07 21:27:19 belinfan Exp $
# =========================================================================
# interface to explorer

proc explorer_init {} {
	global sidinfo
	global explorer
	global timeout
	set timeout -1
	log_user 0
	match_max -d 10000

	if {[catch {spawn -pty} foo]} {
		fatal "spawn for stderr of $program fails: $foo"
	}
	set sidinfo(explorer,stderr) $spawn_id
	set specinfo(pty,stderr) $spawn_out(slave,name)
	send_error "A_DEBUG EXPLORER Stderr match_max=[match_max]\n"

	set cdcmd "cd $explorer(rundir)"
	set explorer(cmd) [list spawn -noecho "/bin/sh" -c "exec 2> $specinfo(pty,stderr); $cdcmd; exec $explorer(program) " ]
#	set explorer(cmd) [list spawn -noecho "/bin/sh" -c "stty -echo; exec 2> $specinfo(pty,stderr); $cdcmd; exec $explorer(program) " ]
#	set explorer(cmd) [list spawn -noecho "/bin/sh" -c "stty -echo; $cdcmd; exec $explorer(program) " ]
	send_error "Starting explorer: $explorer(cmd)\n"
	if {[catch $explorer(cmd) foo]} {
		send_error "spawn for $explorer(cmd) fails: $foo\n"
		do_cleanup
		exit 1
	}
	set sidinfo(explorer) $spawn_id
	set explorer(pid) [exp_pid -i $sidinfo(explorer)]
	send_error "A_DEBUG EXPLORER match_max=[match_max]\n"
	expect_background {
		-i $sidinfo(explorer,stderr) full_buffer {
			send_error "A_WARNING EXPLORER Stderr Buffer Overflow!!!\n"
			set txt $expect_out(buffer)
			send_error "A_WARNING EXPLORER Stderr Buffer Contents Skipped: $txt\n"
			send_user "A_ERROR EXPLORER Stderr Buffer Contents Skipped: $txt\n"
			exit
		}
		-i $sidinfo(explorer,stderr) -re "^A_DEBUG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			send_error "EXPLORER Debug: $txt\n"
		}
		-i $sidinfo(explorer,stderr) -re "(\[^\r\n]*)\r*\n" {
			set l $expect_out(1,string)
			send_error "EXPLORER Stderr: $l\n"
		}
	}
	expect_before {
		-i $sidinfo(explorer) full_buffer {
			send_error "A_WARNING EXPLORER Buffer Overflow!!!\n"
			set txt $expect_out(buffer)
			send_error "A_WARNING EXPLORER Buffer Contents Skipped: $txt\n"
			send_user "A_ERROR EXPLORER Buffer Contents Skipped: $txt\n"
			exit
		}
		-i $sidinfo(explorer) -re "^(A_DEBUG\[^\r\n]*)\r*\n" {
			debugMsg "Explorer: $expect_out(1,string)"
			exp_continue
		}
		-i $sidinfo(explorer) -re "^(\[a-z]\[^\r\n]*)\r*\n" {
			# echo-ed back command
			debugMsg "ExplorerCmd: $expect_out(1,string)"
			exp_continue
		}
	}
	eval "
	expect_after {
		-i $sidinfo(explorer) eof {
			global sidinfo explorer
			if {\[info exists sidinfo(explorer)]} {
				unset sidinfo(explorer)
			}
			if {\[info exists explorer(pid)]} {
				unset explorer(pid)
			}
			if {!\[info exists explorer(quit)]} {
				fatal \"End-of-file from explorer\\n\"
				set explorer(quit) 1
			}
		}
		-i $sidinfo(explorer)  -re \"^(\\\[^\\r\\n]*)\\r*\\n\" {
			send_error \"A_WARNING EXPLORER unexpected Stdout: \$expect_out(1,string)\\n\"
			exp_continue
		}
	}
	"
}

# return tuple of
#  <nr, freevars>
# or
#  <nr, preds, freevars>
proc explorer_start {} {
	global sidinfo
	set initial {}
	exp_send -i $sidinfo(explorer) "r\r"
	expect {
		-i $sidinfo(explorer) -re "^R0(\[^\r\n]*)\r*\n" {
			# error
		}
		-i $sidinfo(explorer) -re "^R\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
			set s $expect_out(1,string)
			debugMsg "start-string: $s"
			set initial [split $s \t]
		}
	}
	debugMsg "explorer_start: $initial"
	return $initial
}

# return list of tuples
#  <nr, visible, preds_solved, label, preds, freevars, identicalnode,time>
proc explorer_unfold {ev} {
	global sidinfo
	set menu {}
	exp_send -i $sidinfo(explorer) "e $ev\r"
	expect {
		-i $sidinfo(explorer) -re "^E0(\[^\r\n]*)\r*\n" {
			# error
		}
		-i $sidinfo(explorer) -re "^EB(\[^\r\n]*)\r*\n" {
			expect {
				-i $sidinfo(explorer) -re "^Ee\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
					# event
					set s $expect_out(1,string)
					set menu_el [split $s \t]
					lappend menu $menu_el
					exp_continue
				}
				-i $sidinfo(explorer) -re "^EE(\[^\r\n]*)\r*\n" {
					# end of list
				}
			}
		}
	}
	debugMsg "unfold ($ev): $menu"
	return $menu
}

# instantiate event or initial-state
# return either: [list 1 newstate]
# or:            [list 0]
proc explorer_instantiate {ev preds} {
	global sidinfo
	exp_send -i $sidinfo(explorer) "i $ev $preds\r"
	expect {
		-i $sidinfo(explorer) -re "^I0(\[^\r\n]*)\r*\n" {
			# error
			set res [list 0 $expect_out(1,string)]
		}
		-i $sidinfo(explorer) -re "^I\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
			set s $expect_out(1,string)
			set inst_ev [split $s \t]
			set res [list 1 $inst_ev]
		}
	}
	debugMsg "explorer_instantiate($ev)($preds): $res"
	return $res
}


# delete events/states
proc explorer_delete {evs} {
}

# match event with list of events
proc explorer_match_events {ev evs} {
	global sidinfo
	set identical 0
	if {[llength $evs] > 0} {
		set evs_str [join $evs " "]
		exp_send -i $sidinfo(explorer) "m $ev $evs_str\r"
		expect {
			-i $sidinfo(explorer) -re "^M0(\[^\r\n]*)\r*\n" {
				# error
			}
			-i $sidinfo(explorer) -re "^M\[\t ]\[\t ]*(\[^\r\n]*)\r*\n" {
				set identical [split $expect_out(1,string)]
			}
		}
	}
	debugMsg "explorer_match_events ($ev)($evs): $identical"
	return $identical
}

# =========================================================================
