# =========================================================================
# $Id: main_loop.tcl,v 1.1.2.23 2008/07/09 00:44:01 belinfan Exp $
# =========================================================================

proc primer_interface {seed file exp rundir algo ioct complete} {
	infoMsg "SEED $seed"
	infoMsg "CONFIGFILE $file"
	infoMsg "ALGORITHM $algo"
	infoMsg "IOCT $ioct"
	infoMsg "COMPLETE $complete"
	infoMsg "EXPLORER $exp"
	infoMsg "RUNDIR $rundir"
	if {[catch {do_primer_interface $seed $file $exp $rundir $algo $ioct $complete} foo]} {
		bgerror $foo
	}
}
proc do_primer_interface {seed file exp rundir algo ioct complete} {
	global explorer
	set explorer(program) $exp
	set explorer(rundir) $rundir
	primer_init $file $algo $ioct $complete
	RandomInit $seed
	install_pending_events_handler
	disable_selection_exportation
	primer_start ;# compute initial menu
	restore_selection_exportation
	primer_main
	vwait explorer(quit)
}

proc primer_main {} {
	## Currently, we prefere to use blocking IO
	## fconfigure -nonblocking stdin
	fileevent stdin readable primer_try_handle_line	
}

# ====================================

proc disable_selection_exportation {} {
#	global selexp_stack
#	lappend selexp_stack [export_selection 0]
}
proc restore_selection_exportation {} {
#	global selexp_stack
#	set old [lindex $selexp_stack end]
#	set selexp_stack [lreplace $selexp_stack end end]
#	update idletasks ;# make selections before enabling export
#	export_selection $old
}


# ====================================

proc get_arguments {arg} {
	global env
	# we expect the seed in env var SEED.
	# we expect the delta-event representation in env var DELTA.
	# we expect the config file name in env var CONFIGFILE.
	set result ""
	switch -- [string tolower $arg] {
		  li {
		  	if {[info exists env(INGATES)]} {
		  		set result [split [string trim $env(INGATES)] ,]
		  	}
		} lu {
		  	if {[info exists env(OUTGATES)]} {
				set result [split [string trim $env(OUTGATES)] ,]
		  	}
		} seed {
		  	if {[info exists env(SEED)]} {
				set result [string trim $env(SEED)]
		  	}
		} delta {
		  	if {[info exists env(DELTA)]} {
				set result [string trim $env(DELTA)]
		  	}
		} config {
		  	if {[info exists env(CONFIGFILE)]} {
				set result [string trim $env(CONFIGFILE)]
		  	}
		} suspension {
		  	if {[info exists env(SUSPENSION)]} {
				set result [string trim $env(SUSPENSION)]
		  	}
		} algorithm {
		  	if {[info exists env(ALGORITHM)]} {
				set result [string trim $env(ALGORITHM)]
		  	}
		} ioct {
			set result 0
		  	if {[info exists env(IOCT)]} {
				set result [string trim $env(IOCT)]
		  	}
		} complete {
		  	if {[info exists env(COMPLETE)]} {
				set result [string trim $env(COMPLETE)]
		  	}
		} explorer {
		  	if {[info exists env(EXPLORER)]} {
				set result [string trim $env(EXPLORER)]
		  	}
		} rundir {
		  	if {[info exists env(RUNDIR)]} {
				set result [string trim $env(RUNDIR)]
 		  	}
 		}
	}
	return $result
}

# ====================================

# We set bgerror, because we use fileevent to
# handle the input, and fileevent runs the input-handler
# as a background command
proc bgerror {msg} {
	global errorInfo
	warningMsg "Fatal InternalError: $msg"
	set lines [split $errorInfo \n]
	warningMsg "Stack Begin"
	foreach l $lines {
		warningMsg "Stack: $l"
	}
	warningMsg "Stack End"
	# exit
}


# ====================================

proc install_pending_events_handler {} {
#	after idle do_handle_pending_events
}

proc stop_pending_events_handler {} {
	global pending_events_handler_signal
	catch { unset pending_events_handler_signal }
}

proc do_handle_pending_events {} {
	global pending_events_handler_signal
	# debugMsg do_handle_pending_events
	handle_pending_events
	if {[info exists pending_events_handler_signal]} {
		after 1 install_pending_events_handler
	}
}

# ====================================

proc primer_try_handle_line {} {
	fileevent stdin readable ""
	set stop 0
	set skip 0
	set line ""
	if {[gets stdin line] < 0} {
		exit
	}
	if {! $skip} {
		disable_selection_exportation
		set speed [time {primer_handle_line $line}]
		debugMsg "time: $speed"
		restore_selection_exportation
	}
	if {! $stop} {
		fileevent stdin readable primer_try_handle_line
	}
}

# ====================================

proc primer_handle_line {line} {
	global explorer
	debugMsg "Line: ($line)"
	set line [string trim $line]
	if {![string compare $line ""] == 0} {
		set parsed_line [primer_parse $line]
		set parsed_ok [lindex $parsed_line 0]
		debugMsg "Parsed: ($parsed_line)"
		if {$parsed_ok} {
			foreach e [mr_list] {
				debugMsg "beginmenu: $e"
			}
			set parsed_cmd [lindex $parsed_line 1]
			set cmd_keyword [lindex $parsed_cmd 0]
			switch -- [string toupper $cmd_keyword] {
			    C_QUIT {
					send_menu_verdict
			    	set cont 0
			    	set explorer(quit) 1
			    	puts "A_QUIT"
			  } C_CHANNELS {
					send_menu_verdict
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set kind_sug [field_value $fields iokind]
			  		primer_do_channels $kind_sug
			  } C_IOKIND {
					send_menu_verdict
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set chan_sug [field_value $fields channel]
			  		set kind_sug [field_value $fields iokind]
			  		primer_do_iokind $kind_sug $chan_sug
			  } C_GETINPUT {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set ev [field_value $fields event]
			  		set chan [field_value $fields channel]
			  		set s [field_value $fields suspension]
			  		set p [field_value $fields predicates]
			  		primer_get_input $ev $chan $s $p
			  } C_INPUT {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set ev [field_value $fields event]
			  		set chan [field_value $fields channel]
			  		set s [field_value $fields suspension]
			  		set p [field_value $fields predicates]
			  		primer_do_input $ev $chan $s $p
			  } C_GETOUTPUT {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set ev [field_value $fields event]
			  		set chan [field_value $fields channel]
			  		set s [field_value $fields suspension]
			  		set p [field_value $fields predicates]
			  		primer_get_output $ev $chan $s $p
			  } C_OUTPUT {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set ev [field_value $fields event]
			  		set chan [field_value $fields channel]
			  		set s [field_value $fields suspension]
			  		set p [field_value $fields predicates]
			  		primer_do_output $ev $chan $s $p
			  } C_INPUTS {
					send_menu_verdict
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set chan_sug [field_value $fields channel]
			  		set kinds [list input inputsuspension]
			    	primer_menu "A_EVENT" "INPUTS" $kinds $chan_sug
			  } C_OUTPUTS {
					send_menu_verdict
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set chan_sug [field_value $fields channel]
			  		set kinds [list output outputsuspension]
			    	primer_menu "A_EVENT" "OUTPUTS" $kinds $chan_sug
			  } C_STATS {
					send_menu_verdict
			    	puts "A_STATS no stats yet"
			  } C_STATE {
					send_menu_verdict
			    	puts "A_STATE_START"
			    	puts "no stats yet"
			    	puts "A_STATE_END"
			  } C_STATEID {
					send_menu_verdict
					set super [primer_get_super]
			    	puts "A_STATEID super=[lindex $super 0]\tinit=[lindex $super 1]\ttrans=[lindex $super 2]"
			  } C_GOTO {
			  		set fields [fields_split [lindex $parsed_cmd 1]]
			  		set sup [field_value $fields super]
					primer_goto_super_id $sup
			  } default {
					puts "A_ERROR UnknownCommand: $cmd_keyword"
			  }
			}
			debugMsg "start of endmenu"
			foreach e [mr_list] {
				debugMsg "endmenu: $e"
			}
			debugMsg "end of endmenu"
		} else {
			puts "A_ERROR UnknownCommand: $line"
		}
		debugMsg "End of Parsed block: ($parsed_line)"
	}
	debugMsg "HandledLine: ($line)"
}

# =========================================================================
