/* IA-Trojka - Inter Active Trojka analyser software - Version 1.0 - 1996 */

/* Copyright (c) by the author.  All Rights Reserved.			  */
/* No guarantee is expressed or implied by the distribution of this code. */
/* IA-Trojka is written by Rene G. de Vries.

/* The Trojka analyser driver (trojka.o) software is written by		  */
/* Rene G. de Vries based upon the SPIN software by			  */
/* Gerard J. Holzmann, as part of					  */
/* ``Design and Validation of Protocols,'' ISBN 0-13-539925-4,		  */
/* 1991, Prentice Hall, Englewood Cliffs, NJ, 07632			  */
/**************************************************************************/

/* updated Fri Sep  4 12:25:39 MET DST 1998
   showing use of Qid2Name
*/

#include "trojka.h"

Trojka_SettingsT        Settings;	
Trojka_statT		Stats;		/* statistics */

void
ia_usage(FILE *out)
{	fprintf(out, "\nInter-Active Trojka\n\n");
        fprintf(out, "Valid Options are:\n");
        fprintf(out, "   -kF Create logfile <F>\n");
        fprintf(out, "   -sS Create searchfile <F>\n");
        fprintf(out, "   -hN choose other hash-function 1..32\n");
        fprintf(out, "   -lN set seedvalue N of randomgenerator\n");
        fprintf(out, "   -mN set cut depth (default 100)\n");
        fprintf(out, "   -wN hashtable of 2^N entries\n");
        fprintf(out, "       (defaults to -w%d)\n", Settings.ssize);
	fprintf(out, "\n");
        exit(1);
}

void
OwnPrtAction(FILE *out, ActionT Action)
{	int i;
	if (Action.atype == output) fprintf(out,"%s => ",Qid2Name(Action.Queue));
	if (Action.sfield[0] == variable)
			fprintf(out,"( var");
	else
			fprintf(out,"( %d", Action.vfield[0]);
	for (i=1;i<Action.nr_fields;i++) {
		if (Action.sfield[i] == variable) fprintf(out, ", var");
	 	else fprintf(out, ", %d ",Action.vfield[i]);
		}
	fprintf(out, " ) ");
	if (Action.atype == input) fprintf(out, "=> %s", Qid2Name(Action.Queue));
}	



void
PrtStats(FILE *out, Trojka_statT s)
{
	fprintf(out, "\t\t*** Trojka engine statistics ***\n\n");
	fprintf(out, "Number of created statevectors:\t\t\t%8d\n", s.nr_vector);
	fprintf(out, "Number of created statevector structures:\t%8d\n",
		s.nr_svector);
	fprintf(out, "Memory usage (bytes):\t\t\t\t%8g\n\n",s.memcnt);
	fprintf(out, "Number of cut branches:\t\t%8d\n",s.nr_cuts);
	fprintf(out, "Number of analysed states:\t%8d\n",s.nr_analysed);
	fprintf(out, "Number of matched states:\t%8d\n",s.nr_matched);
	fprintf(out, "Maximum depth reached:\t\t%8d\n",s.depth);
	fprintf(out, "\nNumber of superstate elements:\t%8d\n",s.nr_supers);
	fprintf(out, "Number of executed actions:\t%8d\n\n",s.nr_executed);
	fprintf(out, "Number of different actions possible: %8d\n",s.nr_actions);
	fprintf(out, "%sOutput suspension in current superstate\n"
			,s.delta?"":"No ");
}

void
userinterface(void)
{	char	useraction[80];		/* user input */
	int	i, actcode, commandnr = 1;
	ActionT Action;

	/* start of main loop */
	Stats = getstats();	/* we need initial info */
	while (1) {
	printf("\n trojka command [h - help] %d%%  ",commandnr++);
	scanf("%s",&useraction); printf("\n");

	if (useraction[0]<='9' && useraction[0]>='0') { /* we're moving */
		actcode = atoi((char *)&useraction);
		if (actcode>Stats.nr_actions || !actcode)
		 printf("sorry, action [%d] does not exist\n", actcode);
		else {
		 Action = ExecMove(actcode);
		 Stats = getstats();	/* need an update */
		 printf("Executed %s action: ", Action.atype==input?"input":"output");
		 OwnPrtAction(stdout,Action);
		 printf("\n");
		}	
	} else
	switch (useraction[0]) {
	case 'p': Analyse(); /* analyses the current superstate */
		  Stats = getstats();	/* get results */
		  for (i=1;i<=Stats.nr_actions;i++) {
			printf("%3d. ",i);
			PrtAction(stdout,i);printf("\n");
			}
		  printf("%s",Stats.delta? \
			"\n\tDelta suspension in current superstate\n":"");
		  break;

	case 'd': ExecDelta();
		  Stats = getstats();
		  printf("\t * Delta move executed * \n");
		  break;
	case '#': PrtStats(stdout,Stats);
		  break;
	case 'h': printf("Possible commands are:\n");
		  printf("  p  - give prediction\n");
		  printf("snnn - show data action nnn (=number)\n");
		  printf("  d  - move a deltatransition\n");
		  printf("  #  - shows Trojka engine statistics\n");
		  printf("  h  - shows this briefly help\n");
		  printf("  q  - quits trojka\n");
		  printf(" nnn - execute action nnn (=number)\n");
		  break;
	case 'q': return; break;
	case 's': if (useraction[1]<='9' && useraction[1]>='0') { 
			actcode = atoi((char *) &useraction[1]);
			if (actcode>Stats.nr_actions || !actcode)
				printf("sorry, action [%d] does exist\n", actcode);
			else {
		 		Action = GetAction(actcode);
		 		printf("Data %s action: ",
				Action.atype==input?"input":"output");
		 		OwnPrtAction(stdout,Action);
		 		printf("\n");
			} }
		  break;
	default : printf("unknown command\n"); break;
	}
	} /* while */
}

int
main(int argc, char *argv[])
{
	
	/* setting defaults critical parameters */
	Settings.logfile = NULL;
	Settings.searchfile = NULL;
	Settings.hashnumber = 0;	/* like pan */
	Settings.maxdepth = 100;
	Settings.ssize = 18;		/* like pan */
 
        while (argc > 1) {
	if (argv[1][0] != '-') ia_usage(stderr); else
               switch (argv[1][1]) {
		case 'k': Settings.logfile = fopen(&argv[1][2], "w");break;
		case 's': Settings.searchfile = fopen(&argv[1][2], "w");
			  break;
                case 'h': Settings.hashnumber = atoi(&argv[1][2]); break;
                case 'l': Settings.seed = atoi(&argv[1][2]); break;
                case 'm': Settings.maxdepth = atoi(&argv[1][2]); break;
                case 'w': Settings.ssize = atoi(&argv[1][2]); break;
                default : ia_usage(stderr); break;
                }
                argc--; argv++;
        }
        Init_Trojka(Settings);
	userinterface();
}
