/***** trojka: trojka.c *****/

/* Copyright (c) 1991,1995 by AT&T Corporation.  All Rights Reserved.     */
/* This software is for educational purposes only.                        */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.att.com          */

/*	Modified code and additional routines for the trojka analyser	*/
/*	No guarantee is expressed or implied!				*/
/*									*/
/* No permission is given to distribute the modified code and		*/
/* additional routines without permission of the author.		*/
/*									*/
/* Modified code and additional routines written by Rene G. de Vries	*/
/*									*/
/* trojka interface							*/
/* version 1.0 :	Tue Jun 11 20:32:02 MET DST 1996		*/
/* version 1.7.4:	Thu Mar  4 14:54:07 MET 1999			*/
/* version 1.8.0:	Tue Jun 22 17:40:03 MET DST 1999		*/
/* version 1.9.0:	Fri Jan 28 18:40:09 CET 2000			*/
/* removed DriverVersion; handling of version is done at top; version.h */
/*----------------------------------------------------------------------*/

#include <stdarg.h>
#include <stdio.h>
#include <time.h>

#ifdef TANSTATIC
#include "trojka.h"
#include "tan-sman.h"
#include "tan-action.h"
#endif

/*
 * globals: lf = logfile
*/

FILE	*lf, *sf, *efd;

static char *title[] = {
	"\n\n",
	"\t***     "
	DriverVersion
	"      ***",
	"\t***  Written by Rene G. de Vries based on SPIN-code by    *** ",
	"\t***       Gerard J. Holzmann (AT&T Corporation)           *** ",
	"\n\n",
	0

	};

StateSetT *SSvector = NULL;
uchar delta;

/* === Internal procedures and functions === */
void
PrtStr (FILE *out, char* c[])
{
	int i;
        for (i = 0; c[i]; i++)
        {       fprintf(out, c[i]);
                fprintf(out, "\n");
        }
}


/* === Exported functions and procedures === */

Trojka_statT
getstats(void)
{
	Trojka_statT Result;

	/* Global parameters */
	Result.nr_vector = createdstates;	/* tan-sman */
	Result.nr_svector = createdstructs;	/* tan-sman */
	Result.memcnt = memcnt;			/* tan.c */

	/* Parameters after each trojka_engine command */
	Result.nr_cuts = cut;			/* tan-engine */
	Result.nr_analysed = analyse;		/* tan-engine */
	Result.nr_matched = matched;		/* tan-engine */
	Result.nr_supers =  SSnrel(&SSvector);
	Result.nr_executed = executed;		/* tan-engine */
	Result.depth = reacheddepth;		/* tan-engine */
	Result.delta = delta;
	Result.nr_actions = actionident;	/* tan-action */

	return Result;
}

void
Init_Trojka( Trojka_SettingsT Settings )       /* installs trojka_engine */
{
	Stack	*reset_stack;	/* knife to cutoff old context */
	Svtack	*reset_svtack;	/* knife to cutoff old context */
	
	Maxdiepte = Settings.maxdepth;
	if (Settings.maxdepth > maxdepth)
		maxdepth = Settings.maxdepth;
	ssize = Settings.ssize;
	mask = ((1<<ssize)-1);  /* hash init, to be used later on.*/
        Tinit();	/* initialize hashtable */
        trail = (Trail *) emalloc((maxdepth+2)*sizeof(Trail));

	memset((char *)&now,0,sizeof(State));	/* clear state */
	vsize=sizeof(State)-VECTORSZ;

	/* initialize stack to undo deletes of processes and queues */
	reset_stack = stack   = ( Stack *) emalloc(sizeof(Stack));
	/* initialize stack to save to undo-info receives from queues*/
        reset_svtack = svtack  = (Svtack *) emalloc(sizeof(Svtack)); 
	/* a place to point for Pptr of non-running procs: */
        noptr   = (uchar *) emalloc(Maxbody * sizeof(char));

	settable();	/* init transition table */
	all_retrans();	/* optimize transitions table (required) */

	iniglobals();
	UnBlock;	/* disable rendez-vous */

	depth = 0;
	trpt = &trail[depth];	/* init the pointer trpt */
	
	active_procs();	/* set initial procs in state */

	/* set rootstate in the superstatevecor */	
	nstates = 0; /* Just make sure */

	IncludeState(&SSvector);	/* set root superstate */
	/* initial no atomic action TODO */

	/* Passings settings to trojka-engine */

	lf = Settings.logfile;
	sf = Settings.searchfile;
	HASH_NR = Settings.hashnumber;
	srandom(Settings.seed);
	
	PrtStr(stderr, title);

	if (lf && lf != stderr) PrtStr(lf, title);


}	

void
Analyse(void)
{
	ClearAction();	/* new search for actions */
	do_with_spec(&SSvector,ADD, &delta);		
} 


void
ExecDelta()
{	ClearAction();
	do_with_spec(&SSvector, DMOVE,  &delta);
}
   
ActionT
ExecMove(int move)
{	ActionT Action;	
        Action = SetAction(move);
	ClearAction();
        do_with_spec(&SSvector,AMOVE, &delta);
	return Action;
}

void  logprintf(char *format, ...)
{
   int sock, len, x;
   char* tijd;
   va_list args;

   time_t tijd_t;
   tijd_t=time(NULL);
   tijd=ctime(&tijd_t);
   tijd[strlen(tijd)-1]=(char) 0;
  
   va_start(args, format);
   if (lf) {fprintf(lf,"[%s] ",tijd); vfprintf(lf, format, args);}
   va_end(args);
}



