#! /bin/sh
 
##############################################################################
#                              T O R X _ O P E N
#-----------------------------------------------------------------------------
#   INRIA - Unite de Recherche Rhone-Alpes
#   655, avenue de l'Europe
#   38330 Montbonnot Saint Martin
#   FRANCE
#-----------------------------------------------------------------------------
#   Module             :       torx_open
#   Auteur             :       Axel Belinfante
#   (based on ncrl_open by Hubert GARAVEL and Jaco van de POL)
#   Version            :       1.0
#   Date               :       2006-02-10
#############################################################################/

set -e

COMMAND=`basename $0`
ARCH=`$CADP/com/arch`
CC=$CADP/src/com/cadp_cc

BIN=$CADP/bin.$ARCH
SRC=$CADP/src/open_caesar
LIB=$CADP/bin.$ARCH

# recording the command-line for OCIS
 
OPEN_CAESAR_COMMAND="$0 $*"
export OPEN_CAESAR_COMMAND

# the following variables are initialised by "configure"
TORX_VERSION=@VERSION@
libdir=@libdir@
TORX_LIB=${libdir}

if test "$TORX_DEBUG" != ""
then
	# debugging mode used internally by the TORX team
	TORX_LD_FLAGS="-I$CADP/incl ./torx_open.c"
else
	TORX_LD_FLAGS="$TORX_LIB/torx_open.o"
fi

# phase 1: parsing the options

while true
do
	if test $# -lt 1
	then
		echo "$COMMAND: missing torx-explorer-program argument"
		exit 1
	fi
	case $1 in
		-version )
			echo $TORX_VERSION
			exit 0
			;;
		* )
			break
			;;
	esac
done

# phase 2: copying the program+args (the value of $TORXPROG may contain spaces)

TORXPROG="$1"
shift

# phase 3: passing the TORXPROG to the C library

TORX_OPEN_EXPLORER="$TORXPROG"
export TORX_OPEN_EXPLORER

# phase 4: parsing the options for cc(1) and preparing other options

CPPFLAGS="-I. -I$CADP/incl -I$SRC"
CFLAGS=""
LDFLAGS="$TORX_LD_FLAGS -L$LIB -lcaesar"

while true
do
	if test $# -lt 1
	then
		echo "$COMMAND: missing C filename argument"
		exit 1
	fi
	case $1 in
		-[DI]* ) CPPFLAGS="$CPPFLAGS $1"
			 shift
			 ;;
		-[lL]* ) LDFLAGS="$LDFLAGS $1"
			 shift
			 ;;
		-* ) CFLAGS="$CFLAGS $1"
		     shift
		     ;;
		*) break
		   ;; 
	esac
done

# phase 5: parsing the ``.c'', ``.o'', or ``.a'' filename, possibly looking in
#          $SRC or $BIN if the file is not present in the current directory

NAME=$1
shift

case "$NAME" in
	*.c )	
		if test -f $NAME
		then
			USER="$NAME"
		elif test -f $SRC/$NAME
		then
			USER="$SRC/$NAME"
		else
			echo "$COMMAND: file \`\`$NAME'' does not exist"
			exit 1
		fi
		;;

	*.o | *.a )
		if test -f $NAME
		then
			USER="$NAME"
		elif test -f $BIN/$NAME
		then
			USER="$BIN/$NAME"
		else
			echo "$COMMAND: file \`\`$NAME'' does not exist"
			exit 1
		fi
		;;

	* )
		if test -f $NAME.c
		then
			USER="$NAME.c"
		elif test -f $SRC/$NAME.c
		then
			USER="$SRC/$NAME.c"
		elif test -f $NAME.a
		then
			USER="$NAME.a"
		elif test -f $BIN/$NAME.a
		then
			USER="$BIN/$NAME.a"
		elif test -f $NAME.o
		then
			USER="$NAME.o"
		elif test -f $BIN/$NAME.o
		then
			USER="$BIN/$NAME.o"
		else
			echo "$COMMAND: cannot locate \`\`$NAME''"
			exit 1
		fi
esac

if test "$USER" != "$NAME"
then
	echo "$COMMAND: using \`\`$USER''"
fi

# phase 6: updating compiler options, if necessary

case "$USER" in
	*.c )
		CPPFLAGS_AUX=`grep '@(#)CPPFLAGS' $USER | sed -e 's/.*".*=\(.*\)".*/\1/'`
		CFLAGS_AUX=`grep '@(#)CFLAGS' $USER | sed -e 's/.*".*=\(.*\)".*/\1/'`
		LDFLAGS_AUX=`grep '@(#)LDFLAGS' $USER | sed -e 's/.*".*=\(.*\)".*/\1/'`
		LD_LIBRARY_PATH_AUX=`grep '@(#)LD_LIBRARY_PATH' $USER | sed -e 's/.*".*=\(.*\)".*/\1/'`
		;;
	*.o | *.a )
		LC_CTYPE="C" ; export LC_CTYPE
		CPPFLAGS_AUX=`strings $USER 2> /dev/null | grep '@(#)CPPFLAGS' | sed -e 's/.*=//'`
		CFLAGS_AUX=`strings $USER 2> /dev/null | grep '@(#)CFLAGS' | sed -e 's/.*=//'`
		LDFLAGS_AUX=`strings $USER 2> /dev/null | grep '@(#)LDFLAGS' | sed -e 's/.*=//'`
		LD_LIBRARY_PATH_AUX=`strings $USER 2> /dev/null | grep '@(#)LD_LIBRARY_PATH' | sed -e 's/.*=//'`
		;;
esac

CPPFLAGS_AUX=`eval echo $CPPFLAGS_AUX`
if test "$CPPFLAGS_AUX" != ""
then
	CPPFLAGS="$CPPFLAGS $CPPFLAGS_AUX"
fi

CFLAGS_AUX=`eval echo $CFLAGS_AUX`
if test "$CFLAGS_AUX" != ""
then
	CFLAGS="$CFLAGS $CFLAGS_AUX"
fi

LDFLAGS_AUX=`eval echo $LDFLAGS_AUX`
if test "$LDFLAGS_AUX" != ""
then
	LDFLAGS="$LDFLAGS $LDFLAGS_AUX"
fi

LD_LIBRARY_PATH_AUX=`eval echo $LD_LIBRARY_PATH_AUX`
if test "$LD_LIBRARY_PATH_AUX" != ""
then
	LD_LIBRARY_PATH="$LD_LIBRARY_PATH_AUX:$LD_LIBRARY_PATH"
	export LD_LIBRARY_PATH
fi

# phase 7: compilation and link edition of $USER

case "$USER" in
	*.c )
		PROG=`basename $USER .c`
		$CC $CPPFLAGS $CFLAGS -c $USER -o $PROG.o
		$CC $CFLAGS $PROG.o $LDFLAGS -o $PROG
		rm -f $PROG.o
		;;
	*.o )
		PROG=`basename $USER .o`
		$CC $CFLAGS $USER $LDFLAGS -o $PROG
		;;
	*.a )
		PROG=`basename $USER .a`
		$CC $CFLAGS $USER $LDFLAGS -o $PROG
		;;
esac

# phase 8: program execution (here, "$@" denotes the arguments, if any, to
# be passed to the program, and $# denotes the number of these arguments)

if test "$#" -eq 0
then
	echo "$COMMAND: running \`\`$PROG'' for \`\`$TORXPROG''"
else
	echo "$COMMAND: running \`\`$PROG $@'' for \`\`$TORXPROG''"
fi
./$PROG "$@"

exit 0

