#! @PERL@

# usage: log2anifsm [ -p pattern ] [-p pattern] ... [ files ]

require 5.002;
use Getopt::Long;

$debug = 0;

$| = 1; # do not buffer!

%optcl = ();
$result = GetOptions(
			\%optcl,
			"p=s@",
		);

$patlistref = $optcl{"p"} if (defined $optcl{"p"});
@pat = @{$patlistref};

# ======================================

initialize();

loop:
while(<>) {
	chomp;
	/^CONFIG/ && do handle_config($_);
	/^ABSTRACT/ && do handle_action($_ , 0);
	/^STATEID/ && do {
		foreach $pat (@pat) {
			# print STDERR "trying: ($pat)($_)\n";
			if (! /$pat/) {
				print STDERR "no match: ($pat)($_)\n" if $debug;;
				next loop;
			}
		}
		handle_stateid($_ , 0);
	}
}

finalize();

sub initialize {
}

sub handle_config {
	my($line) = @_;
	print STDERR "line: ($line) " if $debug;
	my(@words) = split(/ /, $line);
	my($word, $keyword);
	print STDERR "words: (@words) " if $debug;
	$word = shift(@words); # shift out CONFIG
	if ($word eq "CONFIG") {
		$keyword = shift(@words); # shift out config keyword
		if ($keyword eq "CONFIGFILE") {
			$cfgfilename = shift(@words);
			$cfgfilename =~ s/\.if$//;
		} elsif ($keyword eq "PRIMER") {
			$cfgprimer = '\\n';
			$cfgprimer .= shift(@words);
		} elsif ($keyword eq "MUTANT") {
			$cfgmutant = '\\n';
			$cfgmutant .= shift(@words);
		}
	} elsif ($word eq "CONFIGEND") {
		$result = 'dot node ___all shape box label ';
		$result .= $cfgfilename;
		$result .= $cfgprimer;
		$result .= $cfgmutant;
		$result .= "\n";
		print $result;
	}
}

sub handle_action {
	my($line, $expected) = @_;
	print STDERR "line: ($line) " if $debug;
	my(@words) = split(/ /, $line);
	my($ignore, $stepnr, $iokind_chan_pco, $suspension, $action,
	   $iokind,$chan_pco_opt, $chan_pco, $chan, $pco, $inst);
	$ignore = shift(@words); # shift out ABSTRACT
	$stepnr = shift(@words);
	$iokind_chan_pco = shift(@words);
	$suspension = pop(@words);
	$suspension = strip_parentheses($suspension);
	$action = join(" ", @words);
	$action = strip_parentheses($action);
	if ($theSource eq '' ) { # no STATEID
		$nostateids = 1;
		startgraph($stepnr - 1);
		colorgraph();
	}
	if ($iokind_chan_pco =~ /^(input|output)(\(([^\)]*)\))?$/) {
		$iokind=$1;
		$chan_pco_opt = $2;
		$chan_pco = $3;
		if ($chan_pco =~ /^([^:]+):([^:]+)$/) {
			$chan = $1;
			$pco = $2;
			add_nonsuspension($stepnr, $iokind, $action );
		} elsif ($chan_pco =~ /^([^:]+)$/) {
			$chan = $1;
			$pco = '';
			$inst = '('.$chan.')';
			if ($suspension eq "1") {
				add_suspension($stepnr, $iokind);
			} else {
				msg("warning: no pco for non-suspension action nr: $stepnr");
				add_nonsuspension($stepnr, $iokind, $action );
			}
		} else {
			$chan = '';
			$pco = '';
			$inst = 'tester';
			if ($suspension eq "1") {
				add_suspension($stepnr, $iokind);
			} else {
				add_nonsuspension($stepnr, $iokind, $action );
			}
		}
		if ($nostateids) { # no STATEID
			extendgraph($stepnr);
			colorgraph();
		}
	} else {
		msg("parse error: could not parse iokind/channel/pco for action nr: $stepnr");
	}
}

sub add_suspension {
	my($stepnr, $iokind) = @_;
	print STDERR "add_suspension ($stepnr, $iokind)" if $debug;
	if ($iokind eq "input") {
		$theAction='suspension?';
	} elsif ($iokind eq "output") {
		$theAction='quiescence!';
	} else {
		msg("unknown action kind \"$iokind\" for nr: $stepnr");
	}
}
sub add_nonsuspension {
	my($stepnr, $iokind, $action) = @_;
	my($i, $iut, $actpfx);
	print STDERR "add_nonsuspension ($stepnr, $iokind, $action)" if $debug;
	if ($iokind eq "input") {
		$theAction=$action;
	} elsif ($iokind eq "output") {
		$theAction=$action;
	} else {
		msg("unknown action kind \"$iokind\" for nr: $stepnr");
	}
}

# STATEID 0 SPEC super 0 init 0
sub handle_stateid {
	my($line, $expected) = @_;
	print STDERR "line: ($line) " if $debug;
	my(@words) = split(/ /, $line);
	my($ignore, $stepnr, $role, $super, $k);
	my($word, $init, $trans);
	while(@words > 0) {
		$word = shift(@words);
		if ($word eq "init") {
			$init = shift(@words);
		} elsif ($word eq "trans") {
			$trans = shift(@words);
		} elsif ($word eq "super") {
			$super = shift(@words);
		}
	}
	# ignore the rest
	if ($theSource ne '' && $theAction ne '') {
		extendgraph($super);
	} else {
		startgraph($super);
	}
	colorgraph();
}

sub startgraph {
	my($snr) = @_;

	$theInit = $snr;
	print 'dot edge ___all->S'.$theInit.' name enil'."\n";
	$new = '-n S'.$theInit.' -e enil';
	$theSource = $snr;
}
sub extendgraph {
	my($snr) = @_;

	$theDest = $snr;
	$k = join($;, ($theSource, $theAction, $theDest));
	$trans{$k}++;
	$state{$theSource} = 1;
	$state{$theDest} = 1;
	if ($trans{$k} == 1) {
		$ntrans++;
		$transnr{$k} = $ntrans;
		$enr = $ntrans;
		print 'dot edge S'.$theSource.'->S'.$theDest.' label "'.$theAction.'" name e'.$enr."\n";
	} else {
		$enr = $transnr{$k};
	}
	$new = '-n S'.$theDest.' -e e'.$enr;
	$theSource = $snr;
}
sub colorgraph {
	if ($prev ne '') {
		print '-c grey '.$prev.' -c red '.$new."\n";
	} else {
		print $new."\n";
	}
	$prev = $new;
}

sub finalize {
}


sub msg {
	my($s) = @_;
	print STDERR "log2msc: ".$s."\n";
}

sub strip_parentheses {
	my($s) = @_;
	if ($s =~ /^\((.*)\)$/) {
		return $1;
	} else {
		return $s;
	}
}

sub strip_quotes {
	my($s) = @_;
	if ($s =~ /^"(.*)"$/) {
		return $1;
	} else {
		return $s;
	}
}
